# Copyright 2013 Canonical Ltd.
# This software is licensed under the GNU Affero General Public License
# version 3 (see the file LICENSE).

import logging

import sst.actions
import u1testutils.sso.sst.pages

from base64 import b32decode
from urlparse import urlparse, parse_qs

from u1testutils.sst import log_action


class LoggedUserPage(u1testutils.sst.Page):

    def __init__(self, open_page=False):
        super(LoggedUserPage, self).__init__(open_page)
        self.menu = Menu()

    def get_user_name(self):
        user_name_element = sst.actions.get_element_by_css(
            '*[data-qa-id="user_name"]')
        return user_name_element.text

    @log_action(logging.info)
    def log_out(self):
        link = sst.actions.get_element_by_css('*[data-qa-id="logout_link"]')
        sst.actions.click_element(link)
        return YouHaveBeenLoggedOut()


class Menu(object):
    """The menu of all the SSO pages with a user logged in."""

    @log_action(logging.info)
    def go_to_authentication_devices(self):
        link = sst.actions.get_element_by_css('*[data-qa-id="devices_link"]')
        sst.actions.click_link(link)
        return YourAuthenticationDevices()

    @log_action(logging.info)
    def go_to_your_account(self):
        link = sst.actions.get_element_by_css('*[data-qa-id="account_link"]')
        sst.actions.click_link(link)
        return YourAccount()


class LogIn(u1testutils.sso.sst.pages.LogIn):
    """Log in page of the Ubuntu Single Sign On website.

    Extends the class in u1testutils to return the Your Account page  with all
    the actions available.

    """

    @log_action(logging.info)
    def log_in_to_site_recognized(self, user=None):
        """Fill the log in form and continue to the site that requested it.

        Keyword arguments:
        user -- The user credentials. It must have the attributes email and
            password. If None is passed as the user, it means that the user
            has already started session on the identity provider and it's not
            necessary to enter the credentials again.

        """
        self._log_in(user)
        return YourAccount()

    @log_action(logging.info)
    def log_in_to_site_not_recognized(self, user=None):
        """Fill the log in form and continue to the next step.

        As the site is not recognized, the next step is the page where the
        user can select the information that will be sent to the site.

        Keyword arguments:
        user -- The user credentials. It must have the attributes email and
            password. If None is passed as the user, it means that the user
            has already started session on the identity provider and it's not
            necessary to enter the credentials again.

        """
        self._log_in(user)
        return u1testutils.sso.sst.pages.SiteNotRecognized()

    @log_action(logging.info)
    def log_in_with_two_factor(self, user=None):
        """Fill the log in form and continue to two factor authentication."""
        self._log_in(user)
        return TwoFactorVerify()


class TwoFactorVerify(LoggedUserPage):
    """The page where customer is prompted for a 2FA key"""

    url_path = '/two_factor_auth.*'
    is_url_path_regex = True
    qa_anchor = 'two_factor_verify'

    @log_action(logging.info)
    def two_factor_authentication(self, one_time_password):
        self._fill_one_time_password_form(one_time_password)
        self._click_authenticate()
        return YourAccount(self.get_user_name())

    @log_action(logging.info)
    def authenticate_2F_with_error(self, bad_password):
        self._fill_one_time_password_form(bad_password)
        self._click_authenticate()
        return TwoFactorVerify()

    @log_action(logging.info)
    def get_form_errors(self):
        try:
            one_time_password_error = sst.actions.get_element(
                css_class='error').text
        except AssertionError:
            one_time_password_error = None
        return one_time_password_error

    @log_action(logging.info)
    def get_sync_link(self):
        try:
            sync_link = self._get_sync_selector()
        except AssertionError:
            sync_link = None
        return sync_link

    @log_action(logging.info)
    def click_sync_link(self):
        sst.actions.click_element(self._get_sync_selector())
        return TwoFactorSync()

    def _get_sync_selector(self):
        return sst.actions.get_element_by_css('*[data-qa-id="sync-2fa-link"]')

    def _fill_one_time_password_form(self, one_time_password):
        sst.actions.write_textfield('id_oath_token', one_time_password)

    def _click_authenticate(self):
        auth_button = sst.actions.get_element_by_css(
            '*[data-qa-id="auth_button"]')
        sst.actions.click_button(auth_button)


class TwoFactorSync(LoggedUserPage):
    """The page where customer is shown 2FA keys to sync her device."""

    url_path = '/sync_2fa'
    title = 'Log in'

    @log_action(logging.info)
    def get_token_fields(self):
        return sst.actions.get_elements_by_css('.oath-token input')

    @log_action(logging.info)
    def write_token(self, field, token):
        sst.actions.write_textfield(field, token)

    @log_action(logging.info)
    def synchronize(self):
        submit_css = '*[data-qa-id="sync_button"]'
        submit = sst.actions.get_element_by_css(submit_css)
        sst.actions.click_button(submit)


class YouAreSuspended(u1testutils.sst.Page):
    """This is the page you are given if you fail 2 Factor 18 times."""

    url_path = '/two_factor_auth.*'
    is_url_path_regex = True
    qa_anchor = 'account_suspended'


class YouHaveBeenLoggedOut(u1testutils.sso.sst.pages.YouHaveBeenLoggedOut):
    """The logged out landing page of the Ubuntu Single Sign On website."""

    @log_action(logging.info)
    def click_log_in_or_create_account(self):
        link = sst.actions.get_element_by_css('*[data-qa-id="login_link"]')
        sst.actions.click_link(link)
        return LogIn()


class YourAccount(LoggedUserPage):
    """Your account page of the Ubuntu Single Sign On website.

    Extends the class in u1testutils to use the menu with all the actions
    available.

    """

    url_path = '/'
    qa_anchor = 'edit_account'

    def _set_twofactor(self, always_required):
        radio = sst.actions.get_element(tag="input", name="twofactor_required",
                                        value=always_required)
        sst.actions.set_radio_value(radio)

    @log_action(logging.info)
    def set_twofactor_to_always_required(self):
        self._set_twofactor(always_required=True)

    @log_action(logging.info)
    def set_twofactor_to_required_as_needed(self):
        self._set_twofactor(always_required=False)

    @log_action(logging.info)
    def click_update_button(self):
        sst.actions.click_button(sst.actions.get_element_by_css(
            '*[data-qa-id="update_preferences"]'))

    def is_device_options_displayed(self):
        try:
            device_options = sst.actions.get_element_by_css(
                '*[data-qa-id="authentication_devices"]')
            return device_options.is_displayed()
        except AssertionError:
            return False


class YourEmailAddresses(LoggedUserPage):

    url_path = '/+emails'
    qa_anchor = 'account_emails'

    @log_action(logging.info)
    def add_email(self, email_address):
        sst.actions.write_textfield('id_newemail', email_address)
        sst.actions.click_button(
            sst.actions.get_element_by_css('*[data-qa-id="add_new_email"]'))

    @log_action(logging.info)
    def delete_email(self, email_address):
        verified_email_link_css = (
            'a[data-qa-id="delete_verified_%s"]' % email_address)
        unverified_email_link_css = (
            'a[data-qa-id="delete_unverified_%s"]' % email_address)
        try:
            remove_link = sst.actions.get_element_by_css(
                verified_email_link_css)
        except:
            try:
                remove_link = sst.actions.get_element_by_css(
                    unverified_email_link_css)
            except:
                # email is not found, or cannot be deleted
                return
        sst.actions.click_link(remove_link)
        return DeleteEmail()

    def get_form_errors(self):
        try:
            email_error = sst.actions.get_element(
                css_class='error').text
        except AssertionError:
            email_error = None
        return email_error


class EnterConfirmationCode(LoggedUserPage):

    url_path = '/+enter_token'
    qa_anchor = 'enter_token'

    @log_action(logging.info)
    def confirm(self, confirmation_code, email_address):
        self._fill_confirmation_form(confirmation_code, email_address)
        self._continue()
        return u1testutils.sso.sst.pages.CompleteEmailValidation()

    def _fill_confirmation_form(self, confirmation_code, email_address):
        sst.actions.write_textfield('id_confirmation_code', confirmation_code)
        sst.actions.write_textfield('id_email', email_address)

    def _continue(self):
        sst.actions.click_button(sst.actions.get_element_by_css(
            '*[data-qa-id="confirm_enter_token"]'))

    @log_action(logging.info)
    def confirm_with_error(self, confirmation_code, email_address):
        self._fill_confirmation_form(confirmation_code, email_address)
        self._continue()
        return self


class DeleteEmail(LoggedUserPage):

    url_path = '/+remove-email'
    qa_anchor = 'delete_email'

    @log_action(logging.info)
    def confirm(self):
        confirm_button = sst.actions.get_element_by_css(
            '*[data-qa-id="confirm_delete_email"]')
        sst.actions.click_button(confirm_button)
        return YourEmailAddresses()


class ResetPassword(u1testutils.sso.sst.pages.PageWithAnonymousSubheader):

    url_path = '/+forgot_password'
    qa_anchor = 'forgot_password_step_1'

    @log_action(logging.info)
    def request_password_reset(self, email_address):
        sst.actions.write_textfield('id_email', email_address)
        sst.actions.click_button(sst.actions.get_element_by_css(
            '*[data-qa-id="send_password_reset_token"]'))


class ResetPasswordStep3(u1testutils.sso.sst.pages.PageWithAnonymousSubheader):

    url_path = '/token/.+/\+resetpassword/.+@.+'
    is_url_path_regex = True
    qa_anchor = 'reset_password'


class LogInFromRedirect(u1testutils.sso.sst.pages.LogIn):

    url_path = '/.*/\+decide'
    is_url_path_regex = True


# Devices pages.

class YourAuthenticationDevices(LoggedUserPage):

    url_path = '/device-list'
    qa_anchor = 'device_list'

    def get_devices(self):
        """Get a list with the name of the devices added by the user."""
        if sst.actions.exists_element(id='device-list'):
            get_text = lambda x: x.text
            return map(
                get_text, sst.actions.get_elements_by_css(
                    '#device-list td.name'))
        else:
            return []

    @log_action(logging.info)
    def add_new_authentication_device(self):
        add_link = sst.actions.get_element_by_css(
            '*[data-qa-id="add_new_device"]'
        )
        sst.actions.click_link(add_link)
        return AddNewAuthenticationDevice()

    @log_action(logging.info)
    def delete_authentication_device(self, device_index=0):
        delete_link = sst.actions.get_elements_by_css(
            '*[data-qa-id="delete_device"]')
        sst.actions.click_link(delete_link[device_index])
        return DeleteAuthenticationDevice()

    @log_action(logging.info)
    def rename_authentication_device(self, device_index=0):
        rename_link = sst.actions.get_elements_by_css(
            '*[data-qa-id="rename_device"]')
        sst.actions.click_link(rename_link[device_index])
        return RenameAuthenticationDevice()

    def is_warning_displayed(self):
        try:
            warning_message = sst.actions.get_element_by_css(
                '#missing_backup_device')
            return warning_message.is_displayed()
        except AssertionError:
            return False


class DeleteAuthenticationDevice(LoggedUserPage):

    url_path = '/device-removal/.*'
    is_url_path_regex = True
    qa_anchor = 'device_removal'

    @log_action(logging.info)
    def cancel(self):
        cancel_link = sst.actions.get_element_by_css(
            '*[data-qa-id="cancel_deleting_this_device"]'
        )
        sst.actions.click_link(cancel_link)
        return YourAuthenticationDevices()

    @log_action(logging.info)
    def confirm_delete_device(self):
        delete_anchor = sst.actions.get_element_by_css(
            '*[data-qa-id="delete_this_device"]')
        sst.actions.click_element(delete_anchor)
        return YourAuthenticationDevices()


class RenameAuthenticationDevice(LoggedUserPage):

    url_path = '/device-rename/.*'
    is_url_path_regex = True
    qa_anchor = 'device_rename'

    @log_action(logging.info)
    def cancel(self):
        cancel_link = sst.actions.get_element_by_css(
            '*[data-qa-id="cancel_renaming_this_device"]'
        )
        sst.actions.click_link(cancel_link)
        return YourAuthenticationDevices()

    @log_action(logging.info)
    def rename_device(self, new_name):
        self._fill_device_form_and_click(new_name)
        return YourAuthenticationDevices()

    @log_action(logging.info)
    def rename_device_with_errors(self, new_name):
        self._fill_device_form_and_click(new_name)
        return self

    def _fill_device_form_and_click(self, new_name):
        new_name_field = sst.actions.get_element(tag='input', name='name')
        sst.actions.write_textfield(new_name_field, new_name)
        rename_button = sst.actions.get_element_by_css(
            '*[data-qa-id="rename_this_device"]')
        sst.actions.click_element(rename_button)

    def get_form_errors(self):
        try:
            name_error = sst.actions.get_element(id='name-error').text
        except AssertionError:
            name_error = None
        return name_error


class AddNewAuthenticationDevice(LoggedUserPage):

    url_path = '/device-addition'
    qa_anchor = 'device_addition'

    @log_action(logging.info)
    def cancel(self):
        cancel_link = sst.actions.get_element_by_css(
            '*[data-qa-id="cancel_add_device_from_type"]')
        sst.actions.click_link(cancel_link)
        return YourAuthenticationDevices()

    @log_action(logging.info)
    def get_selected_device(self):
        identifiers = ['type_google', 'type_yubi', 'type_generic',
                       'type_paper']
        for identifier in identifiers:
            radio = sst.actions.get_element(id=identifier)
            if radio.is_selected():
                return identifier

    def is_paper_device_displayed(self):
        try:
            paper_device_radio = sst.actions.get_element(id='type_paper')
            return paper_device_radio.is_displayed()
        except AssertionError:
            return False

    @log_action(logging.info)
    def add_generic_device(self):
        self._add_device('type_generic')
        return AddGenericDevice()

    @log_action(logging.info)
    def add_google_device(self):
        self._add_device('type_google')
        return AddGoogleDevice()

    @log_action(logging.info)
    def add_yubikey_device(self):
        self._add_device('type_yubi')
        return AddYubikeyDevice()

    @log_action(logging.info)
    def add_paper_device(self):
        self._add_device('type_paper')
        return PaperDevice()

    def _add_device(self, device_radio_identifier):
        sst.actions.set_radio_value(device_radio_identifier)
        add_button = sst.actions.get_element_by_css(
            '*[data-qa-id="add_device_from_type"]')
        sst.actions.click_button(add_button)


class AddDevice(LoggedUserPage):

    url_path = '/device-addition'
    qa_anchor = 'device_addition'

    @log_action(logging.info)
    def add_device(self, name, one_time_password):
        self._fill_device_form(name, one_time_password)
        self._click_add_device()
        return YourAuthenticationDevices()

    def _fill_device_form(self, name, one_time_password):
        if name is not None:
            sst.actions.write_textfield(self._get_name_text_field(), name)
        if one_time_password is not None:
            sst.actions.write_textfield(
                self._get_one_time_password_text_field(), one_time_password)

    def _click_add_device(self):
        add_button = sst.actions.get_element_by_css(
            '*[data-qa-id="confirm_add_device"]')
        sst.actions.click_button(add_button)

    @log_action(logging.info)
    def add_device_with_errors(self, name, one_time_password):
        self._fill_device_form(name, one_time_password)
        self._click_add_device()
        return self

    def _get_name_text_field(self):
        return sst.actions.get_element(tag='input', name='name')

    def _get_one_time_password_text_field(self):
        return sst.actions.get_element(tag='input', name='otp')

    def get_form_values(self):
        name = self._get_name_text_field().get_attribute('value')
        one_time_password_text_field = self._get_one_time_password_text_field()
        one_time_password = one_time_password_text_field.get_attribute('value')
        return name, one_time_password

    def get_form_errors(self):
        try:
            name_error = sst.actions.get_element(id='name-error').text
        except AssertionError:
            name_error = None
        try:
            one_time_password_error = sst.actions.get_element(
                id='otp-error').text
        except AssertionError:
            one_time_password_error = None
        return name_error, one_time_password_error

    @log_action(logging.info)
    def cancel(self):
        cancel_link = sst.actions.get_element_by_css(
            '*[data-qa-id="cancel_add_device"]')
        sst.actions.click_link(cancel_link)
        return YourAuthenticationDevices()


class AddGenericDevice(AddDevice):

    qa_anchor = 'generic_device_addition'

    def get_key(self):
        return sst.actions.get_element(name='hex_key').get_attribute('value')


class AddGoogleDevice(AddDevice):

    qa_anchor = 'google_device_addition'

    def get_key(self, email):
        return self._get_key_from_qrcode(email)

    def _get_key_from_qrcode(self, email):
        img = sst.actions.get_element(tag='img', css_class='qrcode')
        src = img.get_attribute('src')
        # check the url is well formed
        url = urlparse(src)
        assert url.scheme == 'https', "incorrect google charts protocol"
        msg = "incorrect google charts domain"
        assert url.netloc == 'chart.googleapis.com', msg
        qs = parse_qs(url.query)['chl'][0]
        otpauth = urlparse(qs)
        assert email in otpauth.path
        # python2.7.3 on quantal has a backport from 2.7 trunk (presumably
        # will be 2.7.4) and now urlparse correctly handles query string on
        # *all* url types
        if otpauth.query:
            # urlparse has handled query string
            query = otpauth.query
        else:
            # we need to handle query string parsing
            query = otpauth.path.split('?')[1]
        b32_key = parse_qs(query)['secret'][0]
        aes_key = b32decode(b32_key).encode('hex')
        return aes_key


class AddYubikeyDevice(AddDevice):

    qa_anchor = 'yubikey_device_addition'

    def get_key(self):
        return sst.actions.get_element(name='hex_key').get_attribute('value')

    # TODO make a new test_open_add_yubikey_page on the yubikey page object,
    # add two functions: is_warning_displayed and get_warning_message
    # and assert both of them on the new test. This test is really small and
    # gives no real value to a user, so it's a perfect candidate to be
    # converted into a django unit test.
    def assert_warning(self):
        # Check that the YubiKey warning is showing
        warning = ('Warning: The YubiKey is shipped with a credential in the '
                   'short-press slot')
        sst.actions.assert_element(css_class='warning', text_regex=warning)


class PaperDevice(LoggedUserPage):

    url_path = '/device-print/.*'
    is_url_path_regex = True
    qa_anchor = 'paper_device_addition'

    # TODO This test is really small and gives no real value to a user, so it's
    # a perfect candidate to be converted into a django unit test.
    def assert_codes_present(self):
        # codelist has list of N codes
        sst.actions.assert_element('*[data-qa-id="codelist"]')

    # TODO This test is really small and gives no real value to a user, so it's
    # a perfect candidate to be converted into a django unit test.
    def assert_print_button_visible(self):
        sst.actions.assert_element('*[data-qa-id="print_codes"]')

    @log_action(logging.info)
    def go_back_to_device_list(self):
        device_list_link = sst.actions.get_element_by_css(
            '*[data-qa-id="cancel_print_codes"]')
        sst.actions.click_link(device_list_link)
        return YourAuthenticationDevices()

    @log_action(logging.info)
    def generate_new_codes(self):
        generate_link = sst.actions.get_element_by_css(
            '*[data-qa-id="generate_codes"]')
        sst.actions.click_link(generate_link)
        return GenerateNewPaperCodes()

    def get_first_code(self):
        return sst.actions.get_elements_by_css(
            '*[data-qa-id="codelist"]')[0].text

    # TODO migrate the paper device functions below
    #def store_paper_device(self, name='Printable Backup Codes'):
    #    acceptance.devices.store_paper_device(name)

    #def update_paper_device(self, name='Printable Backup Codes'):
    #    acceptance.devices.update_paper_device(name)


class GenerateNewPaperCodes(LoggedUserPage):

    url_path = '/device-generate/.*'
    is_url_path_regex = True
    qa_anchor = 'generate_new_codes'

    @log_action(logging.info)
    def confirm_new_codes(self):
        confirm_button = sst.actions.get_element_by_css(
            '*[data-qa-id="confirm_new_codes"]')
        sst.actions.click_button(confirm_button)
        return PaperDevice()

    @log_action(logging.info)
    def cancel(self):
        cancel_link = sst.actions.get_element_by_css(
            '*[data-qa-id="cancel_new_codes"]')
        sst.actions.click_link(cancel_link)
        return PaperDevice()


class Consumer(u1testutils.sst.Page):

    url_path = '/consumer/'
    qa_anchor = 'consumer'

    @log_action(logging.info)
    def begin(self, mode='Setup', teams=None):
        self._select_mode(mode)
        if teams:
            self._select_teams(teams)
        self._click_begin()
        return LogInFromRedirect()

    def _select_mode(self, mode):
        if mode == 'Setup':
            setup_radio = sst.actions.wait_for(
                sst.actions.get_element_by_css, '*[data-qa-id="setup_mode"]')
            sst.actions.set_radio_value(setup_radio)
        elif mode == 'Immediate':
            raise NotImplementedError
        else:
            raise ValueError('Unknown mode {0}.'.format(mode))

    def _select_teams(self, teams):
        sst.actions.set_checkbox_value('id_teams', True)
        teams_to_request = ','.join(teams)
        sst.actions.write_textfield('id_request_teams', teams_to_request)

    def _click_begin(self):
        begin_button = sst.actions.get_element_by_css('*[data-qa-id="begin"]')
        sst.actions.click_button(begin_button)


class ConsumerFinish(Consumer):

    url_path = '/consumer/finish/'

    def get_teams(self):
        team_elements = sst.actions.wait_for(
            sst.actions.get_elements_by_css, '*[data-qa-id="teams"] > li')
        result = map(sst.actions.get_text, team_elements)
        return result
