# Copyright 2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import copy
import json
import os
import textwrap
from contextlib import contextmanager

from fabric.api import env, local, settings
from fabric.context_managers import hide

from .constants import PG_VERSION
from .django import get_django_settings, manage, syncdb


def createdb():
    """Create the database."""
    _set_postgres_environment()
    local("%(ENV)s %(BIN)s/createdb -U postgres -O postgres %(DATABASE)s" %
          env.postgres)


def dropdb(warn_only=False):
    """Remove the database."""
    _set_postgres_environment()
    if isinstance(warn_only, basestring):
        warn_only = warn_only.lower() == 'yes'
    with settings(warn_only=warn_only):
        local("%(ENV)s %(BIN)s/dropdb -U postgres %(DATABASE)s" % env.postgres)


def resetdb():
    """Drop and recreate then sync the database."""
    with settings(hide='warnings'):
        dropdb(warn_only=True)
    createdb()
    syncdb()
    setup_db_access()


def setup_postgresql_server():
    """Setup the PostgreSQL server."""
    _set_postgres_environment()
    local("mkdir -p %(DATA)s" % env.postgres)
    local("%(ENV)s %(BIN)s/initdb -A trust -D %(DATA)s" % env.postgres)
    config = textwrap.dedent("""
        fsync = off
        standard_conforming_strings = off
        escape_string_warning = off
    """)
    with open("%(DATA)s/postgresql.conf" % env.postgres, 'a') as pg_conf:
        pg_conf.write(config)
    start_database()
    setup_database()


def start_database():
    """Start the PostgreSQL server."""
    _set_postgres_environment()
    cmd = ['%(ENV)s', '%(BIN)s/pg_ctl', 'start', '-w', '-l',
        '%(HOST)s/postgresql.log', '-o', '"-F -k %(HOST)s -h \'\'"']
    local(' '.join(cmd) % env.postgres)


def stop_database():
    """Stop the PostgreSQL server."""
    _set_postgres_environment()
    cmd = ['%(ENV)s', '%(BIN)s/pg_ctl', 'stop', '-w', '-m', 'fast']
    local(' '.join(cmd) % env.postgres)


def shutdown_postgresql_server():
    """Shutdown the PostgreSQL server."""
    _set_postgres_environment()
    dropdb()
    stop_database()
    local("rm -rf %s" % env.postgres['HOST'])


def setup_database():
    """Setup the database."""
    success = _check_database()
    if not success:
        _createrole()
        createdb()
    syncdb()
    setup_db_access()


def setup_db_access():
    """Grant access to the database."""
    _set_postgres_environment()
    django_settings = get_django_settings('INSTALLED_APPS')
    if 'pgtools' in django_settings['INSTALLED_APPS']:
        with patch_database_settings(USER='postgres'):
            manage('grantuser', env.database['USER'])


@contextmanager
def patch_database_settings(**kwargs):
    # __enter__
    _set_postgres_environment()
    django_settings = get_django_settings('DATABASES')
    database = django_settings['DATABASES']['default']
    new_database = copy.deepcopy(database)
    new_database.update(**kwargs)

    databases = {'default': new_database, 'nop': django_settings['DATABASES']['nop']}
    os.environ['CONFIGGLUE_DJANGO_DATABASES'] = json.dumps(databases)
    yield
    # __exit__
    del os.environ['CONFIGGLUE_DJANGO_DATABASES']


# helpers
# =======

def _set_postgres_environment():
    """Update the environment with the PostgreSQL settings."""
    if 'postgres' in env:
        # environment already set up
        return

    settings = get_django_settings('DATABASES')
    env.database = settings['DATABASES']['default']

    pg_env = []
    env.postgres = {
        'BIN': "/usr/lib/postgresql/%s/bin" % PG_VERSION,
        'DATABASE': env.database['NAME'],
    }

    host = os.environ.get('PGHOST', env.database['HOST'])
    if host:
        data = "%s/data" % host
        env.postgres['HOST'] = host
        env.postgres['DATA'] = data
        pg_env.append("PGHOST=%s" % host)
        pg_env.append("PGDATA=%s" % data)

    port = env.database['PORT']
    if port:
        pg_env.append("PGPORT=%s" % port)

    env.postgres['ENV'] = ' '.join(pg_env)


def _check_database():
    """Check the database is accessible."""
    with hide('aborts', 'running'):
        try:
            _set_postgres_environment()
            local("%(ENV)s psql -U postgres -c '\q'"
                % env.postgres, capture=False)
            success = True
        except SystemExit:
            # there was an error connecting to the db,
            # presumably the db didn't exist
            success = False
    return success


def _createrole():
    """Create required users/roles."""
    _set_postgres_environment()
    local("%(ENV)s %(BIN)s/createuser --superuser --createdb postgres" %
          env.postgres)
    user = env.database['USER']
    if user != 'postgres':
        cmd = "%(ENV)s %(BIN)s/createuser -S -d -r %%(USER)s"
        local((cmd % env.postgres) % env.database)
