import json
import os
import sys
import urllib2

from urlparse import urlparse

from .django import (
    get_django_settings,
    manage,
)
from .environment import virtualenv_local


class ArgumentError(Exception):
    pass


def is_staging(url):
    if url is not None:
        url.rstrip('/')
    return url in [
        'https://login.staging.ubuntu.com',
        'https://login.staging.launchpad.net',
    ]


def is_production(url):
    if url is not None:
        url.rstrip('/')
    return url in [
        'https://login.ubuntu.com',
        'https://login.launchpad.net',
    ]


def is_devel():
    url = os.getenv('SST_BASE_URL', None)
    return url is None or not (is_staging(url) or is_production(url))


def test(*args, **kwargs):
    """Run unit tests."""
    cmd = ['python django_project/manage.py', 'test', '--noinput']
    cmd.extend(list(args))
    virtualenv_local(' '.join(cmd), capture=False)


def lint(*args, **kwargs):
    static_test_names = [
        'api.tests.test_static_code_analysis',
        'identityprovider.tests.test_static_code_analysis',
        'webui.tests.test_static_code_analysis',
    ]
    test(*static_test_names + list(args), **kwargs)


def _is_true(arg, name):
    if arg.lower() in ('t', 'true', 'on', '1', 'yes'):
        return True
    elif arg.lower() in ('f', 'false', 'off', '0', 'no'):
        return False
    raise ArgumentError(
        "Argument {!r} should be boolean, was {!r}".format(name, arg))


def acceptance(headless='true', screenshot='false', report='', quiet='true',
        failfast='false', tests='', flags=None, debug='false',
        extended='false', concurrency=1):
    """Run acceptance tests only."""

    extended = _is_true(extended, 'extended')
    headless = _is_true(headless, 'headless')
    quiet = _is_true(quiet, 'quiet')
    screenshot = _is_true(screenshot, 'screenshot')
    failfast = _is_true(failfast, 'failfast')
    directory = '-d acceptance/tests'
    debug = _is_true(debug, 'debug')

    os.environ['DJANGO_SETTINGS_MODULE'] = 'django_project.settings'
    os.environ['PYTHONPATH'] = '.:lib:src'
    cmd = ['sst-run']
    if extended:
        cmd.append('--extended-tracebacks')
    if headless:
        cmd.append('-x')
    if screenshot:
        cmd.append('-s')
    if report:
        cmd.append('-r %s' % report)
    if failfast:
        cmd.append('--failfast')
    if tests:
        cmd.append(tests)

    sst_flags = set(_get_sst_flags())
    if flags is not None:
        flags = set(flags.strip(';').split(';'))
        sst_flags = sst_flags.union(flags)
    cmd.append('--with-flags="%s"' % ','.join(sst_flags))

    if quiet:
        cmd.append('-q')
    if debug:
        cmd.append('--debug')

    cmd.append('--concurrency=%s' % concurrency)
    cmd.append(directory)
    virtualenv_local(' '.join(cmd), capture=False)


def run(*args, **kwargs):
    """Run development server."""
    if not args:
        parsed = urlparse(get_django_settings('SSO_ROOT_URL')['SSO_ROOT_URL'])
        args = [parsed.netloc]
    manage('runserver', *args, **kwargs)


def docs():
    virtualenv_local('pip install -r requirements_docs.txt')
    virtualenv_local('sphinx-build docs docs/html')


def _get_sst_flags(url=None, username=None, password=None):
    defaults = get_django_settings(
        'SSO_ROOT_URL',
        'SSO_TEST_ACCOUNT_EMAIL',
        'SSO_TEST_ACCOUNT_PASSWORD'
    )
    if url is None:
        url = os.environ.get('SST_BASE_URL', defaults['SSO_ROOT_URL'])
    if username is None:
        username = defaults['SSO_TEST_ACCOUNT_EMAIL']
    if password is None:
        password = defaults['SSO_TEST_ACCOUNT_PASSWORD']

    preflight_url = url.rstrip('/') + '/preflight/'

    # can't use the builtin basic auth as that waits for a 401 challenge before
    # sending the Authenticate header
    encoded = ('%s:%s' % (username, password)).encode('base64')
    req = urllib2.Request(preflight_url, headers={
        'Accept': 'application/json',
        'Authorization': 'Basic %s' % encoded.replace('\n', '')
    })

    try:
        response = urllib2.urlopen(req)
    except urllib2.URLError as e:
        print "Problem retrieving preflight:", e
        sys.exit(1)

    try:
        data = json.loads(response.read())
    except ValueError as e:
        print "Problem parsing preflight response", e
        sys.exit(1)

    try:
        switches = data['switches']['gargoyle']
    except KeyError:
        print 'Couldn\'t find gargoyle switches in preflight json response'
        sys.exit(1)

    output = []

    for switch in switches:
        if 'active' in switch['status_text'].lower():
            output.append(switch['name'].lower())

    return output


def sst_flags(url=None, username=None, password=None):
    """Output in a format suitable for fab acceptance."""
    print ";".join(_get_sst_flags(url, username, password))
