# Copyright 2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import os
import shutil
import sys
import tempfile
import textwrap

from fabric.api import env, local
from fabric.context_managers import lcd

from django_project.paths import LOCAL_CFG_DIR, LOCAL_CFG_PATH
from .constants import BUILD_DEPENDENCIES, PSYCOPG2_CONFLICTS, VIRTUALENV

LOCAL_CFG = textwrap.dedent("""
    [__noschema__]
    db_host = {envdir}/db

    [__main__]
    includes =
        {curdir}/django_project/config/devel.cfg
    """)


def bootstrap(download_cache_path=None):
    """Bootstrap the development environment."""
    _check_bootstrap_dependencies()
    _check_psycopg2_conflicts()

    setup_virtualenv()
    install_dependencies(download_cache_path)
    setup_configuration()


def clean():
    """Clean up compiled and backup files."""
    with lcd('django_project'):
        local("rm -rf .coverage coverage.d coverage.xml")
    local("find . -name '*.~*' -delete")
    local("find . -name '*.pyc' -delete")


def setup_virtualenv():
    """Create the virtualenv."""
    created = False
    virtual_env = os.environ.get('VIRTUAL_ENV', None)
    if virtual_env is None:
        if not os.path.exists(VIRTUALENV):
            _create_virtualenv()
            created = True
        virtual_env = VIRTUALENV
    env.virtualenv = os.path.abspath(virtual_env)
    _activate_virtualenv()
    return created


def install_dependencies(download_cache_path=None):
    """Install all dependencies into the virtualenv."""
    install_sourcedeps()
    install_pip_dependencies(download_cache_path)


def get_patched_config_manager_config():
    output = []
    with open('config-manager.txt') as config:
        for line in config.readlines():
            line = line.strip()
            if not line or line.startswith('#'):
                # ignore empty lines and comments
                continue

            path, url = line.split(' ', 1)
            url = url.replace('bazaar.isd', 'bazaar.launchpad.net')
            path = path.replace('canonical-identity-provider', '.')
            if path in ('.', './branches/translations'):
                # ignore trunk branch (as we already are in the code branch)
                # ignore translations branch (this will be removed completely
                # soon enough)
                continue
            output.append("%s %s" % (path, url))
    return '\n'.join(output)


def install_sourcedeps():
    cm_path = '/usr/lib/config-manager/cm.py'
    patched_config = get_patched_config_manager_config()
    with tempfile.NamedTemporaryFile() as dev_config:
        dev_config.write(patched_config)
        dev_config.flush()
        local("%s update %s" % (cm_path, dev_config.name))


def install_pip_dependencies(download_cache_path=None):
    if download_cache_path:
        cwd = os.getcwd()
        with lcd(download_cache_path):
            virtualenv_local(
                'make install PACKAGES="-r %s/requirements.txt"' %
                cwd, capture=False)
    else:
        virtualenv_local('pip install -r requirements.txt', capture=False)


def setup_configuration():
    """Setup the configuration."""
    if not os.path.exists(LOCAL_CFG_DIR):
        os.makedirs(LOCAL_CFG_DIR)

    LEGACY_LOCAL_CFG_PATH = 'django_project/local.cfg'
    if (os.path.exists(LEGACY_LOCAL_CFG_PATH) and
            not os.path.exists(LOCAL_CFG_PATH)):
        shutil.move(LEGACY_LOCAL_CFG_PATH, LOCAL_CFG_DIR)

    if not os.path.exists(LOCAL_CFG_PATH):
        config = LOCAL_CFG.format(
            curdir=os.path.abspath(os.curdir),
            envdir=env.virtualenv)
        with open(LOCAL_CFG_PATH, 'w') as local_cfg:
            local_cfg.write(config)


def virtualenv_local(command, capture=True):
    """Run a command inside the virtualenv."""
    prefix = ''
    virtual_env = env.get('virtualenv', None)
    if virtual_env:
        prefix = ". %s/bin/activate && " % virtual_env
    command = prefix + command
    return local(command, capture=capture)


# helpers
# =======

def _check_bootstrap_dependencies():
    """Check dependencies required for bootstrap."""
    required = []
    cmd = "dpkg -l %s 2> /dev/null | grep '^ii' | wc -l"
    for pkg in BUILD_DEPENDENCIES:
        output = local(cmd % pkg, capture=True).strip()
        if output != '1':
            required.append(pkg)
    if required:
        print "Please install the following packages, as they are required"
        print "in order to build some of the dependencies:"
        for pkg in required:
            print pkg
        sys.exit(1)


def _check_psycopg2_conflicts():
    """Check for libraries conflicting with psycopg2."""
    conflicting = []
    cmd = "dpkg -l %s 2> /dev/null | grep '^ii' | wc -l"
    for pkg in PSYCOPG2_CONFLICTS:
        output = local(cmd % pkg, capture=True).strip()
        if output != '0':
            conflicting.append(pkg)
    if conflicting:
        print "Please remove the following packages, as their presence will"
        print "produce the psycopg2 library to be built incorrectly:"
        for pkg in conflicting:
            print pkg
        sys.exit(1)


def _activate_virtualenv():
    """Activate the virtualenv."""
    activate_this = os.path.abspath(
        "%s/bin/activate_this.py" % env.virtualenv)
    execfile(activate_this, dict(__file__=activate_this))


def _create_virtualenv(clear=False):
    """Create the virtualenv."""
    if not os.path.exists(VIRTUALENV) or clear:
        virtualenv_bin_path = local('which virtualenv', capture=True)
        args = '--distribute --clear --system-site-packages'
        local("%s %s %s %s" % (sys.executable,
            virtualenv_bin_path, args, VIRTUALENV), capture=False)
