# Copyright 2010-2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import json
import logging

from piston.resource import Resource

from api.v20.utils import errors
from identityprovider.auth import SSOOAuthAuthentication


logger = logging.getLogger(__name__)


class ApiResource(Resource):

    def __call__(self, request, *args, **kwargs):
        request.args = args
        request.kwargs = kwargs
        response = super(ApiResource, self).__call__(request, *args, **kwargs)
        # cope with piston stripping out all response headers
        if response.status_code == 429:
            retry_after = json.loads(response.content)['extra'].get(
                'Retry-After', '')
            if retry_after:
                response['Retry-After'] = retry_after
        return response


class ApiOAuthAuthentication(SSOOAuthAuthentication):

    def _has_ownership(self, request):
        return True

    def is_authenticated(self, request):
        result = super(ApiOAuthAuthentication, self).is_authenticated(request)
        if result:
            # valid oauth credentials is not enough, we need to guarantee this
            # user also has ownership over the resource
            result = self._has_ownership(request)

            if not result:
                # oauth creds are valid, but user does not owns the resource,
                # check if user has special perms over the account
                if request.method == 'GET':
                    perm = 'identityprovider.api_view_account_details'
                else:
                    perm = 'identityprovider.api_edit_account_details'
                account = request.user
                result = account.user.has_perm(perm)

        return result

    def challenge(self):
        """Return a json body 401 response."""
        response = errors.INVALID_CREDENTIALS()
        response.content = json.dumps(response._container)
        response['content-type'] = 'application/json; charset=utf-8'
        # set headers
        # urgh - hardcoded realm - but piston has it hardcoded...
        for k, v in self.builder(realm='API').iteritems():
            response[k] = v
        return response


class ApiAccountsAuthentication(ApiOAuthAuthentication):

    def _has_ownership(self, request):
        openid = request.kwargs.get('openid')  # given openid
        result = request.user.openid_identifier == openid
        return result


class ApiEmailsAuthentication(ApiOAuthAuthentication):

    def _has_ownership(self, request):
        email = request.kwargs.get('email')  # given email
        emails = request.user.emailaddress_set.filter(email=email)
        result = emails.count() > 0
        return result


class ApiTokensAuthentication(ApiOAuthAuthentication):

    def _has_ownership(self, request):
        token = request.kwargs.get('token')  # given token key
        result = request.user.oauth_token_by_key(token) is not None
        return result
