# Copyright 2010-2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import logging

from django.core.exceptions import (
    MultipleObjectsReturned,
    ObjectDoesNotExist,
    ValidationError,
)
from django.conf import settings
from gargoyle import gargoyle
from oauth import oauth
from oauth_backend.models import Token
from piston.handler import AnonymousBaseHandler, BaseHandler
from piston.utils import rc, require_mime

from api.v20 import registration
from api.v20.decorators import throttle
from api.v20.utils import (
    errors,
    get_account_data,
    get_email_data,
    get_minimal_account_data,
    get_token_data,
)
from identityprovider import emailutils
from identityprovider.login import (
    AccountDeactivated,
    AccountSuspended,
    AuthenticationError,
    EmailInvalidated,
    PasswordPolicyError,
    authenticate_user,
)
from identityprovider.models import (
    Account,
    AuthToken,
    EmailAddress,
    InvalidatedEmailAddress,
    twofactor,
)
from identityprovider.models.captcha import Captcha, VerifyCaptchaError
from identityprovider.models.const import (
    AccountStatus,
    TokenType,
)
from identityprovider.store import SSODataStore
from identityprovider.timeline import get_request_timing_function
from identityprovider.utils import redirection_url_for_token


logger = logging.getLogger(__name__)


def get_openid_from_args(request, openid):
    return openid


def get_email_from_request_data(request):
    return request.data.get('email')


def get_email_from_args(request, email):
    return email


def get_token_from_args(request, token):
    return token


class AnonymousAccountsHandler(AnonymousBaseHandler):
    allowed_methods = ('GET',)

    # limit to 1 request per second
    # cache expires after an hour
    @throttle(max_requests=3600, timeout=3600,
              extra_callback=get_openid_from_args)
    def read(self, request, openid):
        account = Account.objects.active_by_openid(openid)
        if account is None:
            return errors.RESOURCE_NOT_FOUND()

        return get_minimal_account_data(account)


class AccountsHandler(BaseHandler):
    allowed_methods = ('GET',)
    anonymous = AnonymousAccountsHandler

    # limit to 1 request per second
    # cache expires after an hour
    @throttle(max_requests=3600, timeout=3600,
              extra_callback=get_openid_from_args)
    def read(self, request, openid):
        """Return account information."""
        account = Account.objects.active_by_openid(openid)
        if account is None:
            return errors.RESOURCE_NOT_FOUND()

        expand = (request.GET.get('expand', '').lower() == 'true')
        return get_account_data(account, expand=expand)


class PasswordResetTokenHandler(BaseHandler):
    allowed_methods = ('POST',)

    # limit to 1 request per second
    # cache expires after an hour
    @throttle(max_requests=3600, timeout=3600,
              extra_callback=get_email_from_request_data)
    @require_mime('json')
    def create(self, request):
        data = request.data
        email = data.get('email')
        if email is None:
            return errors.INVALID_DATA(email='Field required')

        token_string = data.get('token')

        account = Account.objects.get_by_email(email)
        invalidated_email = InvalidatedEmailAddress.objects.filter(
            email__iexact=email)

        if account is None:
            # they've tried to reset with an invalid email, so send them
            # an email on how to create an account
            emailutils.send_invitation_after_password_reset(email)

            if invalidated_email.exists():
                # there is no account with this email
                # but there was some with it invalidated
                return errors.EMAIL_INVALIDATED()
            else:
                # there is no account, neither with this email invalidated
                return errors.INVALID_DATA(
                    email="No account associated with %s" % email)

        if not account.can_reset_password:
            if account.status == AccountStatus.SUSPENDED:
                # log why email was not sent
                condition = ("account '%s' is not active" %
                             account.displayname)
                logger.debug("PasswordResetTokenHandler.create: email was "
                             "not sent out because %s" % condition)
                return errors.ACCOUNT_SUSPENDED()
            elif account.status == AccountStatus.DEACTIVATED:
                # log why email was not sent
                condition = ("account '%s' is not active" %
                             account.displayname)
                logger.debug("PasswordResetTokenHandler.create: email was "
                             "not sent out because %s" % condition)
                return errors.ACCOUNT_DEACTIVATED()

            # user does not have any verified email address
            # he should contact support
            condition = ("account '%s' is not allowed to reset password" %
                         account.displayname)
            logger.debug("PasswordResetTokenHandler.create: email was not "
                         "sent out because %s" % condition)
            return errors.CAN_NOT_RESET_PASSWORD()

        validated = account.emailaddress_set.verified().filter(
            email__iexact=email).exists()
        if not validated and account.preferredemail is not None:
            email = account.preferredemail.email

        tokens = AuthToken.objects.filter(
            token_type=TokenType.PASSWORDRECOVERY, requester_email=email,
            email=email, date_consumed=None)
        if tokens.count() >= settings.MAX_PASSWORD_RESET_TOKENS:
            return errors.TOO_MANY_TOKENS()

        # create new token and send email to user
        redirection_url = redirection_url_for_token(token_string)
        auth_token = emailutils.send_password_reset_email(
            account, email, redirection_url)

        # return response
        data = dict(email=auth_token.email)
        response = rc.CREATED
        response.content = data
        return response


class AccountRegistrationHandler(BaseHandler):
    allowed_methods = ('POST',)

    # limit to 1 request per second
    # cache expires after an hour
    @throttle(max_requests=3600, timeout=3600,
              extra_callback=get_email_from_request_data)
    @require_mime('json')
    def create(self, request):
        """Create/register a new account."""
        data = request.data

        try:
            email = data['email']
            password = data['password']
            displayname = data['displayname']
        except KeyError:
            expected = set(('email', 'password', 'displayname'))
            missing = dict((k, 'Field required') for k in expected - set(data))
            return errors.INVALID_DATA(**missing)

        captcha_id = data.get('captcha_id')
        captcha_solution = data.get('captcha_solution')
        if (not (captcha_solution and captcha_id) and
                gargoyle.is_active('CAPTCHA')):
            extra = {}
            if data.get('create_captcha', True):
                timer = get_request_timing_function(request)
                try:
                    extra = Captcha.new(timer=timer).serialize()
                except Exception:
                    logger.exception('failed to create reCaptcha')
            return errors.CAPTCHA_REQUIRED(**extra)

        elif captcha_id:
            remote_addr = request.environ['REMOTE_ADDR']
            captcha = Captcha(captcha_id)
            verified = False
            try:
                timer = get_request_timing_function(request)
                verified = captcha.verify(
                    captcha_solution, remote_addr, data['email'],
                    timer=timer)
            except VerifyCaptchaError as e:
                # only expose HTTP error codes to client
                code = e.response.code if e.response.code > 200 else None
                return errors.CAPTCHA_ERROR(
                    recaptcha_reason=e.response.reason,
                    recaptcha_status_code=code,
                    recaptcha_body=e.response.body)
            except Exception:
                logger.exception("reCaptcha error")

            if not verified:
                message = getattr(captcha, 'message', '')
                return errors.CAPTCHA_FAILURE(captcha_message=message)

        try:
            account = registration.register(
                email, password, displayname,
            )
        except ValidationError as e:
            return errors.INVALID_DATA(**e.message_dict)
        except registration.EmailAlreadyRegistered:
            return errors.ALREADY_REGISTERED(email=data['email'])

        data = get_account_data(account)
        response = rc.CREATED
        response.content = data
        return response


class AccountLoginHandler(BaseHandler):
    allowed_methods = ('POST',)

    # limit to 1 request per second
    # cache expires after an hour
    @throttle(max_requests=3600, timeout=3600,
              extra_callback=get_email_from_request_data)
    @require_mime('json')
    def create(self, request):
        data = request.data
        try:
            email = data['email']
            password = data['password']
            token_name = data['token_name']
        except KeyError:
            expected = set(('email', 'password', 'token_name'))
            missing = dict((k, 'Field required') for k in expected - set(data))
            return errors.INVALID_DATA(**missing)

        try:
            account = authenticate_user(email, password)
        except AccountSuspended:
            return errors.ACCOUNT_SUSPENDED()
        except AccountDeactivated:
            return errors.ACCOUNT_DEACTIVATED()
        except EmailInvalidated:
            return errors.EMAIL_INVALIDATED()
        except PasswordPolicyError as e:
            return errors.PASSWORD_POLICY_ERROR(
                reason=unicode(e), location=settings.SSO_ROOT_URL)
        except AuthenticationError:
            return errors.INVALID_CREDENTIALS()

        otp = data.get('otp')
        if otp is not None:
            try:
                twofactor.authenticate(account, otp)
            except AuthenticationError:
                return errors.TWOFACTOR_FAILURE()
        elif account.twofactor_required:
            return errors.TWOFACTOR_REQUIRED()

        token, created = account.get_or_create_oauth_token(token_name)
        if created:
            response = rc.CREATED
        else:
            response = rc.ALL_OK

        data = get_token_data(token)
        data['openid'] = account.openid_identifier
        response.content = data

        return response


class AccountPhoneLoginHandler(BaseHandler):
    allowed_methods = ('POST', )

    # limit to 1 request per second
    # cache expires after an hour
    @throttle(max_requests=3600, timeout=3600,
              extra_callback=get_email_from_request_data)
    @require_mime('json')
    def create(self, request):
        data = request.data
        # email should be available the first time, to bind account/phone
        # after that, get email from phoneid in the token name
        email = data.get('email')
        try:
            phone_id = data['phone_id']
            password = data['password']
            token_name = data['token_name']
        except KeyError:
            expected = set(('phone_id', 'password', 'token_name'))
            missing = dict((k, 'Field required') for k in expected - set(data))
            return errors.INVALID_DATA(**missing)

        try:
            phone_email = EmailAddress.objects.get_from_phone_id(phone_id)
            login_email = phone_email.email
        except EmailAddress.DoesNotExist:
            phone_email = None
            login_email = email

        if login_email is None:
            return errors.INVALID_DATA(phone_id='invalid value')

        try:
            account = authenticate_user(login_email, password)
        except AccountSuspended:
            return errors.ACCOUNT_SUSPENDED()
        except AccountDeactivated:
            return errors.ACCOUNT_DEACTIVATED()
        except PasswordPolicyError as e:
            return errors.PASSWORD_POLICY_ERROR(
                reason=unicode(e), location=settings.SSO_ROOT_URL)
        except AuthenticationError:
            return errors.INVALID_CREDENTIALS()

        # TODO: here we should check for 2fa, avoiding for demo

        if phone_email is None:
            # user is authenticated through email,
            # adding phone id as email for future login
            EmailAddress.objects.create_from_phone_id(phone_id, account)

        token, created = account.get_or_create_oauth_token(token_name)
        if created:
            response = rc.CREATED
        else:
            response = rc.ALL_OK

        data = get_token_data(token)
        data['openid'] = account.openid_identifier
        response.content = data

        return response


class EmailsHandler(BaseHandler):
    allowed_methods = ('DELETE', 'GET')
    model = EmailAddress

    # limit to 1 request per second
    # cache expires after an hour
    @throttle(max_requests=3600, timeout=3600,
              extra_callback=get_email_from_args)
    def read(self, request, email):
        try:
            result = get_email_data(self.model.objects.get(email=email))
        except ObjectDoesNotExist:
            result = rc.NOT_FOUND
        except MultipleObjectsReturned:
            result = rc.BAD_REQUEST
        return result

    # limit to 1 request per second
    # cache expires after an hour
    @throttle(max_requests=3600, timeout=3600,
              extra_callback=get_email_from_args)
    def delete(self, request, email):
        result = super(EmailsHandler, self).delete(request, email=email)
        # piston is returning 410 (Gone) when the email does not exist
        # this is not correct, we should return 404
        if result.status_code == rc.NOT_HERE.status_code:
            result = errors.RESOURCE_NOT_FOUND()
        return result


class RequestsHandler(BaseHandler):
    allowed_methods = ('POST',)

    def serialize_error(self, exception):
        result = exception.__class__.__name__
        msg = ', '.join(repr(i) for i in exception.args)
        if msg:
            result += ': %s' % msg
        return result

    @require_mime('json')
    def create(self, request):
        """Return whether a given oauth authorization header is valid.

        Receive an oauth Authorization header, an url and an http method and
        decide if the signature is valid (was created using valid credentials
        and belongs to an activa account).

        """
        data = request.data
        http_method = data.get('http_method', '')
        http_url = data.get('http_url', '')
        authorization = data.get('authorization', None)
        query_string = data.get('query_string', None)

        headers = None
        if authorization is not None:
            headers = {'Authorization': authorization}

        result = {'is_valid': False}

        if not http_url or not http_method:
            return result

        try:
            http_url = http_url.encode('utf-8')
            oauth_request = oauth.OAuthRequest.from_request(
                http_method, http_url, headers=headers,
                query_string=query_string)
        except Exception as e:
            result['error'] = self.serialize_error(e)
            logger.exception(
                'RequestsHandler.create: could not build OAuthRequest with '
                'the given parameters %r:', data)
            oauth_request = None

        if oauth_request is None:
            return result

        oauth_datastore = SSODataStore(oauth_request)
        oauth_server = oauth.OAuthServer(oauth_datastore)
        oauth_server.add_signature_method(
            oauth.OAuthSignatureMethod_PLAINTEXT())
        oauth_server.add_signature_method(
            oauth.OAuthSignatureMethod_HMAC_SHA1())

        try:
            consumer, _, _ = oauth_server.verify_request(oauth_request)
        except oauth.OAuthError:
            # do nothing special, signature was invalid
            valid = False
        except Exception as e:
            result['error'] = self.serialize_error(e)
            logger.exception(
                'RequestsHandler.create: could not verify request:')
            valid = False
        else:
            # signature is valid, need to also check the account exists
            # and is active
            valid = Account.objects.active_by_openid(consumer.key) is not None

        result['is_valid'] = valid
        return result


class TokensHandler(BaseHandler):
    allowed_methods = ('DELETE', 'GET')
    model = Token

    # limit to 1 request per second
    # cache expires after an hour
    @throttle(max_requests=3600, timeout=3600,
              extra_callback=get_token_from_args)
    def read(self, request, token):
        """Return token information."""
        result = super(TokensHandler, self).read(request, token=token)
        if isinstance(result, self.model):
            result = get_token_data(result)
            result.pop('token_secret', None)
            result.pop('consumer_secret', None)

        return result

    # limit to 1 request per second
    # cache expires after an hour
    @throttle(max_requests=3600, timeout=3600,
              extra_callback=get_token_from_args)
    def delete(self, request, token):
        """Return token information."""
        result = super(TokensHandler, self).delete(request, token=token)
        # piston is returning 410 (Gone) when the token does not exist
        # this is not correct, we should return 404
        if result.status_code == rc.NOT_HERE.status_code:
            result = errors.RESOURCE_NOT_FOUND()
        return result
