# -*- coding: utf-8 -*-
import json
import time
from urllib import urlencode

from django.core import mail
from django.core.serializers.json import DjangoJSONEncoder
from django.core.urlresolvers import NoReverseMatch, reverse
from django.contrib.auth.models import AnonymousUser
from django.test.client import RequestFactory
from mock import ANY, patch
from oauth_backend.models import Token
from oauth.oauth import OAuthRequest

from identityprovider.models import (
    Account,
    AuthToken,
    EmailAddress,
)
from identityprovider.models.const import (
    AccountStatus,
    EmailStatus,
    TokenType,
)
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import (
    SSOBaseTestCase,
    authorization_header_from_token,
    patch_settings,
)
from identityprovider.utils import redirection_url_for_token

from api.v20 import handlers


class POSTOnlyResourceMixin(object):

    def do_post(self, data=None):
        if data is None:
            data = getattr(self, 'data', {})
        json_data = json.dumps(data)
        return self.client.post(self.url, data=json_data,
                                content_type='application/json')

    def test_get_not_allowed(self):
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 405)
        self.assertEqual(response.get('Allow', ''), 'POST')

    def test_requires_post_data_json_encoded(self):
        response = self.client.post(self.url, data={'foo': ['bar', 'baz']})
        # django-piston doesn't set this required attribute
        response._base_content_is_iter = False
        self.assertContains(response, 'Bad Request', status_code=400)


class BaseTestCase(SSOBaseTestCase):

    url = None

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self.account = self.factory.make_account(
            password=DEFAULT_USER_PASSWORD)
        self.token = self.account.create_oauth_token('token-name')

    def get_response(self, url=None, token=None, method='GET', **kwargs):
        if url is None:
            url = self.url

        if token is not None:
            headers = authorization_header_from_token(
                url, token, http_method=method.upper())
            kwargs.update(headers)

        return getattr(self.client, method.lower())(url, **kwargs)

    def to_json(self, expected):
        return json.loads(json.dumps(expected, cls=DjangoJSONEncoder))


class AccountsHandlerTestCase(BaseTestCase):
    url_name = 'api-account'

    def setUp(self):
        super(AccountsHandlerTestCase, self).setUp()
        self.url = reverse(self.url_name,
                           args=(self.account.openid_identifier,))

    def test_openid_empty(self):
        self.assertRaises(NoReverseMatch, reverse, self.url_name, args=('',))

    def test_unauthenticated_gives_minimal(self):
        response = self.get_response()

        self.assertEqual(response.status_code, 200)
        expected = handlers.get_minimal_account_data(self.account)
        self.assertEqual(json.loads(response.content), self.to_json(expected))

    def test_authenticated_with_django_gives_minimal(self):
        assert self.client.login(username=self.account.preferredemail.email,
                                 password=DEFAULT_USER_PASSWORD)
        response = self.get_response()

        self.assertEqual(response.status_code, 200)
        expected = handlers.get_minimal_account_data(self.account)
        self.assertEqual(json.loads(response.content), self.to_json(expected))

    def test_authenticated_with_oauth(self):
        response = self.get_response(token=self.token)

        self.assertEqual(response.status_code, 200)
        expected = handlers.get_account_data(self.account)
        self.assertEqual(json.loads(response.content), self.to_json(expected))

    def test_authenticated_openid_no_match(self):
        assert self.account.openid_identifier != 'foo'
        url = reverse(self.url_name, args=('foo',))
        response = self.get_response(url=url, token=self.token)

        self.assertEqual(response.status_code, 404)

    def test_authenticated_openid_no_match_but_has_permission(self):
        self.factory.add_permission_to_account(
            self.account, 'api_view_account_details')

        other = self.factory.make_account()
        assert self.account.openid_identifier != other.openid_identifier
        url = reverse(self.url_name, args=(other.openid_identifier,))
        response = self.get_response(url=url, token=self.token)

        self.assertEqual(response.status_code, 200)
        expected = handlers.get_account_data(other)
        self.assertEqual(json.loads(response.content), self.to_json(expected))

    def test_authenticated_openid_no_match_but_has_permission_404(self):
        self.factory.add_permission_to_account(
            self.account, 'api_view_account_details')

        assert self.account.openid_identifier != 'foo'
        assert Account.objects.active_by_openid('foo') is None
        url = reverse(self.url_name, args=('foo',))
        response = self.get_response(url=url, token=self.token)

        self.assertEqual(response.status_code, 404)

    def test_get_account_data_no_preferred_email(self):
        account = self.account
        account.emailaddress_set.all().delete()

        assert account.preferredemail is None
        result = handlers.get_account_data(account)

        self.assertIsNone(result['email'])

    def test_get_account_data_verified(self):
        account = self.account
        assert account.emailaddress_set.verified().count() > 0

        result = handlers.get_account_data(account)
        self.assertTrue(result['verified'])

    def test_get_account_data_no_verified(self):
        account = self.account
        account.emailaddress_set.verified().delete()

        result = handlers.get_account_data(account)
        self.assertFalse(result['verified'])

    def assert_account_data_expanded(self, expand_key, expand):
        url = self.url + '?expand=%s' % expand_key
        response = self.get_response(url=url, token=self.token)

        self.assertEqual(response.status_code, 200)
        expected = handlers.get_account_data(self.account, expand=expand)
        self.assertEqual(json.loads(response.content), self.to_json(expected))

    def test_expanded_data(self):
        self.assert_account_data_expanded('true', True)

    def test_expanded_data_case_insensitive(self):
        self.assert_account_data_expanded('True', True)
        self.assert_account_data_expanded('TRUE', True)
        self.assert_account_data_expanded('trUE', True)

    def test_not_expanded_data(self):
        self.assert_account_data_expanded('false', False)

    def test_not_expanded_data_other_values(self):
        self.assert_account_data_expanded('False', False)
        self.assert_account_data_expanded('0', False)
        self.assert_account_data_expanded('no', False)
        self.assert_account_data_expanded('', False)


class EmailsHandlerTestCase(BaseTestCase):
    url_name = 'api-email'

    def setUp(self):
        super(EmailsHandlerTestCase, self).setUp()
        self.email = self.account.preferredemail
        self.url = reverse(
            self.url_name, kwargs=dict(email=self.email.email))

    def assert_email_result(self, result, email):
        expected = dict(
            email=email.email, verified=True,
            date_created=email.date_created,
            href=reverse(self.url_name, kwargs=dict(email=email.email)),
        )
        self.assertEqual(result, self.to_json(expected))

    def test_email_empty(self):
        self.assertRaises(
            NoReverseMatch, reverse, self.url_name, kwargs=dict(email=''))

    def test_unauthenticated_401(self):
        response = self.get_response()

        self.assertEqual(response.status_code, 401)

    def test_authenticated_with_django_401(self):
        assert self.client.login(username=self.account.preferredemail.email,
                                 password=DEFAULT_USER_PASSWORD)
        response = self.get_response()

        self.assertEqual(response.status_code, 401)

    def test_get_authenticated_with_oauth(self):
        response = self.get_response(token=self.token)
        self.assertEqual(response.status_code, 200)

        result = json.loads(response.content)
        self.assert_email_result(result, self.email)

    def test_can_get_others_if_elevated_permission(self):
        other_email = self.factory.make_account().preferredemail
        self.factory.add_permission_to_account(
            self.account, 'api_view_account_details')

        url = reverse(self.url_name, kwargs=dict(email=other_email.email))
        response = self.get_response(url=url, token=self.token)
        self.assertEqual(response.status_code, 200)

        result = json.loads(response.content)
        self.assert_email_result(result, other_email)

    def test_forbidden_if_not_owner_not_elevated_permission_on_get(self):
        other_email = self.factory.make_account().preferredemail

        url = reverse(self.url_name, kwargs=dict(email=other_email.email))
        response = self.get_response(url=url, token=self.token)

        self.assertEqual(response.status_code, 401)

    def test_get_email_does_not_exist(self):
        self.factory.add_permission_to_account(
            self.account, 'api_view_account_details')

        url = reverse(self.url_name, kwargs=dict(email='foobarbaz'))
        response = self.get_response(url=url, token=self.token)

        self.assertEqual(response.status_code, 404)

    def test_delete_authenticated_with_oauth(self):
        response = self.get_response(token=self.token, method='DELETE')

        self.assertEqual(response.status_code, 204)
        self.assertEqual(response.content, '')
        self.assertEqual(
            EmailAddress.objects.filter(email=self.email.email).count(), 0)

    def test_can_delete_others_if_elevated_permission(self):
        other_email = self.factory.make_account().preferredemail
        self.factory.add_permission_to_account(
            self.account, 'api_edit_account_details')

        url = reverse(self.url_name, kwargs=dict(email=other_email.email))
        response = self.get_response(
            url=url, token=self.token, method='DELETE')

        self.assertEqual(response.status_code, 204)
        self.assertEqual(response.content, '')
        self.assertEqual(
            EmailAddress.objects.filter(email=other_email.email).count(), 0)

    def test_forbidden_if_not_owner_not_elevated_permission_on_delete(self):
        other_email = self.factory.make_account().preferredemail

        url = reverse(self.url_name, kwargs=dict(email=other_email.email))
        response = self.get_response(
            url=url, token=self.token, method='DELETE')

        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            EmailAddress.objects.filter(email=other_email.email).count(), 1)

    def test_delete_email_does_not_exist(self):
        self.factory.add_permission_to_account(
            self.account, 'api_edit_account_details')

        url = reverse(self.url_name, kwargs=dict(email='foobarbaz'))
        response = self.get_response(
            url=url, token=self.token, method='DELETE')

        self.assertEqual(response.status_code, 404)


class RequestsTestCase(BaseTestCase, POSTOnlyResourceMixin):
    url_name = 'api-requests'
    default_http_method = 'GET'

    def setUp(self):
        super(RequestsTestCase, self).setUp()
        self.url = reverse(self.url_name)

        self.external_url = 'http://example.com/foo/bar?some=1&nothing=foo'
        self.data = self.build_data()

        self.mock_logger = self._apply_patch('api.v20.handlers.logger')

    def build_data(self, external_url=None, token=None):
        if external_url is None:
            external_url = self.external_url
        if token is None:
            token = self.token
        header = authorization_header_from_token(
            external_url, token, http_method=self.default_http_method)
        data = dict(
            http_url=external_url,
            http_method=self.default_http_method,
            authorization=header['HTTP_AUTHORIZATION'],
        )
        return data

    def assert_request_response(self, response, expected, error=None):
        self.assertEqual(response.status_code, 200)
        try:
            response = json.loads(response.content)
        except:
            self.fail('Could not JSON-decode response, content is %r.' %
                      response.content)
        else:
            expected = dict(is_valid=expected)
            if error:
                expected['error'] = error
            self.assertEqual(response, expected)

    def test_invalid_if_not_method_given(self):
        self.data.pop('http_method')
        response = self.do_post()
        self.assert_request_response(response, False)
        self.assertFalse(self.mock_logger.exception.called)

    def test_invalid_if_method_none(self):
        self.data['http_method'] = None
        response = self.do_post()
        self.assert_request_response(response, False)
        self.assertFalse(self.mock_logger.exception.called)

    def test_invalid_if_not_url_given(self):
        self.data.pop('http_url')
        response = self.do_post()
        self.assert_request_response(response, False)
        self.assertFalse(self.mock_logger.exception.called)

    def test_invalid_if_url_none(self):
        self.data['http_url'] = None
        response = self.do_post()
        self.assert_request_response(response, False)
        self.assertFalse(self.mock_logger.exception.called)

    def test_url_encoded(self):
        self.data = self.build_data(external_url=u'http://foo/ñoño ñandú/')
        response = self.do_post()

        self.assert_request_response(response, True)
        self.assertFalse(self.mock_logger.exception.called)

    def test_error_in_response_is_robust_to_encodings(self):
        msg = u'Ñoño ñandú'
        exc = TypeError(msg, msg.encode('utf-8'), msg.encode('latin-1'))
        self._apply_patch(
            'api.v20.handlers.oauth.OAuthRequest.from_request',
            side_effect=exc)

        response = self.do_post()

        msg = ', '.join(repr(i) for i in exc.args)
        self.assert_request_response(
            response, False, error="TypeError: %s" % msg)

    def test_invalid_if_from_request_fails(self):
        self._apply_patch(
            'api.v20.handlers.oauth.OAuthRequest.from_request',
            side_effect=ValueError(u'foo ñoño'.encode('latin-1')))

        response = self.do_post()
        self.assert_request_response(
            response, False, error="ValueError: 'foo \\xf1o\\xf1o'")
        self.mock_logger.exception.assert_called_once_with(
            'RequestsHandler.create: could not build OAuthRequest with the '
            'given parameters %r:', self.data)

    def test_invalid_if_from_verify_request_fails(self):
        self._apply_patch(
            'api.v20.handlers.oauth.OAuthServer.verify_request',
            side_effect=KeyError(u'\xf1'))

        response = self.do_post()
        self.assert_request_response(
            response, False, error="KeyError: u'\\xf1'")
        self.mock_logger.exception.assert_called_once_with(
            'RequestsHandler.create: could not verify request:')

    def test_invalid_if_header_none(self):
        self.data['authorization'] = None
        response = self.do_post()
        self.assert_request_response(response, False)
        self.assertFalse(self.mock_logger.exception.called)

    def test_invalid_if_not_header_given(self):
        self.data.pop('authorization')
        response = self.do_post()
        self.assert_request_response(response, False)
        self.assertFalse(self.mock_logger.exception.called)

    def test_invalid_if_no_header_or_query_string(self):
        del self.data['authorization']
        response = self.do_post()
        self.assert_request_response(response, False)
        self.assertFalse(self.mock_logger.exception.called)

    def test_invalid_if_bad_query_string(self):
        self.data['query_string'] = 'invalid'
        # make sure there is no authorization header
        del self.data['authorization']
        response = self.do_post()
        self.assert_request_response(response, False)
        self.assertFalse(self.mock_logger.exception.called)

    def test_invalid_if_query_string_invalid(self):
        # set a query string
        params = OAuthRequest._split_header(self.data['authorization'])
        # tamper with query string parameters
        params['oauth_nonce'] = '1'
        query_string = urlencode(params)
        self.data['query_string'] = query_string
        # make sure we don't have an authorization header set
        del self.data['authorization']

        # do a post with correct params
        response = self.do_post()
        self.assert_request_response(response, False)
        self.assertFalse(self.mock_logger.exception.called)

    def test_url_mismatch(self):
        new_url = self.external_url[:-1]
        assert new_url != self.external_url
        self.data['http_url'] = new_url
        response = self.do_post()

        self.assert_request_response(response, False)
        self.assertFalse(self.mock_logger.exception.called)

    def test_method_mismatch(self):
        self.data['http_method'] = self.default_http_method * 2
        response = self.do_post()

        self.assert_request_response(response, False)
        self.assertFalse(self.mock_logger.exception.called)

    def test_valid_request(self):
        # do a post with correct params
        response = self.do_post()

        self.assertEqual(response.status_code, 200)
        self.assert_request_response(response, True)
        self.assertFalse(self.mock_logger.exception.called)

    def test_valid_request_for_non_existent_account(self):
        self.account.delete()
        # do a post with correct params for a non existent account
        response = self.do_post()

        self.assert_request_response(response, False)

    def test_valid_request_for_inactive_account(self):
        for status, name in AccountStatus._get_choices():
            if status == AccountStatus.ACTIVE:
                continue
            self.account.status = status
            self.account.save()

            # restore oauth token since some account changes invalidate those
            token = self.account.create_oauth_token('another')
            self.data = self.build_data(token=token)

            # do a post with correct params for an inactive account
            response = self.do_post()

            self.assert_request_response(response, False)
            self.assertFalse(self.mock_logger.exception.called)

    def test_valid_request_with_only_query_string(self):
        params = OAuthRequest._split_header(self.data['authorization'])
        query_string = urlencode(params)
        self.data['query_string'] = query_string
        del self.data['authorization']
        # do a post with correct params
        response = self.do_post()

        self.assertEqual(response.status_code, 200)
        self.assert_request_response(response, True)
        self.assertFalse(self.mock_logger.exception.called)

    def test_query_string_ignored_if_authorization_header_found(self):
        self.data['query_string'] = 'invalid'
        assert self.data['authorization']
        # do a post with correct params
        response = self.do_post()

        self.assertEqual(response.status_code, 200)
        self.assert_request_response(response, True)
        self.assertFalse(self.mock_logger.exception.called)


class RequestsWithHEADTestCase(RequestsTestCase):
    default_http_method = 'HEAD'


class RequestsWithDELETETestCase(RequestsTestCase):
    default_http_method = 'DELETE'


class RequestsWithPATCHTestCase(RequestsTestCase):
    default_http_method = 'PATCH'


class RequestsWithPOSTTestCase(RequestsTestCase):
    default_http_method = 'POST'


class RequestsWithPUTTestCase(RequestsTestCase):
    default_http_method = 'PUT'


class PasswordResetTokenHandlerTestCase(SSOBaseTestCase,
                                        POSTOnlyResourceMixin):
    url_name = 'api-password-reset'

    def setUp(self):
        super(PasswordResetTokenHandlerTestCase, self).setUp()

        self.account = self.factory.make_account()
        self.email = EmailAddress.objects.get(account=self.account)
        self.token = self.factory.make_authtoken(
            token_type=TokenType.PASSWORDRECOVERY, email=self.email.email)
        self.url = reverse(self.url_name)
        self.mock_logger = self._apply_patch('api.v20.handlers.logging')

        # cleanup tokens from commands above
        AuthToken.objects.all().delete()

    def test_email_missing(self):
        response = self.do_post()
        self.assertEqual(response.status_code, 400)
        content = json.loads(response.content)
        self.assertEqual(content, {
            'message': 'Invalid request data',
            'code': 'INVALID_DATA',
            'extra': {'email': 'Field required'}})

    def test_invalid_email(self):
        name = ('api.v20.handlers.emailutils.'
                'send_invitation_after_password_reset')
        with patch(name) as mock_send_invitation:
            response = self.do_post({'email': 'invalid@foo.com'})

        mock_send_invitation.assert_called_once_with('invalid@foo.com')
        self.assertEqual(response.status_code, 400)
        content = json.loads(response.content)
        self.assertEqual(content, {
            'message': 'Invalid request data',
            'code': 'INVALID_DATA',
            'extra': {
                'email': 'No account associated with invalid@foo.com'}})

    def test_suspended_account(self):
        self.account.suspend()

        assert not self.account.can_reset_password
        response = self.do_post({'email': self.email.email})

        self.assertEqual(response.status_code, 403)
        content = json.loads(response.content)
        self.assertEqual(content, {
            'message': 'Your account has been suspended. Please contact '
                       'login support to re-enable it',
            'code': 'ACCOUNT_SUSPENDED',
            'extra': {}})

        condition = "account '%s' is not active" % self.account.displayname
        self.mock_logger.debug("PasswordResetTokenHandler.create: email was "
                               "not sent out because %s" % condition)

    def test_deactivated_account(self):
        self.account.status = AccountStatus.DEACTIVATED
        self.account.save()

        name = 'identityprovider.models.account.Account.can_reset_password'
        with patch(name, False):
            response = self.do_post({'email': self.email.email})

        self.assertEqual(response.status_code, 403)
        content = json.loads(response.content)
        self.assertEqual(content, {
            'message': 'Your account has been deactivated. To reactivate it, '
                       'please reset your password',
            'code': 'ACCOUNT_DEACTIVATED',
            'extra': {}})

        condition = "account '%s' is not active" % self.account.displayname
        self.mock_logger.debug("PasswordResetTokenHandler.create: email was "
                               "not sent out because %s" % condition)

    def test_can_not_reset_password(self):
        name = 'identityprovider.models.account.Account.can_reset_password'
        with patch(name, False):
            response = self.do_post({'email': self.email.email})

        self.assertEqual(response.status_code, 403)
        content = json.loads(response.content)
        self.assertEqual(content, {
            'message': 'Can not reset password. Please contact login support',
            'code': 'CAN_NOT_RESET_PASSWORD',
            'extra': {}})

        condition = "account '%s' is not active" % self.account.displayname
        self.mock_logger.debug("PasswordResetTokenHandler.create: email was "
                               "not sent out because %s" % condition)

    def test_reset_password(self):
        email = self.email.email
        response = self.do_post({'email': email})

        self.assertEqual(response.status_code, 201)
        body = json.loads(response.content)
        self.assertEqual(body['email'], email)

        tokens = AuthToken.objects.filter(
            email=self.account.preferredemail,
            token_type=TokenType.PASSWORDRECOVERY)
        token = tokens.order_by('-date_created')[0]

        self.assertEqual(len(mail.outbox), 1)
        mail_content = unicode(mail.outbox[0].message())
        self.assertIn(token.token, mail_content)
        self.assertIn(email, mail_content)

    def test_reset_password_uses_preferred_if_given_not_validated(self):
        email = self.factory.make_email_for_account(
            self.account, status=EmailStatus.NEW)
        response = self.do_post({'email': email.email})

        self.assertEqual(response.status_code, 201)

        tokens = AuthToken.objects.filter(
            token_type=TokenType.PASSWORDRECOVERY,
            email=email)
        self.assertEqual(tokens.count(), 0)

        tokens = AuthToken.objects.filter(
            token_type=TokenType.PASSWORDRECOVERY,
            email=self.account.preferredemail)
        self.assertEqual(tokens.count(), 1)

        self.assertEqual(len(mail.outbox), 1)
        mail_content = unicode(mail.outbox[0].message())
        self.assertNotIn(email.email, mail_content)
        self.assertIn(self.account.preferredemail.email, mail_content)

    def test_reset_password_uses_given_email_if_validated(self):
        self.factory.make_email_for_account(
            self.account, status=EmailStatus.PREFERRED)
        assert self.account.preferredemail != self.email
        response = self.do_post({'email': self.email.email})

        self.assertEqual(response.status_code, 201)

        tokens = AuthToken.objects.filter(
            token_type=TokenType.PASSWORDRECOVERY,
            email=self.account.preferredemail)
        self.assertEqual(tokens.count(), 0)

        tokens = AuthToken.objects.filter(
            token_type=TokenType.PASSWORDRECOVERY,
            email=self.email)
        self.assertEqual(tokens.count(), 1)

        self.assertEqual(len(mail.outbox), 1)
        mail_content = unicode(mail.outbox[0].message())
        self.assertNotIn(self.account.preferredemail.email, mail_content)
        self.assertIn(self.email.email, mail_content)

    def test_too_many_tokens(self):
        with patch_settings(MAX_PASSWORD_RESET_TOKENS=0):
            response = self.do_post({'email': self.email.email})

        self.assertEqual(response.status_code, 403)
        content = json.loads(response.content)
        self.assertEqual(content, {
            'message': 'Too many non-consumed tokens exist. Further token '
                       'creation is not allowed until existing tokens '
                       'are consumed.',
            'code': 'TOO_MANY_TOKENS',
            'extra': {}})

    def test_too_many_active_tokens(self):
        self.token.consume()
        tokens = AuthToken.objects.filter(
            token_type=TokenType.PASSWORDRECOVERY, email=self.email)
        assert tokens.count() == 1
        assert tokens.filter(date_consumed=None).count() == 0

        with patch_settings(MAX_PASSWORD_RESET_TOKENS=1):
            response = self.do_post({'email': self.email.email})
            self.assertEqual(response.status_code, 201)

            response = self.do_post({'email': self.email.email})
            self.assertEqual(response.status_code, 403)
            content = json.loads(response.content)
            self.assertEqual(content, {
                'message': 'Too many non-consumed tokens exist. Further '
                           'token creation is not allowed until existing '
                           'tokens are consumed.',
                'code': 'TOO_MANY_TOKENS',
                'extra': {}})

    def test_invalidated_email(self):
        invalidated_email = self.email.invalidate()

        name = ('api.v20.handlers.emailutils.'
                'send_invitation_after_password_reset')
        with patch(name) as mock_send_invitation:
            response = self.do_post({'email': invalidated_email.email})

        mock_send_invitation.assert_called_once_with(self.email.email)
        self.assertEqual(response.status_code, 403)
        content = json.loads(response.content)
        self.assertEqual(content, {
            'message': ANY,
            'code': 'EMAIL_INVALIDATED',
            'extra': {}})

    @patch('api.v20.handlers.emailutils.send_password_reset_email')
    def test_redirect_from_token(self, mock_send_password_reset_email):
        token = 'ABCDEFGH' * 2
        redirection_url = redirection_url_for_token(token)
        response = self.do_post({'email': self.email.email,
                                 'token': token})
        self.assertEqual(response.status_code, 201)

        mock_send_password_reset_email.assert_called_once_with(
            self.account, self.email.email, redirection_url)


class TokensHandlerTestCase(BaseTestCase):
    url_name = 'api-token'

    def setUp(self):
        super(TokensHandlerTestCase, self).setUp()
        self.url = reverse(
            self.url_name, kwargs=dict(token=self.token.token))

    def assert_token_result(self, result, token):
        expected = dict(
            consumer_key=token.consumer.key,
            token_key=token.token, token_name=token.name,
            date_created=token.created_at,
            date_updated=token.updated_at,
            href=reverse(self.url_name, kwargs=dict(token=token.token)),
        )
        self.assertEqual(result, self.to_json(expected))

    def test_token_empty(self):
        self.assertRaises(
            NoReverseMatch, reverse, self.url_name, kwargs=dict(token=''))

    def test_unauthenticated_401(self):
        response = self.get_response()

        self.assertEqual(response.status_code, 401)

    def test_authenticated_with_django_401(self):
        assert self.client.login(username=self.account.preferredemail.email,
                                 password=DEFAULT_USER_PASSWORD)
        response = self.get_response()

        self.assertEqual(response.status_code, 401)

    def test_get_authenticated_with_oauth(self):
        response = self.get_response(token=self.token)
        self.assertEqual(response.status_code, 200)

        result = json.loads(response.content)
        self.assert_token_result(result, self.token)

    def test_secrets_not_returned(self):
        response = self.get_response(token=self.token)
        result = json.loads(response.content)

        for k in ('token_secret', 'consumer_secret'):
            self.assertNotIn(k, result)

    def test_can_get_others_if_elevated_permission(self):
        other_token = self.factory.make_account().create_oauth_token('other')
        self.factory.add_permission_to_account(
            self.account, 'api_view_account_details')

        url = reverse(self.url_name, kwargs=dict(token=other_token.token))
        response = self.get_response(url=url, token=self.token)
        self.assertEqual(response.status_code, 200)

        result = json.loads(response.content)
        self.assert_token_result(result, other_token)

    def test_forbidden_if_not_owner_not_elevated_permission_on_get(self):
        other_token = self.factory.make_account().create_oauth_token('other')

        url = reverse(self.url_name, kwargs=dict(token=other_token.token))
        response = self.get_response(url=url, token=self.token)

        self.assertEqual(response.status_code, 401)

    def test_get_token_does_not_exist(self):
        self.factory.add_permission_to_account(
            self.account, 'api_view_account_details')

        url = reverse(self.url_name, kwargs=dict(token='foobarbaz'))
        response = self.get_response(url=url, token=self.token)

        self.assertEqual(response.status_code, 404)

    def test_delete_authenticated_with_oauth(self):
        response = self.get_response(token=self.token, method='DELETE')

        self.assertEqual(response.status_code, 204)
        self.assertEqual(response.content, '')
        self.assertEqual(
            Token.objects.filter(token=self.token.token).count(), 0)

    def test_can_delete_others_if_elevated_permission(self):
        other_token = self.factory.make_account().create_oauth_token('other')
        self.factory.add_permission_to_account(
            self.account, 'api_edit_account_details')

        url = reverse(self.url_name, kwargs=dict(token=other_token.token))
        response = self.get_response(
            url=url, token=self.token, method='DELETE')

        self.assertEqual(response.status_code, 204)
        self.assertEqual(response.content, '')
        self.assertEqual(
            Token.objects.filter(token=other_token.token).count(), 0)

    def test_forbidden_if_not_owner_not_elevated_permission_on_delete(self):
        other_token = self.factory.make_account().create_oauth_token('other')

        url = reverse(self.url_name, kwargs=dict(token=other_token.token))
        response = self.get_response(
            url=url, token=self.token, method='DELETE')

        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            Token.objects.filter(token=other_token.token).count(), 1)

    def test_delete_token_does_not_exist(self):
        self.factory.add_permission_to_account(
            self.account, 'api_edit_account_details')

        url = reverse(self.url_name, kwargs=dict(token='foobarbaz'))
        response = self.get_response(
            url=url, token=self.token, method='DELETE')

        self.assertEqual(response.status_code, 404)


class ThrottledTestCase(BaseTestCase):
    def throttle_all(self):
        p = patch('piston.utils.cache')
        mock_cache = p.start()
        self.addCleanup(p.stop)

        # hit max number of requests allowed
        mock_cache.get.return_value = (3600, time.time() + 42.1)

    def throttle_by_openid(self, openid):
        p = patch('piston.utils.cache')
        mock_cache = p.start()
        self.addCleanup(p.stop)

        def mock_get(key, default=None, version=None):
            # only look for openid in extra part
            # the first part of the key is the requester username
            _, extra = key.split(':', 1)
            if openid in extra:
                return (3600, time.time() + 42.1)
            else:
                return (42, time.time() + 42.1)
        mock_cache.get = mock_get

    def throttle_by_email(self, email):
        p = patch('piston.utils.cache')
        mock_cache = p.start()
        self.addCleanup(p.stop)

        def mock_get(key, default=None, version=None):
            if email in key:
                return (3600, time.time() + 42.1)
            else:
                return (42, time.time() + 42.1)
        mock_cache.get = mock_get

    def throttle_by_token(self, token):
        p = patch('piston.utils.cache')
        mock_cache = p.start()
        self.addCleanup(p.stop)

        def mock_get(key, default=None, version=None):
            if token in key:
                return (3600, time.time() + 42.1)
            else:
                return (42, time.time() + 42.1)
        mock_cache.get = mock_get

    def assert_handler_throttled(self, request, token=None,
                                 throttle_func=None, **kwargs):
        if throttle_func is None:
            self.throttle_all()
        else:
            throttle_func()

        response = self.get_response(request.path, token=token,
                                     method=request.method, **kwargs)

        self.assertEqual(response.status_code, 429)
        self.assertEqual(response['Retry-After'], '42')

    def assert_handler_not_throttled(self, request, token=None,
                                     throttle_func=None, **kwargs):
        if throttle_func is None:
            self.throttle_all()
        else:
            throttle_func()

        response = self.get_response(request.path, token=token,
                                     method=request.method, **kwargs)

        self.assertNotEqual(response.status_code, 429)

    def test_anonymous_accounts_handler_throttled_by_openid(self):
        openid = self.account.openid_identifier
        url = reverse('api-account', args=[openid])
        request = RequestFactory().get(url)
        request.user = AnonymousUser()

        self.assert_handler_throttled(
            request, throttle_func=lambda: self.throttle_by_openid(openid))

        assert 'foo1234' != openid
        path = reverse('api-account', args=['foo1234'])
        request.path = path
        self.assert_handler_not_throttled(
            request, throttle_func=lambda: self.throttle_by_openid(openid))

    def test_accounts_handler_throttled_by_openid(self):
        openid = self.account.openid_identifier
        url = reverse('api-account', args=[openid])
        request = RequestFactory().get(url)
        request.user = self.account

        self.assert_handler_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_openid(openid))

        assert 'foo1234' != openid
        path = reverse('api-account', args=['foo1234'])
        request.path = path
        self.assert_handler_not_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_openid(openid))

    def test_accounts_handler_whitelisted(self):
        url = reverse('api-account', args=[self.account.openid_identifier])
        request = RequestFactory().get(url)
        request.user = self.account

        with patch_settings(THROTTLE_WHITELIST_EMAIL_REGEXP_LIST=[
                            self.account.preferredemail.email]):
            self.assert_handler_not_throttled(request, token=self.token)

    def test_password_reset_token_handler_throttled_by_email(self):
        url = reverse('api-password-reset')
        request = RequestFactory().post(url)
        request.user = AnonymousUser()
        email = self.account.preferredemail.email

        self.assert_handler_throttled(
            request, data=json.dumps({'email': email}),
            content_type='application/json',
            throttle_func=lambda: self.throttle_by_email(email))
        self.assert_handler_not_throttled(
            request, data=json.dumps({'email': 'foo@foo.com'}),
            content_type='application/json',
            throttle_func=lambda: self.throttle_by_email(email))

    def test_account_registration_handler_throttled_by_email(self):
        url = reverse('api-registration')
        request = RequestFactory().post(url)
        request.user = AnonymousUser()
        email = 'foo@foo.com'

        data = {
            'captcha_id': 'some-id',
            'captcha_solution': 'some solution',
            'email': 'foo@foo.com',
            'password': 'some-password-123',
            'displayname': 'Some User',
        }

        self.assert_handler_throttled(
            request, data=json.dumps(data),
            content_type='application/json',
            throttle_func=lambda: self.throttle_by_email(email))

        data['email'] = 'bar@foo.com'
        self.assert_handler_not_throttled(
            request, data=json.dumps(data),
            content_type='application/json',
            throttle_func=lambda: self.throttle_by_email(email))

    def test_account_login_handler_throttled_by_email(self):
        url = reverse('api-login')
        request = RequestFactory().post(url)
        request.user = AnonymousUser()
        email = self.account.preferredemail.email

        data = {
            'email': email,
            'password': DEFAULT_USER_PASSWORD,
            'token_name': 'foo',
        }

        self.assert_handler_throttled(
            request, data=json.dumps(data),
            content_type='application/json',
            throttle_func=lambda: self.throttle_by_email(email))

        data['email'] = 'bar@foo.com'
        self.assert_handler_not_throttled(
            request, data=json.dumps(data),
            content_type='application/json',
            throttle_func=lambda: self.throttle_by_email(email))

    def test_account_phone_login_handler_throttled_by_email(self):
        url = reverse('api-login-phone')
        request = RequestFactory().post(url)
        request.user = AnonymousUser()
        email = self.account.preferredemail.email

        data = {
            'email': email,
            'phone_id': 'some-phone-id',
            'password': DEFAULT_USER_PASSWORD,
            'token_name': 'foo',
        }

        self.assert_handler_throttled(
            request, data=json.dumps(data),
            content_type='application/json',
            throttle_func=lambda: self.throttle_by_email(email))

        data['email'] = 'bar@foo.com'
        self.assert_handler_not_throttled(
            request, data=json.dumps(data),
            content_type='application/json',
            throttle_func=lambda: self.throttle_by_email(email))

    def test_emails_handler_get_throttled_by_email(self):
        email = self.account.preferredemail.email
        url = reverse('api-email', args=[email])
        request = RequestFactory().get(url)
        request.user = self.account

        self.assert_handler_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_email(email))

        assert 'foo@foo.com' != email
        url = reverse('api-email', args=['foo@foo.com'])
        request.path = url
        self.assert_handler_not_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_email(email))

    def test_emails_handler_get_whitelisted(self):
        email = self.account.preferredemail.email
        url = reverse('api-email', args=[email])
        request = RequestFactory().get(url)
        request.user = self.account

        with patch_settings(THROTTLE_WHITELIST_EMAIL_REGEXP_LIST=[
                            self.account.preferredemail.email]):
            self.assert_handler_not_throttled(request, token=self.token)

    def test_email_handler_delete_throttled_by_email(self):
        email = self.account.preferredemail.email
        url = reverse('api-email', args=[email])
        request = RequestFactory().delete(url)
        request.user = self.account

        self.assert_handler_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_email(email))

        assert 'foo@foo.com' != email
        url = reverse('api-email', args=['foo@foo.com'])
        request.path = url
        self.assert_handler_not_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_email(email))

    def test_email_handler_delete_whitelisted(self):
        email = self.account.preferredemail.email
        url = reverse('api-email', args=[email])
        request = RequestFactory().delete(url)
        request.user = self.account

        with patch_settings(THROTTLE_WHITELIST_EMAIL_REGEXP_LIST=[
                            self.account.preferredemail.email]):
            self.assert_handler_not_throttled(request, token=self.token)

    def test_tokens_handler_get_throttled(self):
        token = self.token.token
        url = reverse('api-token', args=[self.token])
        request = RequestFactory().get(url)
        request.user = self.account

        self.assert_handler_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_token(token))

        assert 'some-token' != token
        path = reverse('api-token', args=['some-token'])
        request.path = path
        self.assert_handler_not_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_token(token))

    def test_tokens_handler_get_whitelisted(self):
        url = reverse('api-token', args=[self.token])
        request = RequestFactory().get(url)
        request.user = self.account

        with patch_settings(THROTTLE_WHITELIST_EMAIL_REGEXP_LIST=[
                            self.account.preferredemail.email]):
            self.assert_handler_not_throttled(request, token=self.token)

    def test_tokens_handler_delete_throttled(self):
        token = self.token.token
        url = reverse('api-token', args=[self.token])
        request = RequestFactory().delete(url)
        request.user = self.account

        self.assert_handler_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_token(token))

        assert 'some-token' != token
        path = reverse('api-token', args=['some-token'])
        request.path = path
        self.assert_handler_not_throttled(
            request, token=self.token,
            throttle_func=lambda: self.throttle_by_token(token))

    def test_tokens_handler_delete_whitelisted(self):
        url = reverse('api-token', args=[self.token])
        request = RequestFactory().delete(url)
        request.user = self.account

        with patch_settings(THROTTLE_WHITELIST_EMAIL_REGEXP_LIST=[
                            self.account.preferredemail.email]):
            self.assert_handler_not_throttled(request, token=self.token)

    def test_reports_metrics_when_throttled(self):
        email = self.account.preferredemail.email
        url = reverse('api-email', args=[email])
        request = RequestFactory().get(url)
        request.user = self.account
        called_metrics = []

        def incr(metric):
            called_metrics.append(metric)
        with patch('api.v20.decorators.statsd.incr', incr):
            self.assert_handler_throttled(request, token=self.token)
        self.assertEqual(
            [
                'throttling.throttled',
                'throttling.api.v20.handlers.EmailsHandler.read.throttled',
                'response.429',
                'response.auth.429',
            ], called_metrics)

    def test_doesnt_reports_metrics_when_not_throttled(self):
        url = reverse('api-account', args=[self.account.openid_identifier])
        request = RequestFactory().get(url)
        request.user = self.account
        called_metrics = []

        def incr(metric):
            called_metrics.append(metric)
        with patch('api.v20.decorators.statsd.incr', incr):
            with patch('piston.utils.cache') as mock_cache:
                mock_cache.get.return_value = (0, time.time() + 42.1)
                self.get_response(request.path, method=request.method)
        self.assertEqual(
            [
                'response.200',
            ], called_metrics)
