from StringIO import StringIO

from django.core.exceptions import ValidationError
from django.core.urlresolvers import reverse
from gargoyle.testutils import switches
from mock import ANY, patch, call as mock_call
from timeline import Timeline
from u1testutils.django import patch_settings

from identityprovider.models import Account, captcha
from identityprovider.models.const import AccountStatus
from identityprovider.tests.utils import SSOBaseTestCase

from api.v20 import (
    handlers,
    registration,
)
from api.v20.tests.utils import call

MODULE = 'api.v20.registration'
OVERIDES = dict(
    DISABLE_CAPTCHA_VERIFICATION=False,
    EMAIL_WHITELIST_REGEXP_LIST=['^canonicaltest(?:\+.+)?@gmail\.com$']
)


class RegistrationTestCase(SSOBaseTestCase):

    def setUp(self):
        super(RegistrationTestCase, self).setUp()
        self.email = self.factory.make_email_address()

        self.mock_send_impersonation_email = self._apply_patch(
            'api.v20.registration.send_impersonation_email')
        self.mock_send_new_user_email = self._apply_patch(
            'api.v20.registration.send_new_user_email')

    def test_register_already_registered_and_inactive(self):
        account = self.factory.make_account(
            email=self.email, status=AccountStatus.DEACTIVATED)
        assert not account.is_active

        with self.assertRaises(registration.EmailAlreadyRegistered):
            registration.register(self.email, "", "")

        self.assertFalse(self.mock_send_impersonation_email.called)
        self.assertFalse(self.mock_send_new_user_email.called)

    def test_register_already_registered_and_active(self):
        account = self.factory.make_account(email=self.email)
        assert account.is_active

        with self.assertRaises(registration.EmailAlreadyRegistered):
            registration.register(self.email, "", "")

        self.mock_send_impersonation_email.assert_called_once_with(
            self.email)
        self.assertFalse(self.mock_send_new_user_email.called)

    def test_register_success(self):
        assert Account.objects.all().count() == 0

        registration.register(
            self.email, 'SecretPassword1', 'displayname')

        self.assertEqual(Account.objects.all().count(), 1)
        account = Account.objects.get()
        self.assertEqual(account.displayname, 'displayname')
        self.assertEqual(account.preferredemail.email, self.email)
        self.assertFalse(account.is_verified)

        self.assertTrue(
            self.client.login(username=self.email, password='SecretPassword1'),
            'Password was not properly set on new account.')
        self.mock_send_new_user_email.assert_called_once_with(
            account=account, email=self.email)
        self.assertFalse(self.mock_send_impersonation_email.called)

    def test_invalid_email_and_pw(self):
        with self.assertRaises(ValidationError) as ctx:
            registration.register("", '', 'displayname')

        e = ctx.exception
        self.assertEqual(
            e.message_dict['email'][0],
            'This field cannot be blank.'
        )
        self.assertEqual(
            e.message_dict['password'][0],
            'Password must be at least 8 characters long'
        )


class RegistrationHandlerTestCase(SSOBaseTestCase):
    handler = handlers.AccountRegistrationHandler()
    url = reverse('api-registration')

    def setUp(self):
        super(RegistrationHandlerTestCase, self).setUp()
        self.data = {
            'email': self.factory.make_email_address(),
            'password': 'asdfASDF1',
            'displayname': 'Ricardo the Magnificent',
            'captcha_id': 'ignored',
            'captcha_solution': 'ignored',
        }

    def test_registration_handler_invalid_data(self):
        data = {
            'email': 'x',
            'password': 'y',
            'displayname': 'foo',
            'captcha_id': 'ignored',
            'captcha_solution': 'ignored',
        }

        response, json_body = call(self.handler.create, self.url, data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(json_body['code'], "INVALID_DATA")
        self.assertIn("Invalid request data", json_body['message'])
        self.assertIn('email', json_body['extra'])
        self.assertIn('password', json_body['extra'])
        self.assertNotIn('displayname', json_body['extra'])

    def assert_field_required(self, field):
        data = self.data.copy()
        data.pop(field)

        response, json_body = call(self.handler.create, self.url, data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(json_body['code'], "INVALID_DATA")
        self.assertIn("Invalid request data", json_body['message'])
        self.assertEqual(json_body['extra'], {field: 'Field required'})

    def test_no_email(self):
        self.assert_field_required('email')

    def test_no_password(self):
        self.assert_field_required('password')

    def test_no_displayname(self):
        self.assert_field_required('displayname')

    def test_registration_already_registered(self):
        self.account = self.factory.make_account(email=self.data['email'])

        response, json_body = call(self.handler.create, self.url, self.data)

        self.assertEqual(response.status_code, 409)
        self.assertEqual(json_body['code'], "ALREADY_REGISTERED")
        self.assertIn("already registered", json_body['message'])
        self.assertIn('email', json_body['extra'])
        self.assertIn(self.data['email'], json_body['extra']['email'])

    def test_registration_success(self):
        response, json_body = call(self.handler.create, self.url, self.data)

        self.assertEqual(response.status_code, 201)
        self.assertIn('openid', json_body)
        self.assertIn('href', json_body)
        self.assertEqual(json_body['email'], self.data['email'])
        self.assertEqual(json_body['displayname'], self.data['displayname'])
        self.assertEqual(json_body['status'], 'Active')
        self.assertEqual(len(json_body['emails']), 1)
        self.assertIn(self.data['email'], json_body['emails'][0]['href'])

    @patch('api.v20.handlers.Captcha')
    def test_register_captcha_required(self, mock_captcha):
        captcha_data = {'captcha_id': '999', 'image_url': 'somewhere'}
        del self.data['captcha_id']
        del self.data['captcha_solution']
        mock_captcha.new.return_value.serialize.return_value = captcha_data

        with switches(CAPTCHA=True):
            response, json_body = call(
                self.handler.create, self.url, self.data)

        self.assertEqual(response.status_code, 401)
        self.assertEqual(json_body['code'], "CAPTCHA_REQUIRED")
        self.assertIn('A captcha challenge is required', json_body['message'])
        self.assertEqual(Account.objects.all().count(), 0)
        self.assertEqual(json_body['extra'], captcha_data)

    @patch('api.v20.handlers.Captcha')
    def test_register_captcha_success(self, mock_captcha):
        mock_captcha.return_value.verify.return_value = True
        self.data['captcha_id'] = '999'
        self.data['captcha_solution'] = 'foo bar'

        response, json_body = call(self.handler.create, self.url, self.data)

        self.assertEqual(response.status_code, 201)
        expected_calls = mock_call('999').verify(
            'foo bar', '127.0.0.1', self.data['email'], timer=ANY).call_list()
        self.assertEqual(mock_captcha.mock_calls, expected_calls)

        self.assertIn('openid', json_body)
        self.assertIn('href', json_body)
        self.assertEqual(json_body['email'], self.data['email'])
        self.assertEqual(json_body['displayname'], self.data['displayname'])
        self.assertEqual(json_body['status'], 'Active')
        self.assertEqual(len(json_body['emails']), 1)
        self.assertIn(self.data['email'], json_body['emails'][0]['href'])

    @patch('api.v20.handlers.Captcha')
    def test_register_captcha_failure(self, mock_captcha):
        mock_captcha.return_value.verify.return_value = False
        self.data['captcha_id'] = '999'
        self.data['captcha_solution'] = 'foo bar'

        response, json_body = call(self.handler.create, self.url, self.data)
        self.assertEqual(response.status_code, 403)
        expected_calls = mock_call('999').verify(
            'foo bar', '127.0.0.1', self.data['email'], timer=ANY).call_list()
        self.assertEqual(mock_captcha.mock_calls, expected_calls)

        self.assertEqual(json_body['code'], "CAPTCHA_FAILURE")
        self.assertIn('Failed response to captcha challenge.',
                      json_body['message'])
        self.assertEqual(Account.objects.all().count(), 0)

    @patch('identityprovider.models.captcha.Captcha._open')
    def test_register_captcha_whitelist(self, mock_open):
        self.data['email'] = 'canonicaltest@gmail.com'
        self.data['captcha_id'] = '999'
        self.data['captcha_solution'] = 'foo bar'

        with patch_settings(**OVERIDES):
            response, json_body = call(
                self.handler.create, self.url, self.data
            )

        self.assertEqual(Account.objects.all().count(), 1)
        self.assertFalse(mock_open.called)

    @patch('identityprovider.models.captcha.Captcha._open')
    def test_register_captcha_whitelist_with_uuid(self, mock_open):
        self.data['email'] = 'canonicaltest+something@gmail.com'
        self.data['captcha_id'] = '999'
        self.data['captcha_solution'] = 'foo bar'

        with patch_settings(**OVERIDES):
            response, json_body = call(
                self.handler.create, self.url, self.data
            )

        self.assertEqual(Account.objects.all().count(), 1)
        self.assertFalse(mock_open.called)

    @patch('identityprovider.models.captcha.Captcha._open')
    def test_register_captcha_whitelist_fail(self, mock_open):
        self.data['captcha_id'] = '999'
        self.data['captcha_solution'] = 'foo bar'
        self.data['email'] = 'notcanonicaltest@gmail.com'
        mock_open.return_value.is_error = False
        mock_open.return_value.data.return_value = 'false\nmessage'

        with patch_settings(**OVERIDES):
            response, json_body = call(
                self.handler.create, self.url, self.data
            )
        self.assertEqual(Account.objects.all().count(), 0)
        self.assertTrue(mock_open.called)

    @patch('api.v20.handlers.Captcha')
    def test_register_captcha_http_error(self, mock_captcha):
        response = captcha.CaptchaResponse(
            500, None, "", "Server error", "Unexpected recaptcha error")
        error = captcha.VerifyCaptchaError(response)
        mock_captcha.return_value.verify.side_effect = error
        self.data['captcha_id'] = '999'
        self.data['captcha_solution'] = 'foo bar'

        response, json_body = call(self.handler.create, self.url, self.data)

        self.assertEqual(response.status_code, 502)
        self.assertEqual(json_body['code'], "CAPTCHA_ERROR")
        self.assertIn('Unable to get a valid response from reCaptcha service',
                      json_body['message'])
        extra = json_body['extra']
        self.assertEqual(extra['recaptcha_reason'], "Server error")
        self.assertEqual(extra['recaptcha_status_code'], 500)
        self.assertEqual(extra['recaptcha_body'], "Unexpected recaptcha error")
        self.assertEqual(Account.objects.all().count(), 0)

    @patch('api.v20.handlers.Captcha')
    def test_register_captcha_network_error(self, mock_captcha):
        response = captcha.CaptchaResponse(111, None, "", "Connection refused")
        error = captcha.VerifyCaptchaError(response)
        mock_captcha.return_value.verify.side_effect = error
        self.data['captcha_id'] = '999'
        self.data['captcha_solution'] = 'foo bar'

        response, json_body = call(self.handler.create, self.url, self.data)

        self.assertEqual(response.status_code, 502)
        self.assertEqual(json_body['code'], "CAPTCHA_ERROR")
        self.assertIn('Unable to get a valid response from reCaptcha service',
                      json_body['message'])
        extra = json_body['extra']
        self.assertEqual(extra['recaptcha_reason'], "Connection refused")
        self.assertEqual(extra['recaptcha_status_code'], None)
        self.assertEqual(extra['recaptcha_body'], None)
        self.assertEqual(Account.objects.all().count(), 0)

    @patch('identityprovider.models.captcha.Captcha._open')
    @switches(CAPTCHA=True)
    def test_timeline_for_captcha_generation(self, mock_open):
        timeline = Timeline()
        meta = {
            'timeline.timeline': timeline,
            'CONTENT_TYPE': 'application/json',
        }
        data = self.data.copy()
        data.pop('captcha_id')
        data.pop('captcha_solution')
        data['create_captcha'] = True

        response, json_body = call(self.handler.create, self.url, data, meta)

        self.assertEqual(1, len(timeline.actions))
        self.assertEqual('captcha-new', timeline.actions[0].category)
        self.assertEqual(mock_open.call_args[0][0],
                         timeline.actions[0].detail)

    @patch('identityprovider.models.captcha.Captcha._open')
    @switches(CAPTCHA=True)
    def test_timeline_for_captcha_verification(self, mock_open):
        mock_open.response = captcha.CaptchaResponse(500, StringIO(''))
        timeline = Timeline()
        meta = {
            'timeline.timeline': timeline,
            'CONTENT_TYPE': 'application/json',
        }

        with patch_settings(DISABLE_CAPTCHA_VERIFICATION=False):
            response, json_body = call(
                self.handler.create, self.url, self.data, meta)

        self.assertEqual(1, len(timeline.actions))
        self.assertEqual('captcha-verify', timeline.actions[0].category)
