import csv
import time
from datetime import datetime
from optparse import OptionValueError, make_option

from django.contrib.auth.models import make_password
from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand, CommandError
from django.core.validators import validate_email
from django.db import DatabaseError, connection
from psycopg2 import InternalError

from identityprovider.auth import LaunchpadBackend
from identityprovider.models import Account, LeakedCredential


def format_with_timestamp(msg, level=0):
    return "[%.2f] %s%s" % (time.time(), '  '*level, msg)


class Command(BaseCommand):
    args = '<input.csv> <leak source>'
    help = 'Import leaked credential data.'
    option_list = BaseCommand.option_list + (
        make_option('--read-batch-size', action='store',
                    dest='read_batch_size', type=int, default=10000),
        make_option('--insert-batch-size', action='store',
                    dest='insert_batch_size', type=int),
        make_option('--dry-run', action='store_true', dest='dry_run',
                    default=False),
        make_option('--date-leaked', action='store', dest='date_leaked'),
    )

    def handle(self, *args, **kwargs):
        if len(args) < 1:
            raise CommandError('Need to specify an input file.')
        elif len(args) < 2:
            raise CommandError('Need to specify a leak source.')
        elif len(args) > 2:
            raise CommandError('Too many arguments provided.')

        # get args
        filename, self.source = args

        self.verbosity = int(kwargs.get('verbosity', 1))
        # get options
        read_batch_size = kwargs.get('read_batch_size', 10000)
        insert_batch_size = kwargs.get('insert_batch_size', None)
        dry_run = kwargs.get('dry_run', False)
        date_leaked = kwargs.get('date_leaked', None)
        if date_leaked is not None:
            try:
                self.date_leaked = datetime.strptime(
                    date_leaked, '%Y-%m-%d').date()
            except ValueError:
                raise OptionValueError(
                    "option --date-leaked: invalid date: %r" % date_leaked)
        else:
            self.date_leaked = datetime.utcnow().date()

        with open(filename, 'rb') as csvfile:
            reader = self.get_reader(csvfile)

            batch_start = 1
            batch_end = read_batch_size
            self.write('Reading rows %d-%d ...' % (batch_start, batch_end))
            rows = self.read_rows(reader, read_batch_size)
            while rows:
                if not dry_run:
                    self.write('creating items', level=1)
                    items = self.make_items(rows, insert_batch_size)

                    self.write('inserting items', level=1)
                    self.insert_items(items, insert_batch_size)

                batch_start = batch_end + 1
                batch_end += read_batch_size
                self.write('Reading rows %d-%d ...' % (batch_start, batch_end))
                rows = self.read_rows(reader, read_batch_size)

    def write(self, msg, level=0, min_verbosity=0):
        if self.verbosity > min_verbosity:
            self.stdout.write(
                msg,
                style_func=lambda x: format_with_timestamp(x, level=level))

    def get_reader(self, fp):
        def unicode_reader(fp):
            reader = csv.reader(fp)
            for row in reader:
                yield [unicode(cell, 'utf-8') for cell in row]

        return unicode_reader(fp)

    def read_rows(self, reader, count):
        rows = []
        for i in xrange(count):
            try:
                rows.append(reader.next())
            except StopIteration:
                break
        return rows

    def make_item(self, row):
        try:
            email, password = row
            # email must be valid
            validate_email(email)
            # password can't be empty
            assert password

            # check if there is a matching account, so that we hash
            # the compromised password using the same algorithm+salt
            # for easier comparison and better security (not having plain-text
            # passwords lying around)
            account = Account.objects.get_by_email(email)
            if account is not None:
                backend = LaunchpadBackend()
                try:
                    if backend._validate_raw_password(
                            account.accountpassword, password,
                            update=False):
                        # current account's password is leaked
                        # force a password reset
                        account.reset_password(email)
                except ValidationError:
                    # password did not match
                    pass

            # store leak data using current hashing defaults
            item = LeakedCredential(
                email=email, password=make_password(password),
                when_leaked=self.date_leaked, source=self.source)
            self.write('item created', level=2, min_verbosity=1)
        except (AssertionError, ValidationError, ValueError):
            item = None
            self.write('skipping malformed row: %s' % row,
                       level=2, min_verbosity=1)
        return item

    def make_items(self, rows, batch_size):
        if batch_size is None:
            batch_size = len(rows)

        items = []
        for i, row in enumerate(rows):
            if i % batch_size == 0:
                self.write('making items %d-%d ...' % (i+1, i+batch_size),
                           level=1)
            item = self.make_item(row)
            if item is not None:
                items.append(item)
        return items

    def insert_items(self, items, batch_size):
        if batch_size is None:
            # do item by item insert
            for item in items:
                self.insert_item(item)
        else:
            # find last id
            leaks = LeakedCredential.objects.all().order_by('-id')
            if leaks:
                next_id = leaks[0].id + 1
            else:
                next_id = 1
            while items:
                batch = []
                for item in items[:batch_size]:
                    item.id = next_id
                    batch.append(item)
                    next_id += 1
                items = items[batch_size:]
                try:
                    # attempt bulk insert
                    LeakedCredential.objects.bulk_create(batch)
                    self.set_sequence(
                        'identityprovider_leakedcredential_id_seq',
                        next_id - 1)
                except (DatabaseError, InternalError):
                    # but fall back to single insert if bulk insert fails
                    for item in batch:
                        self.insert_item(item)

    def set_sequence(self, sequence, value):
        cursor = connection.cursor()
        cursor.execute(
            "select setval('%s', %%s);" % sequence, (value,))

    def insert_item(self, item):
        try:
            LeakedCredential.objects.get_or_create(
                email=item.email, source=item.source,
                defaults=dict(password=item.password,
                              when_leaked=item.when_leaked))
        except (DatabaseError, InternalError) as e:
            self.stderr.write(
                "error while inserting item: %s. "
                "error was: %s" % (item, str(e)))
