from __future__ import absolute_import

import os.path
from datetime import date
from importlib import import_module

import preflight
from bzrlib.branch import Branch
from bzrlib.errors import (
    NotBranchError,
    UnknownFormatError,
    UnsupportedFormatError,
)

from preflight.views import allow_basic_auth
from gargoyle import gargoyle

# Remember that the preflight may be run from the command-line or
# queried from a web-browser.  Some tests may always pass in one
# environment (e.g. if you can even *hit* the preflight page in a
# browser, you already know the DB is up), and that's okay.
#
# Make sure that each test at least works in both environment: a test
# shouldn't fail in either environment unless there's really a
# problem.


def get_version_info(name):
    version_info = {}

    try:
        mod = import_module(name)
        version_info['name'] = mod.__name__
    except ImportError:
        mod = None
        version_info['name'] = name

    if mod is None:
        path = os.path.realpath(os.path.abspath(os.path.join(
            __file__, '..', '..', '..', 'branches', name)))
    else:
        path = os.path.realpath(os.path.abspath(mod.__file__))

    try:
        branch, _ = Branch.open_containing(path)
        tip = branch.repository.get_revision(
            branch.get_rev_id(branch.revno()))
        tip_date_committed = date.fromtimestamp(
            tip.timestamp).strftime("%y.%m.%d")
        version = "{name}@{revno} ({date})".format(
            name=branch.nick, revno=branch.revno(),
            date=tip_date_committed,
        )
    except (NotBranchError, UnknownFormatError,
            UnsupportedFormatError):
        if mod is None:
            version = 'version not available'
        else:
            version = getattr(mod, '__version__', 'version not available')

    version_info['version'] = version
    return version_info


@preflight.register
class SSOPreflight(preflight.Preflight):

    @allow_basic_auth
    def authenticate(self, request):
        return gargoyle.is_active('PREFLIGHT', request.user)

    def versions(self):
        MODULES = [
            'django',
            'django_openid_auth',
            'identityprovider',
            'openid',
            'piston',
            'preflight',
            'ssoclient',
        ]
        BRANCHES = [
            'config',
        ]

        version_info = []
        for module in MODULES:
            version_info.append(get_version_info(module))
        for branch in BRANCHES:
            version_info.append(get_version_info(branch))

        return version_info

    def check_database(self):
        'Are database connections accepted?'
        from django import db

        cursor = db.connection.cursor()
        cursor.execute('SELECT 42')
        cursor.fetchone()

        return True
