# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import json
import os
import socket
import stat
import urllib
import urllib2

from django.conf import settings

# When a request arrives, the middleware will:
# 1. check_readonly(), to reconcile our settings with the flagfiles.
#      Master DB recovery is attempted within this method call.
# 2. Iterating over (ReadOnlyManager.)connections, find the first connection
#      available for which is_failed() returns False.
#      This method just checks marker files, doesn't actually try to connect.
# 3. Call set_db() for that connection
# 4. Call ping_current_connection() to check if the connection is really live.
# 5. If ping_current_connection() returns False, call mark_failed() for that
#      connection (or mark_current_failed() which is just a shortcut)
#      and go back to step 2, until success or no connections left.
#      It will pass automatic=True into mark_failed, as this is the failover
#      code.

# Readonly commands and/or the readonly views will also use:
# - is_failed, next_recovery_due, to compile information to send
#   out in calls to /readonlydata and show in readonly --list.
# - connections, to iterate over all configured connections.
# - set_readonly, clear_readonly, mark_failed, clear_failed, to manage
#   the different readonly status information.  These views will always omit
#   automatic=True, as they manage manual overrides.


def _remote_req(host, port=None, scheme=None, server_id=None,
                virtual_host=None, post=None):
    """Makes a request to a specific appserver.

    The first five arguments are the same as the keys for each dictionary in
    the APP_SERVERS setting:

    * host: The host at which we can reach the app server (can be an IP)
    * port: The port that should be used (optional)
    * scheme: The scheme to use to contact this app server (http/https)
              (defaults to http)
    * server_id: Some canonical name for this app server (optional)
    * virtual_host: The virtual host that should be used, for app servers that
                    serve multiple sites (optional).

    If post is provided, it should be a sequence of 2-tuples that will be
    encoded in to POST data.
    """
    if post is None:
        post = []
    post.append(('secret', settings.READONLY_SECRET))
    post = urllib.urlencode(post)

    if scheme is None:
        scheme = 'http'
    if port is None:
        portstr = ''
    else:
        portstr = ':' + port
    url = '%s://%s%s/readonlydata' % (scheme, host, portstr)
    headers = {}
    if virtual_host is not None:
        headers['Host'] = virtual_host
    req = urllib2.Request(url, headers=headers, data=post)
    try:
        oldtimeout = socket.getdefaulttimeout()
        socket.setdefaulttimeout(5)
        datafile = urllib2.urlopen(req)
        data = datafile.read()
        socket.setdefaulttimeout(oldtimeout)
    except urllib2.URLError:
        data = None
    return data


def lowercase_keys(dicts):
    """ Convert all keys in a list of dicts to lowercase. """
    return [dict((k.lower(), v) for (k, v) in d.items()) for d in dicts]


def get_server_atts(servers):
    """Provides a report about readonly status of all app servers."""
    appservers = []
    set_all_readonly = False
    clear_all_readonly = False
    for server in lowercase_keys(servers):
        datastr = _remote_req(**server)
        if datastr is None:
            data = {'reachable': False}
        else:
            try:
                data = json.loads(datastr)
            except ValueError:
                # This is probably caused by the remote database being out
                # of sync with our current database
                data = {'reachable': False}
            else:
                data['reachable'] = True
            if data.get('readonly'):
                clear_all_readonly = True
            else:
                set_all_readonly = True
        data['name'] = server['server_id']
        appservers.append(data)
    atts = {
        'appservers': appservers,
        'clear_all_readonly': clear_all_readonly,
        'set_all_readonly': set_all_readonly,
    }
    return atts


def update_server(action, appserver=None):
    if appserver is None:
        appservers = settings.APP_SERVERS
    else:
        appservers = [server for server in settings.APP_SERVERS
                      if server['SERVER_ID'] == appserver]
    appservers = lowercase_keys(appservers)
    for server in appservers:
        post = [('action', action)]
        _remote_req(post=post, **server)


class ReadOnlyManager(object):

    @property
    def readonly_marker_file(self):
        flagdir = getattr(settings, 'DBFAILOVER_FLAG_DIR', '/tmp')
        return os.path.join(flagdir, 'db.readonly')

    def check_readonly(self):
        """Check readonly flag file and settings for consistency."""
        settings.READ_ONLY_MODE = os.path.exists(self.readonly_marker_file)
        return settings.READ_ONLY_MODE

    def set_readonly(self):
        """Create a marker file to indicate we're in readonly mode."""
        settings.READ_ONLY_MODE = True
        open(self.readonly_marker_file, 'a').close()
        mode = os.stat(self.readonly_marker_file)[stat.ST_MODE]
        group_perms = stat.S_IRGRP | stat.S_IWGRP
        if mode & group_perms != group_perms:
            os.chmod(self.readonly_marker_file, mode | group_perms)

    def clear_readonly(self):
        """Clear the marker file to leave readonly mode."""
        settings.READ_ONLY_MODE = False
        if os.path.exists(self.readonly_marker_file):
            os.remove(self.readonly_marker_file)
