from adminaudit.schema import AdminAuditSchema
from configglue.contrib import (
    DjangoOpenIdAuthSchema,
    NexusSchema,
    PreflightSchema,
)
from configglue.schema import (
    BoolOption,
    Section,
    DictOption,
    IntOption,
    ListOption,
    StringOption,
    TupleOption,
    merge,
)
from django_configglue.schema import schemas
from oops_dictconfig.configglue_options import OopsOption

from ubuntu_sso_saml.schema import Saml2IdpSchema


DjangoSchema = schemas.get('1.5.4')


class UpperCaseDictOption(DictOption):
    """ A DictOption with all upper-case keys. """
    def parse(self, section, parser=None, raw=False):
        parsed = super(UpperCaseDictOption, self).parse(
            section, parser, raw)
        result = {}
        for k, v in parsed.items():
            result[k.upper()] = v
        return result


class SSOSchema(DjangoSchema):
    # default
    database_slaves = ListOption(item=StringOption(),
                                 help='List of databases who are configured '
                                      'as read-only slaves.')
    writable_ro_mode_models = ListOption(item=StringOption(),
                                         help='List of app.Model entries '
                                              'which will stay writable in '
                                              'readonly mode.')
    handler_timeout_millis = IntOption(default=10000)
    oops_dir = StringOption(help='Absolute path to the directory oops'
                                 ' reports will be stored in')
    pgconnect_timeout = IntOption(default=10)
    google_analytics_id = StringOption()
    support_phone = StringOption()

    # api
    class api(Section):
        api_host = StringOption()
        api_url = StringOption()
        api_use_internal = BoolOption(default=False)
        oauth_data_store = StringOption()
        throttle_whitelist_email_regexp_list = ListOption(
            item=StringOption(raw=True))

    # branding
    class branding(Section):
        brand = StringOption()
        brand_descriptions = DictOption(item=StringOption(), spec={
            'launchpad': StringOption(default='Launchpad Login Service'),
            'ubuntuone': StringOption(default='Ubuntu One'),
        })

    # captcha
    class captcha(Section):
        captcha_use_proxy = BoolOption()
        captcha_proxies = DictOption()
        captcha_api_url = StringOption()
        captcha_api_url_secure = StringOption()
        captcha_image_url_pattern = StringOption(raw=True)
        captcha_verify_url = StringOption()
        captcha_public_key = StringOption()
        captcha_private_key = StringOption()
        disable_captcha_verification = BoolOption()
        email_whitelist_regexp_list = ListOption(
            item=StringOption(raw=True)
        )

    # debug
    class debug(Section):
        testing = BoolOption(default=False)

    class django_secure(Section):
        secure_ssl_redirect = BoolOption(default=True)
        secure_hsts_seconds = IntOption(default=300)
        secure_hsts_include_subdomains = BoolOption(default=True)
        secure_frame_deny = BoolOption(default=True)
        secure_content_type_nosniff = BoolOption(default=True)
        secure_browser_xss_filter = BoolOption(default=True)

    class django_statsd(Section):
        statsd_patches = ListOption()

    # override django section
    class django(DjangoSchema.django):
        csrf_cookie_httponly = BoolOption(default=False)

    # email
    class email(Section):
        noreply_from_address = StringOption()
        feedback_to_address = StringOption()

    class gargoyle(Section):
        gargoyle_switch_defaults = DictOption(item=DictOption())

    #i18n
    class i18n(Section):
        supported_languages = ListOption(
            item=StringOption(),
            help="List of currently supported languages")
        language_names = DictOption(
            help="Mapping between language-code and name in that language")

    class oopses(Section):
        oopses = OopsOption()

    # openid
    class openid(Section):
        pre_authorization_validity = IntOption()
        openid_preauthorization_acl = ListOption(
            item=TupleOption(length=2))

    class profile(Section):
        profile_filename = StringOption(
            default="sso.profile",
            help="Filename of the file created by the profile middleware")
        profile_pattern = StringOption(
            default="(api|identityprovider|ubuntu_sso_saml|webui)\..*",
            help="Regex pattern for limiting functions to be considered for "
                 "profiling")
        timelog_log = StringOption(
            default='timelog.log',
            help='Filename of the file created by the TimeLogMiddleware')

    # readonly
    class readonly(Section):
        app_servers = ListOption(item=UpperCaseDictOption(
            spec={'server_id': StringOption(),
                  'host': StringOption(),
                  'port': StringOption(),
                  'scheme': StringOption(default='http'),
                  'virtual_host': StringOption(),
                  },
            strict=True))
        readonly_secret = StringOption(raw=True)
        dbfailover_flag_dir = StringOption()
        read_only_mode = BoolOption(default=False)

    class salesforce_portal_attributes(Section):
        organization_id = StringOption()
        portal_id = StringOption()

    class south(Section):
        south_database_adapters = DictOption()

    # sso
    class sso(Section):
        sso_media_root = StringOption()
        sso_restrict_rp = BoolOption()
        sso_root_url = StringOption()
        sso_provider_url = StringOption()
        sso_account_update_url = StringOption(
            null=True,
            help="URL to POST when account data has changed")
        embedded_trust_root = StringOption()
        max_password_reset_tokens = IntOption(default=5)
        combo_url = StringOption(default="/combo/")
        combine = BoolOption(default=False)
        pci_compliance_group = StringOption(default="pci-compliance")

    class static_urls(Section):
        support_form_url = StringOption()
        twofactor_faq_url = StringOption()
        yubikey_personalize_url = StringOption()

        apps_url = StringOption()
        photos_url = StringOption()
        music_url = StringOption()
        cloud_url = StringOption()

        terms_of_service_url = StringOption()
        privacy_url = StringOption()

    class statsd(Section):
        statsd_client = StringOption(
            default='django_statsd.clients.normal')
        statsd_host = StringOption(
            default='localhost')
        statsd_port = IntOption(default=8125)
        statsd_prefix = StringOption(null=True)

    class twofactor(Section):
        hotp_drift = IntOption()
        hotp_backwards_drift = IntOption()
        hotp_sync_window = IntOption(default=10)
        twofactor_service_ident = StringOption(default='UbuntuSSO')
        twofactor_ttl = IntOption(
            default=8 * 3600,
            help="Number of seconds for the twofactor session to be valid")
        twofactor_freshness = IntOption(
            default=15 * 60,
            help="Number of seconds for the twofactor session to be "
                 "considered 'fresh'")
        twofactor_max_attempts = IntOption(
            default=18,
            help="Maximum allowed number of attempts to authenticate with a "
                 "two factor device before the account is locked")
        twofactor_paper_codes = IntOption(
            default=25,
            help="Number of backup codes given to the user in printable lists")
        twofactor_paper_codes_warn_renewal = IntOption(
            default=3,
            help="Number of unused codes left on printed code sheet before "
                 "warning the user")
        twofactor_paper_codes_allow_generation = IntOption(
            default=twofactor_paper_codes.default,
            help="Number of unused codes left on printed code sheet before "
                 "allowing the user to generate new codes")
        twofactor_failure_notification_threshold = IntOption(
            default=2,
            help="Number of times someone can fail at entering the 2fa code "
                 "before an email with a warning notification is sent to "
                 "TWOFACTOR_FAILURE_NOTIFICATION_EMAIL")
        twofactor_failure_notification_teams = ListOption(
            default=[],
            help="Team that someone has to belong to in order to send 2fa "
                 "an email with a warning notification to "
                 "TWOFACTOR_FAILURE_NOTIFICATION_EMAIL")
        twofactor_failure_notification_emails = ListOption(
            default=[],
            help="Email addresses where warning notifications are sent about "
                 "too many failures when entering 2fa code.")

    # testing
    class testing(Section):
        # We use sso_test_account_* credentials also on pay and u1 projects.
        # The helpers from u1testutils.sso.data.User will create a user data
        # object based on this values and the email_address_pattern.
        sso_test_account_full_name = StringOption()
        sso_test_account_email = StringOption()
        sso_test_account_password = StringOption()
        # We shouldn't need test_account_* credentials.
        # TODO refactor and remove them. All the tests that rely on this config
        # should create a new user. -- elopio 2013-04-22
        test_account_email = StringOption()
        test_account_password = StringOption()
        email_address_pattern = StringOption(
            raw=True, default=r'isdtest+%s@canonical.com')
        imap_server = StringOption()
        imap_port = IntOption(default=993)
        imap_use_ssl = BoolOption(default=True)
        imap_username = StringOption()
        imap_password = StringOption()
        test_discover_root = StringOption()
        test_discover_top_level = StringOption()

    # send warning emails/perform cleanup actions on accounts
    class warnings(Section):
        suspend_unverified_account_after_days = IntOption(
            default=90,
            help='Unverified accounts will be suspended after this period',
        )
        delete_unverified_account_after_days = IntOption(
            default=180,
            help=('Suspended and unverified accounts will be deleted after '
                  'this period'),
        )
        warn_suspend_unverified_account_before_days = IntOption(
            default=15,
            help=('Warning emails about upcoming unverified account '
                  'suspension will be sent with this buffer of notification'),
        )
        warn_delete_unverified_account_before_days = IntOption(
            default=15,
            help=('Warning emails about upcoming unverified and suspended '
                  'account deletion will be sent with this buffer of '
                  'notification'),
        )


# merge all contrib schemas into the base schema
# order matters
schema = merge(
    SSOSchema,
    DjangoOpenIdAuthSchema,
    NexusSchema,
    PreflightSchema,
    Saml2IdpSchema,
    AdminAuditSchema,
)
