# encoding: utf-8
# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from django.conf import settings
from django.contrib.auth.hashers import get_hasher
from django.contrib.auth.models import (
    User,
    check_password,
    is_password_usable,
    make_password,
)
from django.db import connection
from mock import (
    Mock,
    patch,
)
from oauth_backend.models import (
    Consumer,
    Token,
)
from piston.oauth import OAuthError as PistonOAuthError

from identityprovider.auth import basic_authenticate
from identityprovider.auth import (
    LaunchpadBackend,
    SSOOAuthAuthentication,
    oauth_authenticate,
)
from identityprovider.models import (
    Account,
    EmailAddress,
)
from identityprovider.models.const import AccountStatus
from identityprovider.tests import DEFAULT_API_PASSWORD, DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import SSOBaseTestCase
from identityprovider.utils import (
    encrypt_launchpad_password,
    validate_launchpad_password,
)


class LaunchpadBackendTestCase(SSOBaseTestCase):

    email = 'mark@example.com'

    def setUp(self):
        super(LaunchpadBackendTestCase, self).setUp()
        self.account = self.factory.make_account(
            email=self.email, password=DEFAULT_USER_PASSWORD)
        self.backend = LaunchpadBackend()

    def test_authenticate_with_email_status_not_in_expected_one(self):
        email_address = EmailAddress.objects.get(
            email__iexact=self.email)
        email_address.status = 9999
        email_address.save()

        result = self.backend.authenticate(self.email,
                                           DEFAULT_USER_PASSWORD)

        self.assertIsNone(result)

    def test_get_user_does_not_exist(self):
        user = self.backend.get_user(9999)
        self.assertIsNone(user)

    def test_authenticate_with_email_case_insensitive(self):
        # Make sure authentication works as expected
        account1 = self.backend.authenticate(self.email,
                                             DEFAULT_USER_PASSWORD)
        self.assertIsNotNone(account1)

        # Try using different case for email
        account2 = self.backend.authenticate('Mark@Example.com',
                                             DEFAULT_USER_PASSWORD)
        self.assertIsNotNone(account2)

        # Make sure both accounts are the same
        self.assertEqual(account1, account2)

    def test_authenticate_with_launchpad_encrypted_password(self):
        """The backend is able to authenticate with Launchpad cypher."""
        ap = self.account.accountpassword
        ap.password = encrypt_launchpad_password(DEFAULT_USER_PASSWORD)
        ap.save()

        account = self.backend.authenticate(self.email, DEFAULT_USER_PASSWORD)
        self.assertEqual(account, self.account)

    def test_changes_cypher_after_verifying_with_launchpad(self):
        """Change to the Django cypher when needed."""
        ap = self.account.accountpassword
        ap.password = encrypt_launchpad_password(DEFAULT_USER_PASSWORD)
        ap.save()
        self.assertFalse(check_password(DEFAULT_USER_PASSWORD, ap.password))
        self.assertTrue(validate_launchpad_password(DEFAULT_USER_PASSWORD,
                                                    ap.password))

        account = self.backend.authenticate(self.email, DEFAULT_USER_PASSWORD)
        password = account.accountpassword.password
        self.assertTrue(check_password(DEFAULT_USER_PASSWORD, password))
        self.assertFalse(validate_launchpad_password(DEFAULT_USER_PASSWORD,
                                                     password))

    def test_changes_cypher_do_not_removes_oauth_tokens(self):
        """Change to the Django cypher when needed."""
        ap = self.account.accountpassword
        ap.password = encrypt_launchpad_password(DEFAULT_USER_PASSWORD)
        ap.save()
        assert not check_password(DEFAULT_USER_PASSWORD, ap.password)
        assert validate_launchpad_password(DEFAULT_USER_PASSWORD, ap.password)

        # create a bunch of tokens for the account
        tokens = []
        for i in xrange(5):
            t = self.account.create_oauth_token(token_name='test-%s' % i)
            tokens.append(t)
        assert self.account.oauth_tokens().count() == 5

        account = self.backend.authenticate(self.email, DEFAULT_USER_PASSWORD)
        assert account.id == self.account.id
        self.assertEqual(account.oauth_tokens().count(), 5)

    def test_authenticate_with_django_encrypted_password(self):
        """The backend is able to authenticate with Django cypher."""
        ap = self.account.accountpassword
        ap.password = make_password(DEFAULT_USER_PASSWORD)
        ap.save()

        account = self.backend.authenticate(self.email, DEFAULT_USER_PASSWORD)
        self.assertEqual(account, self.account)

    def test_changes_hasher_upon_authentication(self):
        """The backend changes from PBKDF2 to bcrypt upon authentication."""
        ap = self.account.accountpassword
        ap.password = make_password(DEFAULT_USER_PASSWORD,
                                    hasher='pbkdf2_sha256')
        ap.save()

        account = self.backend.authenticate(self.email, DEFAULT_USER_PASSWORD)

        encrypted = account.accountpassword.password
        hasher = get_hasher('bcrypt13')
        self.assertTrue(hasher.verify(DEFAULT_USER_PASSWORD, encrypted))

    def test_doesnt_save_password_again_if_not_needed(self):
        """The backend doesn't update the password if not needed."""
        ap = self.account.accountpassword
        ap.password = make_password(DEFAULT_USER_PASSWORD)
        ap.save()

        with patch.object(settings, 'DEBUG', True):
            self.backend.authenticate(self.email, DEFAULT_USER_PASSWORD)

        for query in connection.queries:
            sql = query['sql']
            if sql.startswith('UPDATE "accountpassword" SET "password"'):
                self.fail("Shouldn't have updated the password.")

    def test_authenticate_account_active(self):
        account = Account.objects.get_by_email(self.email)
        # make sure account is active
        self.assertEqual(account.status, AccountStatus.ACTIVE)
        # make sure authentication succeeds
        response = self.backend.authenticate(self.email,
                                             DEFAULT_USER_PASSWORD)
        self.assertEqual(response, account)

    def test_authenticate_by_token_missing_account(self):
        user = User.objects.create_user('test', 'test',
                                        password=DEFAULT_USER_PASSWORD)
        consumer = Consumer.objects.create(user=user)
        token = Token.objects.create(consumer=consumer, name='test')

        self.assertIsNone(self.backend.authenticate(token=token))

    def test_authenticate_account_inactive(self):
        account = Account.objects.get_by_email(self.email)
        _status = account.status

        for status, _ in AccountStatus._get_choices():
            if status == AccountStatus.ACTIVE:
                # skip as this is tested elsewhere
                continue

            account.status = status
            account.save()

            # make sure authentication fails
            response = self.backend.authenticate(self.email,
                                                 DEFAULT_USER_PASSWORD)
            self.assertEqual(response, None)

        # leave everything as it was
        account.status = _status
        account.save()

    def test_authenticate_account_no_password(self):
        account = Account.objects.get_by_email(self.email)
        account.accountpassword.delete()

        response = self.backend.authenticate(self.email,
                                             DEFAULT_USER_PASSWORD)

        self.assertIsNone(response)
        account = Account.objects.get_by_email(self.email)
        self.assertIsNotNone(account)

    def test_authenticate_account_with_leaked_password(self):
        leak = self.factory.make_leaked_credential(
            email=self.email, password=DEFAULT_USER_PASSWORD)
        account = Account.objects.get_by_email(self.email)
        # make sure account is not flagged for password reset
        self.assertFalse(account.need_password_reset)

        assert leak.password != account.encrypted_password

        account = self.backend.authenticate(
            self.email, DEFAULT_USER_PASSWORD)
        # verify account was flagged for password reset
        self.assertTrue(account.need_password_reset)

    def test_oauth_authenticate_account_active(self):
        account = Account.objects.get_by_email(self.email)
        user, _ = User.objects.get_or_create(
            username=account.openid_identifier)
        consumer, created = Consumer.objects.get_or_create(user=user)
        token = account.create_oauth_token('new-token')
        oauth_token = token.oauth_token()

        # make sure the account is active
        self.assertEqual(account.status, AccountStatus.ACTIVE)

        # make sure authentication succeeds
        response = oauth_authenticate(consumer, oauth_token, None)
        self.assertEqual(response, account)

    def test_oauth_authenticate_account_inactive(self):
        account = Account.objects.get_by_email(self.email)
        _status = account.status
        user, _ = User.objects.get_or_create(
            username=account.openid_identifier)

        consumer, created = Consumer.objects.get_or_create(user=user)
        token = account.create_oauth_token('new-token')
        oauth_token = token.oauth_token()

        for status, _ in AccountStatus._get_choices():
            if status == AccountStatus.ACTIVE:
                # skip as this is tested elsewhere
                continue

            account.status = status
            account.save()

            # make sure authentication fails
            response = oauth_authenticate(consumer, oauth_token, None)
            self.assertEqual(response, None)

        # leave everything as it was
        account.status = _status
        account.save()

        if created:
            consumer.delete()

    def test_oauth_authenticate_stolen_token(self):
        victim_account = Account.objects.get_by_email(self.email)
        token = victim_account.create_oauth_token('new-token')
        oauth_token = token.oauth_token()

        malicious_account = self.factory.make_account()
        malicious_user, _ = User.objects.get_or_create(
            username=malicious_account.openid_identifier)
        consumer, created = Consumer.objects.get_or_create(user=malicious_user)

        # make sure the accounts are active
        self.assertEqual(victim_account.status, AccountStatus.ACTIVE)
        self.assertEqual(malicious_account.status, AccountStatus.ACTIVE)

        # make sure authentication succeeds
        response = oauth_authenticate(consumer, oauth_token, None)
        self.assertFalse(response)

    def test_password_leaked_no_account(self):
        self.factory.make_leaked_credential(email=self.email,
                                            password=DEFAULT_USER_PASSWORD)
        self.account.delete()

        leaked = LaunchpadBackend().password_leaked(
            self.email, DEFAULT_USER_PASSWORD, None)
        self.assertTrue(leaked)


class SSOOAuthAuthenticationTestCase(SSOBaseTestCase):

    def setUp(self):
        super(SSOOAuthAuthenticationTestCase, self).setUp()
        self.auth = SSOOAuthAuthentication()
        # create a valid request
        self.request = Mock()
        self.request.META = {}
        self.request.REQUEST = {
            'oauth_consumer_key': '',
            'oauth_token': '',
            'oauth_signature': '',
            'oauth_signature_method': '',
            'oauth_timestamp': '',
            'oauth_nonce': '',
        }

    def test_is_authenticated_invalid_request(self):
        mock_is_valid_request = Mock(return_value=False)
        with patch.object(self.auth, 'is_valid_request',
                          mock_is_valid_request):
            self.assertEqual(self.auth.is_authenticated(self.request), False)

    @patch('sys.stdout')
    def test_is_authenticated_piston_oauth_error(self, sys_mock):
        mock_validate_token = Mock(side_effect=PistonOAuthError)
        with patch.object(self.auth, 'validate_token', mock_validate_token):
            self.assertEqual(self.auth.is_authenticated(self.request), False)

    def test_is_authenticated_no_consumer_token(self):
        consumer, token, params = None, None, None
        mock_validate_token = Mock(return_value=(consumer, token, params))
        # authentication fails as there is no consumer and token in the request
        with patch.object(self.auth, 'validate_token', mock_validate_token):
            self.assertEqual(self.auth.is_authenticated(self.request), False)

    @patch('identityprovider.auth.Account.objects.active_by_openid')
    def test_is_authenticated_inactive_account(self, mock_account_get):
        mock_account_get.return_value = None
        consumer, token, params = Mock(), Mock(), None
        # make sure validate_token returns a consumer and a token
        mock_validate_token = Mock(return_value=(consumer, token, params))
        with patch.object(self.auth, 'validate_token', mock_validate_token):
            self.assertEqual(self.auth.is_authenticated(self.request), False)

    @patch('identityprovider.auth.Account.objects.active_by_openid')
    def test_is_authenticated(self, mock_account_get):
        mock_account = Mock()
        mock_account_get.return_value = mock_account
        mock_token = Mock()
        mock_token.consumer.id = 'foo'
        consumer, params = Mock(), None
        # make sure validate_token returns a consumer and a token
        mock_validate_token = Mock(return_value=(consumer, mock_token, params))
        with patch.object(self.auth, 'validate_token', mock_validate_token):
            response = self.auth.is_authenticated(self.request)

        self.assertEqual(response, True)
        self.assertEqual(self.request.user, mock_account)
        self.assertEqual(self.request.throttle_extra, 'foo')

    def test_challenge(self):
        auth = SSOOAuthAuthentication(realm='foo')
        response = auth.challenge()
        self.assertEqual(response.status_code, 401)
        self.assertEqual(response.content, 'Authorization Required')
        self.assertEqual(response['WWW-Authenticate'], 'OAuth realm="foo"')


class BasicAuthenticateTestCase(SSOBaseTestCase):
    def test_apiuser_authenticate(self):
        apiuser = self.factory.make_apiuser()
        user = basic_authenticate(apiuser.username, DEFAULT_API_PASSWORD)
        self.assertEqual(user, apiuser)

    def test_apiuser_authenticate_with_latin1_username(self):
        username = u'ço'
        apiuser = self.factory.make_apiuser(username=username)
        user = basic_authenticate(username.encode('latin1'),
                                  DEFAULT_API_PASSWORD)
        self.assertEqual(user, apiuser)

    def test_apiuser_authenticate_with_utf8_username(self):
        username = u'ço'
        apiuser = self.factory.make_apiuser(username=username)
        user = basic_authenticate(username.encode('utf-8'),
                                  DEFAULT_API_PASSWORD)
        self.assertEqual(user, apiuser)

    def test_launchpadbackend_authenticate(self):
        account = self.factory.make_account()
        user = basic_authenticate(account.preferredemail.email,
                                  DEFAULT_USER_PASSWORD)
        self.assertEqual(user, account)

    @patch('identityprovider.auth.LaunchpadBackend.authenticate')
    @patch('identityprovider.auth.APIUser.authenticate')
    def test_password_is_decoded(self, mock_apiuser_authenticate,
                                 mock_lp_backend_authenticate):
        # make sure apiuser check fails so lp backend is tried too
        mock_apiuser_authenticate.return_value = None

        username = u'Franço'
        password = u'Franço'

        basic_authenticate(username.encode('latin1'),
                           password.encode('latin1'))

        mock_apiuser_authenticate.assert_called_once_with(
            username, password)
        mock_lp_backend_authenticate.assert_called_once_with(
            username, password)


class CypherLearningTestCase(SSOBaseTestCase):

    def test_launchpad_pass_is_not_usable_by_django(self):
        self.assertFalse(is_password_usable(encrypt_launchpad_password('foo')))

    def test_django_pass_is_usable_by_django(self):
        self.assertTrue(is_password_usable(make_password('foo')))
