from django.core.management import call_command, CommandError
from django.test import TestCase

from identityprovider.models import OpenIDRPConfig
from identityprovider.utils import get_object_or_none


class AddOpenIDRPConfigCommandTestCase(TestCase):

    root = 'http://localhost:8000'

    def call_command(self, **kwargs):
        call_command('add_openid_rp_config', self.root, **kwargs)
        config = get_object_or_none(OpenIDRPConfig, trust_root=self.root)
        self.assertIsNotNone(config)
        return config

    def test_add_without_root(self):
        with self.assertRaises(CommandError) as cm:
            call_command('add_openid_rp_config')
        self.assertEqual(str(cm.exception),
                         'Need to specify trust_root')

    def test_add_default(self):
        config = self.call_command()
        self.assertFalse(config.allow_unverified)
        self.assertEqual(config.allowed_user_attribs, '')

    def test_add_with_allow_unverified(self):
        config = self.call_command(allow_unverified=True)
        self.assertTrue(config.allow_unverified)

    def test_add_with_allowed_user_attribs_value(self):
        config = self.call_command(allowed_user_attribs='nickname,email')
        self.assertFalse(config.allow_unverified)
        self.assertEqual(config.allowed_user_attribs, 'nickname,email')

    def test_auto_authorize_default(self):
        config = self.call_command()
        self.assertFalse(config.auto_authorize)

    def test_auto_authorize_unset(self):
        config = self.call_command(auto_authorize=False)
        self.assertFalse(config.auto_authorize)

    def test_auto_authorize_set(self):
        config = self.call_command(auto_authorize=True)
        self.assertTrue(config.auto_authorize)

    def test_can_query_any_team_default(self):
        config = self.call_command()
        self.assertFalse(config.can_query_any_team)

    def test_can_query_any_team_unset(self):
        config = self.call_command(can_query_any_team=False)
        self.assertFalse(config.can_query_any_team)

    def test_can_query_any_team_set(self):
        config = self.call_command(can_query_any_team=True)
        self.assertTrue(config.can_query_any_team)

    def test_multiple_add_updates(self):
        config = self.call_command()
        self.assertFalse(config.can_query_any_team)
        self.assertEqual(OpenIDRPConfig.objects.count(), 1)
        config = self.call_command(can_query_any_team=True)
        self.assertTrue(config.can_query_any_team)
        self.assertEqual(OpenIDRPConfig.objects.count(), 1)
