# encoding: utf-8
import csv
import random
import string
import tempfile
from datetime import date
from optparse import OptionValueError
from StringIO import StringIO

from django.contrib.auth.hashers import check_password, make_password
from django.core.management import call_command, CommandError
from django.db import DatabaseError, connection
from mock import Mock, patch
from psycopg2 import InternalError

from identityprovider.management.commands.import_leak_data import Command
from identityprovider.models import Account, LeakedCredential
from identityprovider.tests.utils import SSOBaseTestCase
from identityprovider.utils import encrypt_launchpad_password


class ImportLeakDataCommandTestCase(SSOBaseTestCase):
    def setUp(self):
        super(ImportLeakDataCommandTestCase, self).setUp()
        self.command = Command()
        # disable output
        self.command.verbosity = 0
        self.command.stdout = Mock()
        self.command.stderr = Mock()
        self.command.source = 'the source'
        self.command.date_leaked = date.today()

    def make_string(self):
        length = random.randint(1, 16)
        return ''.join(random.choice(string.letters + string.digits)
                       for i in xrange(length))

    def make_email(self):
        return "%s@dom.ain" % self.make_string()

    def make_row(self, flat=False):
        email = self.make_email()
        password = self.make_string()
        row = [email, password]
        if flat:
            return ','.join(row)
        return row

    def make_input_file(self, rows=1, content=None):
        input_file = tempfile.NamedTemporaryFile()
        if content is not None:
            input_file.write(content)
        else:
            for row in xrange(rows):
                input_file.write(self.make_row(flat=True))
        input_file.flush()
        return input_file

    def make_items(self, count=1, source='source', date_leaked=None):
        if date_leaked is None:
            date_leaked = date.today()
        rows = [self.make_row() for i in xrange(count)]
        items = [
            LeakedCredential(email=row[0], password=make_password(row[1]),
                             source=source, when_leaked=date_leaked)
            for row in rows]
        return items

    def test_missing_input_file(self):
        with self.assertRaises(CommandError) as ctx:
            call_command('import_leak_data', verbosity=0)
        self.assertEqual(str(ctx.exception), 'Need to specify an input file.')

    def test_missing_leak_source(self):
        with self.assertRaises(CommandError) as ctx:
            call_command('import_leak_data', 'input.csv', verbosity=0)
        self.assertEqual(str(ctx.exception), 'Need to specify a leak source.')

    def test_too_many_arguments(self):
        with self.assertRaises(CommandError) as ctx:
            call_command('import_leak_data', 'input.csv', 'source',
                         'something', verbosity=0)
        self.assertEqual(str(ctx.exception), 'Too many arguments provided.')

    def test_no_date_leaked(self):
        input_file = self.make_input_file()

        prefix = 'identityprovider.management.commands.import_leak_data'
        with patch(prefix + '.datetime') as mock_datetime:
            today = date.today()
            mock_datetime.utcnow.return_value.date.return_value = today
            self.command.handle(input_file.name, 'source',
                                read_batch_size=1000)
        self.assertEqual(self.command.date_leaked, today)

    def test_invalid_date_leaked(self):
        with self.assertRaises(OptionValueError) as ctx:
            call_command('import_leak_data', 'input.csv', 'source',
                         date_leaked='foo', verbosity=0)
        self.assertEqual(str(ctx.exception),
                         "option --date-leaked: invalid date: 'foo'")

    def test_valid_date_leaked(self):
        input_file = self.make_input_file()

        date_leaked = date(2013, 11, 16)
        self.command.handle(input_file.name, 'source', read_batch_size=1000,
                            date_leaked='2013-11-16')
        self.assertEqual(self.command.date_leaked, date_leaked)

    def test_dry_run_mode(self):
        input_file = self.make_input_file()

        name = ('identityprovider.management.commands.import_leak_data'
                '.Command.make_items')
        with patch(name) as mock_make_items:
            call_command('import_leak_data', input_file.name, 'source',
                         dry_run=True, verbosity=0)
        self.assertFalse(mock_make_items.called)

    def test_no_dry_run_mode(self):
        input_file = self.make_input_file()

        name = ('identityprovider.management.commands.import_leak_data'
                '.Command.make_items')
        with patch(name) as mock_make_items:
            call_command('import_leak_data', input_file.name, 'source',
                         verbosity=0)
        self.assertTrue(mock_make_items.called)

    def test_read_rows_no_data(self):
        input_file = StringIO()
        reader = csv.reader(input_file)

        rows = self.command.read_rows(reader, 10)
        self.assertEqual(rows, [])

    def test_read_rows_partial_data(self):
        input_file = StringIO("""email1,password1
email2,password2
email3,password3
""")
        reader = csv.reader(input_file)

        rows = self.command.read_rows(reader, 10)
        self.assertEqual(rows, [
            ['email1', 'password1'], ['email2', 'password2'],
            ['email3', 'password3'],
        ])

    def test_read_rows_full_data(self):
        input_file = StringIO("""email1,password1
email2,password2
email3,password3
""")
        reader = csv.reader(input_file)

        rows = self.command.read_rows(reader, 2)
        self.assertEqual(rows, [
            ['email1', 'password1'], ['email2', 'password2'],
        ])

    def test_read_rows_unicode_data(self):
        password = u'päŞwöṙd'
        row = u"user@dom.ain,%s" % password
        input_file = self.make_input_file(
            content=row.encode('utf-8'))
        call_command('import_leak_data', input_file.name, 'source',
                     verbosity=0)
        item = LeakedCredential.objects.get(email='user@dom.ain',
                                            source='source')
        self.assertNotEqual(item.password, password)
        self.assertTrue(check_password(password, item.password))

    def test_make_items_clean_data(self):
        rows = [['user1@dom.ain', 'password1'], ['user2@dom.ain', 'password2']]
        items = list(self.command.make_items(rows, None))
        self.assertEqual(len(items), 2)
        for item, row in zip(items, rows):
            self.assertEqual(item.email, row[0])
            self.assertNotEqual(item.password, row[1])
            self.assertTrue(check_password(row[1], item.password))
            self.assertEqual(item.source, 'the source')
            self.assertEqual(item.when_leaked, date.today())

    def test_make_items_noisy_data(self):
        rows = [['user1', 'password1'], ['user2@dom.ain', 'password2'],
                ['user3@dom.ain', '']]
        items = list(self.command.make_items(rows, None))
        self.assertEqual(len(items), 1)
        self.assertEqual(items[0].email, 'user2@dom.ain')
        self.assertTrue(items[0].password, 'password2')
        self.assertEqual(items[0].source, 'the source')
        self.assertEqual(items[0].when_leaked, date.today())

    def test_insert_items_no_batching(self):
        items = self.make_items(count=10)
        prefix = ('identityprovider.management.commands.import_leak_data'
                  '.LeakedCredential.objects')
        with patch(prefix + '.bulk_create') as mock_bulk_create:
            with patch(prefix + '.get_or_create') as mock_get_or_create:
                self.command.insert_items(items, None)
        self.assertFalse(mock_bulk_create.called)
        self.assertEqual(mock_get_or_create.call_count, 10)

    def test_insert_items_in_batches(self):
        items = self.make_items(count=10)
        prefix = ('identityprovider.management.commands.import_leak_data'
                  '.LeakedCredential.objects')
        with patch(prefix + '.bulk_create') as mock_bulk_create:
            with patch(prefix + '.get_or_create') as mock_get_or_create:
                self.command.insert_items(items, 10)
        self.assertEqual(mock_bulk_create.call_count, 1)
        self.assertFalse(mock_get_or_create.called)

    def test_insert_items_in_batches_with_errors(self):
        items = self.make_items(count=10)
        prefix = ('identityprovider.management.commands.import_leak_data'
                  '.LeakedCredential.objects')
        with patch(prefix + '.bulk_create') as mock_bulk_create:
            with patch(prefix + '.get_or_create') as mock_get_or_create:
                mock_bulk_create.side_effect = InternalError
                # now attempt the bulk insert
                self.command.insert_items(items, 5)
        self.assertEqual(mock_bulk_create.call_count, 2)
        self.assertEqual(mock_get_or_create.call_count, 10)

    def test_insert_items_in_batches_with_errors_rollback_single_batch(self):
        real_bulk_create = LeakedCredential.objects.bulk_create

        def bulk_create_with_failure(*args, **kwargs):
            if mock_bulk_create.call_count == 1:
                return real_bulk_create(*args, **kwargs)
            raise InternalError()
        mock_bulk_create = Mock(side_effect=bulk_create_with_failure)

        real_get_or_create = LeakedCredential.objects.get_or_create
        mock_get_or_create = Mock(side_effect=real_get_or_create)

        items = self.make_items(count=10)
        prefix = ('identityprovider.management.commands.import_leak_data'
                  '.LeakedCredential.objects')
        with patch(prefix + '.bulk_create', mock_bulk_create):
            with patch(prefix + '.get_or_create', mock_get_or_create):
                # now attempt the bulk insert
                self.command.insert_items(items, 5)
        self.assertEqual(mock_bulk_create.call_count, 2)
        self.assertEqual(mock_get_or_create.call_count, 5)
        self.assertEqual(LeakedCredential.objects.all().count(), 10)

    def test_insert_items_in_batches_with_error_updating_sequence(self):
        items = self.make_items(count=10)
        name = ('identityprovider.management.commands.import_leak_data'
                '.Command.set_sequence')
        with patch(name) as mock_set_sequence:
            mock_set_sequence.side_effect = DatabaseError
            self.command.insert_items(items, 5)
        self.assertEqual(LeakedCredential.objects.all().count(), 10)
        cursor = connection.cursor()
        cursor.execute(
            "select nextval('identityprovider_leakedcredential_id_seq')")
        next_id = cursor.fetchone()[0]
        self.assertEqual(next_id, 11)

    def test_error_in_insert_fallback(self):
        items = self.make_items(count=5)
        prefix = ('identityprovider.management.commands.import_leak_data'
                  '.LeakedCredential.objects')
        with patch(prefix + '.bulk_create') as mock_bulk_create:
            mock_bulk_create.side_effect = InternalError
            with patch(prefix + '.get_or_create') as mock_get_or_create:
                mock_get_or_create.side_effect = InternalError
                self.command.insert_items(items, 10)
        self.assertEqual(self.command.stderr.write.call_count, 5)
        self.assertEqual(LeakedCredential.objects.all().count(), 0)

    def test_error_in_insert(self):
        items = self.make_items(count=5)
        prefix = ('identityprovider.management.commands.import_leak_data'
                  '.LeakedCredential.objects')
        with patch(prefix + '.get_or_create') as mock_get_or_create:
            mock_get_or_create.side_effect = InternalError
            # insert unbatched to trigger the error
            self.command.insert_items(items, None)
        self.assertEqual(mock_get_or_create.call_count, 5)
        self.assertEqual(self.command.stderr.write.call_count, 5)
        self.assertEqual(LeakedCredential.objects.all().count(), 0)

    def test_make_item_with_matching_account_using_launchpad(self):
        encrypted = encrypt_launchpad_password('foo', salt='salt')
        account = self.factory.make_account(password=encrypted,
                                            password_encrypted=True)
        email = account.preferredemail.email

        name = 'identityprovider.emailutils.send_password_reset_email'
        with patch(name) as mock_send_email:
            item = self.command.make_item([email, 'foo'])

        self.assertEqual(item.email, email)
        self.assertNotEqual(item.password, encrypted)
        self.assertTrue(check_password('foo', item.password))
        self.assertEqual(item.source, 'the source')
        self.assertEqual(item.when_leaked, date.today())

        # refresh account
        account = Account.objects.get(id=account.id)
        self.assertTrue(account.need_password_reset)
        mock_send_email.assert_called_once_with(account, email,
                                                reason='password_leak')

    def test_make_item_with_matching_account_using_pbkdf2(self):
        encrypted = make_password('foo', salt='salt', hasher='pbkdf2_sha256')
        account = self.factory.make_account(password=encrypted,
                                            password_encrypted=True)
        email = account.preferredemail.email

        name = 'identityprovider.emailutils.send_password_reset_email'
        with patch(name) as mock_send_email:
            item = self.command.make_item([email, 'foo'])

        self.assertEqual(item.email, email)
        self.assertNotEqual(item.password, encrypted)
        self.assertTrue(check_password('foo', item.password))
        self.assertEqual(item.source, 'the source')
        self.assertEqual(item.when_leaked, date.today())

        # refresh account
        account = Account.objects.get(id=account.id)
        self.assertTrue(account.need_password_reset)
        mock_send_email.assert_called_once_with(account, email,
                                                reason='password_leak')

    def test_make_item_without_matching_account(self):
        row = self.make_row()

        item = self.command.make_item(row)

        self.assertEqual(item.email, row[0])
        self.assertTrue(check_password(row[1], item.password))
        self.assertEqual(item.source, 'the source')
        self.assertEqual(item.when_leaked, date.today())

    def test_make_item_invalid_email(self):
        item = self.command.make_item(['foo', 'foo'])
        self.assertIsNone(item)

    def test_make_item_no_password(self):
        item = self.command.make_item(['user@dom.ain', ''])
        self.assertIsNone(item)
