# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from django import forms
from django.db.models.query import EmptyQuerySet
from django.test import TestCase

from identityprovider.fields import (
    CommaSeparatedField,
    DisplayNameField,
    EmailField,
    OATHPasswordField,
    PasswordField,
    PreferredEmailField,
    default_errors,
    default_email_placeholder,
    default_password_placeholder,
    email_errors,
)
from identityprovider.widgets import ROAwareSelect, ROAwareTextInput
from identityprovider.tests.utils import SSOBaseTestCase


class CommaSeparatedFieldTestCase(TestCase):

    def test_clean(self):
        choices = (('1', 'One'), ('2', 'Two'), ('3', 'Three'))
        field = CommaSeparatedField(choices=choices)
        value = field.clean(["1", "2", "3"])
        self.assertEqual(value, "1,2,3")


class OATHPasswordFieldTestCase(TestCase):

    def do_clean(self, value):
        field = OATHPasswordField()
        return field.clean(value)

    def test_too_short_not_accepted(self):
        self.assertRaises(forms.ValidationError, self.do_clean, '1')
        self.assertRaises(forms.ValidationError, self.do_clean, '12')
        self.assertRaises(forms.ValidationError, self.do_clean, '123')
        self.assertRaises(forms.ValidationError, self.do_clean, '1234')
        self.assertRaises(forms.ValidationError, self.do_clean, '12345')

    def test_too_long_not_accepted(self):
        self.assertRaises(forms.ValidationError, self.do_clean, '123456789')

    def test_7_digits_not_accepted(self):
        self.assertRaises(forms.ValidationError, self.do_clean, '1234567')

    def test_valid_lengths_accepted(self):
        self.assertEqual(self.do_clean('123456'), '123456')
        self.assertEqual(self.do_clean('12345678'), '12345678')

    def test_non_digits_not_accepted(self):
        self.assertRaises(forms.ValidationError, self.do_clean, '12345X')
        self.assertRaises(forms.ValidationError, self.do_clean, '1234567X')

    def test_whitespace_accepted(self):
        self.assertEqual(self.do_clean(' 123456'), '123456')
        self.assertEqual(self.do_clean('123456 '), '123456')
        self.assertEqual(self.do_clean('123 456'), '123456')


class PreferredEmailFieldTestCase(SSOBaseTestCase):

    error_messages = email_errors
    inherits_from = forms.ModelChoiceField
    widget = ROAwareSelect

    def setUp(self):
        super(PreferredEmailFieldTestCase, self).setUp()
        self.field = self.create_field()

    def create_field(self, **kwargs):
        return PreferredEmailField(queryset=EmptyQuerySet(), **kwargs)

    def test_parent_field(self):
        self.assertIsInstance(self.field, self.inherits_from)

    def test_error_messages(self):
        expected = set(self.error_messages.iteritems())
        actual = set(self.field.error_messages.iteritems())
        self.assertTrue(expected.issubset(actual))

    def test_widget(self):
        self.assertIsInstance(self.field.widget, self.widget)


class EmailFieldTestCase(PreferredEmailFieldTestCase):

    error_messages = email_errors
    inherits_from = forms.EmailField
    input_type = 'email'
    widget = forms.TextInput
    expected_attrs = {'class': 'textType'}
    ubuntuone_attrs = {
        'autofocus': 'autofocus', 'placeholder': default_email_placeholder}

    def create_field(self, **kwargs):
        return EmailField(**kwargs)

    def test_widget_input_type(self):
        self.assertEqual(self.field.widget.input_type, self.input_type)

    def test_placeholder_can_be_overriden(self):
        self.set_brand('ubuntuone')
        new = 'Foo Bar Baz'
        self.field = self.create_field(placeholder=new)
        self.assertEqual(self.field.widget.attrs['placeholder'], new)

    def assert_widget_attrs_per_brand(self, brand, extra=None):
        expected_attrs = self.expected_attrs.copy()
        if brand == 'ubuntuone':
            expected_attrs.update(self.ubuntuone_attrs)

        self.set_brand(brand)
        self.field = self.create_field()
        self.assertEqual(self.field.widget.attrs, expected_attrs)

    def test_widget_attrs_brand_launchpad(self):
        self.assert_widget_attrs_per_brand('launchpad')

    def test_widget_attrs_brand_ubuntuone(self):
        self.assert_widget_attrs_per_brand('ubuntuone')

    def test_widget_attrs_change_brands(self):
        self.assert_widget_attrs_per_brand('launchpad')
        self.assert_widget_attrs_per_brand('ubuntuone')
        self.assert_widget_attrs_per_brand('launchpad')


class PasswordFieldTestCase(EmailFieldTestCase):

    error_messages = default_errors
    inherits_from = forms.CharField
    input_type = 'password'
    widget = forms.PasswordInput
    expected_attrs = {'class': 'disableAutoComplete textType'}
    ubuntuone_attrs = {'placeholder': default_password_placeholder}

    def create_field(self, **kwargs):
        return PasswordField(**kwargs)


class DisplayNameFieldTestCase(EmailFieldTestCase):

    error_messages = default_errors
    inherits_from = forms.CharField
    input_type = 'text'
    widget = ROAwareTextInput
    expected_attrs = {'class': 'textType'}
    ubuntuone_attrs = {'placeholder': 'Your name'}

    def create_field(self, **kwargs):
        return DisplayNameField(**kwargs)
