# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from datetime import datetime, timedelta

from django.conf import settings
from django.contrib.auth.hashers import get_hasher
from django.contrib.auth.models import User, check_password, make_password
from django.core.exceptions import ValidationError

from mock import Mock, call, patch
from oauth_backend.models import Consumer


from identityprovider.login import (
    AuthenticationError,
    AccountDeactivated,
    AccountSuspended,
    PasswordPolicyError,
    authenticate_user,
)
from identityprovider.models.account import (
    Account,
    AccountPassword,
    AccountQuerySet,
)
from identityprovider.models import account as a
from identityprovider.models.emailaddress import EmailAddress
from identityprovider.models.const import (
    AccountStatus,
    AccountCreationRationale,
    EmailStatus,
)
from identityprovider.readonly import ReadOnlyManager
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import SSOBaseTestCase
from identityprovider.validators import PASSWORD_POLICY_ERROR


class MockBackend(object):

    def __init__(self, value):
        self.value = value

    def has_module_perms(self, account, app_label):
        return self.value

    def has_perm(self, account, perm):
        return self.value


class AccountQuerySetTestCase(SSOBaseTestCase):

    def setUp(self):
        super(AccountQuerySetTestCase, self).setUp()
        self.qs = AccountQuerySet(model=Account)

    def test_verified_list_only_verified_accounts(self):
        accounts = self.qs.verified()
        for account in accounts:
            emails = account.emailaddress_set.all()
            self.assertTrue(any([email.is_verified for email in emails]))


class AccountManagerTestCase(SSOBaseTestCase):

    def test_objects_get_by_email_fails(self):
        account = Account.objects.get_by_email('somethin@which.not.exists.com')

        self.assertTrue(account is None)

    def test_create_account_with_invalid_password(self):
        salt = get_hasher('default').salt()
        kwargs = dict(
            displayname='displayname', email_address='email@host.local',
            password='test', salt=salt,
            creation_rationale=AccountCreationRationale.USER_CREATED,
        )
        self.assertRaises(ValidationError, Account.objects.create_account,
                          **kwargs)

    def test_create_account_with_invalid_email_address(self):
        salt = get_hasher('default').salt()
        kwargs = dict(
            displayname='displayname', email_address='not_an_email',
            password=DEFAULT_USER_PASSWORD, salt=salt,
            creation_rationale=AccountCreationRationale.USER_CREATED,
        )
        self.assertRaises(ValidationError, Account.objects.create_account,
                          **kwargs)

    def test_create_account_with_rationale(self):
        salt = get_hasher('default').salt()
        account = Account.objects.create_account(
            displayname='displayname', email_address='email@host.local',
            password=DEFAULT_USER_PASSWORD,
            creation_rationale=AccountCreationRationale.USER_CREATED,
            salt=salt,
        )

        self.assertEqual(account.displayname, 'displayname')
        self.assertEqual(account.creation_rationale,
                         AccountCreationRationale.USER_CREATED)
        self.assertEqual(account.status, AccountStatus.ACTIVE)

        emails = EmailAddress.objects.filter(account=account)
        self.assertEqual(emails.count(), 1)
        email = emails[0]
        self.assertEqual(email.email, 'email@host.local')
        self.assertEqual(email.status, EmailStatus.PREFERRED)

        passwords = AccountPassword.objects.filter(account=account)
        self.assertEqual(passwords.count(), 1)
        password = passwords[0]
        self.assertEqual(password.password,
                         make_password(DEFAULT_USER_PASSWORD, salt=salt))

    def test_create_account_with_no_rationale(self):
        salt = get_hasher('default').salt()
        account = Account.objects.create_account(
            displayname='displayname', email_address='email@host.local',
            password=DEFAULT_USER_PASSWORD, salt=salt)

        self.assertEqual(account.displayname, 'displayname')
        self.assertEqual(account.creation_rationale,
                         AccountCreationRationale.OWNER_CREATED_LAUNCHPAD)
        self.assertEqual(account.status, AccountStatus.ACTIVE)

        emails = EmailAddress.objects.filter(account=account)
        self.assertEqual(emails.count(), 1)
        email = emails[0]
        self.assertEqual(email.email, 'email@host.local')
        self.assertEqual(email.status, EmailStatus.PREFERRED)

        passwords = AccountPassword.objects.filter(account=account)
        self.assertEqual(passwords.count(), 1)
        password = passwords[0]
        self.assertEqual(password.password,
                         make_password(DEFAULT_USER_PASSWORD, salt=salt))

    def test_create_account_with_email_validated(self):
        salt = get_hasher('default').salt()
        account = Account.objects.create_account(
            displayname='displayname', email_address='email@host.local',
            password=DEFAULT_USER_PASSWORD, salt=salt, email_validated=False)

        self.assertEqual(account.displayname, 'displayname')
        self.assertEqual(account.creation_rationale,
                         AccountCreationRationale.OWNER_CREATED_LAUNCHPAD)
        self.assertEqual(account.status, AccountStatus.ACTIVE)

        emails = EmailAddress.objects.filter(account=account)
        self.assertEqual(emails.count(), 1)
        email = emails[0]
        self.assertEqual(email.email, 'email@host.local')
        self.assertEqual(email.status, EmailStatus.NEW)

        passwords = AccountPassword.objects.filter(account=account)
        self.assertEqual(passwords.count(), 1)
        password = passwords[0]
        self.assertEqual(password.password,
                         make_password(DEFAULT_USER_PASSWORD, salt=salt))

    def test_create_account_no_password(self):
        with self.assertRaises(ValidationError) as ctx:
            Account.objects.create_account(
                displayname='John Doe', email_address='email@host.local',
                password=None)

        e = ctx.exception
        self.assertEqual(e.message_dict, {'password': [PASSWORD_POLICY_ERROR]})

    def test_create_account_no_email(self):
        with self.assertRaises(ValidationError) as ctx:
            Account.objects.create_account(
                displayname='John Doe', email_address=None,
                password=DEFAULT_USER_PASSWORD)

        e = ctx.exception
        self.assertEqual(
            e.message_dict, {'email': ['This field cannot be null.']})

    def test_create_account_no_displayname(self):
        with self.assertRaises(ValidationError) as ctx:
            Account.objects.create_account(
                displayname=None, email_address='email@host.local',
                password=DEFAULT_USER_PASSWORD)

        e = ctx.exception
        self.assertEqual(e.message_dict, {'displayname': ['Field required']})

    def test_create_account_displayname_empty(self):
        with self.assertRaises(ValidationError) as ctx:
            Account.objects.create_account(
                displayname='', email_address='email@host.local',
                password=DEFAULT_USER_PASSWORD)

        e = ctx.exception
        self.assertEqual(e.message_dict, {'displayname': ['Field required']})

    def test_active_by_openid_no_accounts(self):
        result = Account.objects.active_by_openid('123456')
        self.assertIsNone(result)

    def test_active_by_openid_no_match(self):
        account = self.factory.make_account()
        assert account.openid_identifier != '123456'
        result = Account.objects.active_by_openid('123456')
        self.assertIsNone(result)

    def test_active_by_openid(self):
        account = self.factory.make_account()
        openid = account.openid_identifier
        for status, name in AccountStatus._get_choices():
            account.status = status
            account.save()
            result = Account.objects.active_by_openid(openid)
            if status == AccountStatus.ACTIVE:
                assert account.is_active
                self.assertEqual(result, account)
            else:
                self.assertIsNone(result)

    def test_verified_list_only_verified_accounts(self):
        accounts = Account.objects.verified()
        for account in accounts:
            emails = account.emailaddress_set.all()
            self.assertTrue(any([email.is_verified for email in emails]))

    def test_verified_no_duplicates(self):
        accounts = Account.objects.verified()
        self.assertEqual(accounts.count(), accounts.distinct().count())


class AccountTestCase(SSOBaseTestCase):

    def prepare_perms(self, has_perm):
        account = self.factory.make_account()

        def mock_get_backends():
            return [MockBackend(has_perm)]

        get_backends = a.get_backends
        a.get_backends = mock_get_backends
        return account, get_backends

    def test_last_login_when_user_doesnt_exists(self):
        account = self.factory.make_account()
        self.assertTrue(account.last_login is not None)

    def test_last_login_when_user_exists(self):
        account = self.factory.make_account()
        User.objects.get_or_create(username=account.openid_identifier)

        self.assertTrue(account.last_login is not None)

    def test_set_last_login_when_user_does_not_exists(self):
        account = self.factory.make_account()

        # we use now() here because last_login maps to django's auth model
        # which uses localtime.
        account.last_login = datetime.now()

        self.assertTrue(account.last_login is not None)

    def test_set_last_login_when_readonly(self):
        readonly_manager = ReadOnlyManager()
        account = self.factory.make_account()
        account.last_login = last_login = datetime(2010, 01, 01)
        self.assertEqual(account.last_login, last_login)

        assert not settings.READ_ONLY_MODE
        readonly_manager.set_readonly()
        self.addCleanup(readonly_manager.clear_readonly)

        # we use now() here because last_login maps to django's auth model
        # which uses localtime.
        account.last_login = now = datetime.now()
        self.assertEqual(account.last_login, last_login)
        self.assertNotEqual(account.last_login, now)

    def test_set_last_login_when_no_preferredemail(self):
        account = self.factory.make_account()
        account.emailaddress_set.all().update(status=EmailStatus.NEW)
        # we use now() here because last_login maps to django's auth model
        # which uses localtime.
        account.last_login = datetime.now()
        # Changed from assertRaises, as using it to test for database
        # exceptions causes transaction to be aborted and all tests following
        # it will fail. (It's only the case if the test case inherits
        # django.test.TestCase)
        # Test is has actually created a User object
        self.assertEqual(
            1, User.objects.filter(username=account.openid_identifier).count())

    def test_is_active(self):
        account = self.factory.make_account()

        self.assertTrue(account.is_active)

    def test_is_staff(self):
        account = self.factory.make_account()

        self.assertFalse(account.is_staff)

    def test_is_authenticated(self):
        account = self.factory.make_account()

        self.assertTrue(account.is_authenticated)

    def test_is_superuser(self):
        account = self.factory.make_account()

        self.assertFalse(account.is_superuser)

    def test_is_not_verified(self):
        account = self.factory.make_account(email_validated=False)
        self.assertFalse(account.is_verified)

    def test_is_verified(self):
        account = self.factory.make_account()
        self.assertTrue(account.is_verified)

    def test_first_name(self):
        account = self.factory.make_account(displayname='Sample Person')
        self.assertEqual(account.first_name, "Sample")

    def test_full_name(self):
        account = self.factory.make_account(displayname='Sample Person')
        self.assertEqual(account.get_full_name(), "Sample Person")

    def test_twofactor_attempts_default(self):
        account = self.factory.make_account()
        self.assertEqual(account.twofactor_attempts, 0)

    def test_increment_twofactor_attempts(self):
        account = self.factory.make_account()
        assert account.twofactor_attempts == 0

        account.increment_twofactor_attempts()

        # reload account
        account = Account.objects.get(id=account.id)
        self.assertEqual(account.twofactor_attempts, 1)

    def test_reset_twofactor_attempts(self):
        account = self.factory.make_account()
        assert account.twofactor_attempts == 0

        account.increment_twofactor_attempts()
        account.increment_twofactor_attempts()
        account.increment_twofactor_attempts()
        account.increment_twofactor_attempts()
        assert account.twofactor_attempts == 4

        account.reset_twofactor_attempts()

        # reload account
        account = Account.objects.get(id=account.id)
        self.assertEqual(account.twofactor_attempts, 0)

    def test_has_module_perms_when_not_active(self):
        account = self.factory.make_account()
        account.status = AccountStatus.NOACCOUNT

        self.assertFalse(account.has_module_perms('app'))

    def test_has_module_perms_returning_true(self):
        account, get_backends = self.prepare_perms(True)
        try:
            self.assertTrue(account.has_module_perms('app'))
        finally:
            a.get_backends = get_backends

    def test_has_module_perms_returning_false(self):
        account, get_backends = self.prepare_perms(False)
        try:
            self.assertFalse(account.has_module_perms('app'))
        finally:
            a.get_backends = get_backends

    def test_has_perm_returning_true(self):
        account, get_backends = self.prepare_perms(True)
        try:
            self.assertTrue(account.has_perm('can_change_account'))
        finally:
            a.get_backends = get_backends

    def test_has_perm_returning_false(self):
        account, get_backends = self.prepare_perms(False)
        try:
            self.assertFalse(account.has_perm('can_change_account'))
        finally:
            a.get_backends = get_backends

    def test_has_perm_when_account_is_not_active(self):
        account = self.factory.make_account()
        account.status = AccountStatus.NOACCOUNT

        self.assertFalse(account.has_perm('can_change_account'))

    def test_new_account_has_preferred_email_with_flag(self):
        account = self.factory.make_account(email_validated=False)
        self.assertIsNotNone(account.preferredemail)

    def test_set_preferred_with_unverified_email(self):
        account = self.factory.make_account(email_validated=False)
        with self.assertRaises(ValidationError):
            account.preferredemail = account.emailaddress_set.get()

    def test_can_reset_password_without_verified_email(self):
        account = self.factory.make_account(email_validated=False)
        self.assertTrue(account.can_reset_password)

    def test_set_password_validates(self):
        account = self.factory.make_account()
        new_password = 'some NEW password 123'
        original_password = account.accountpassword.password

        VALIDATE = 'identityprovider.models.account.validate_password_policy'
        with patch(VALIDATE) as mock_validate:
            account.set_password(new_password)

        mock_validate.assert_called_once_with(new_password, account)
        self.assertNotEqual(account.accountpassword.password,
                            original_password)

    def test_password_change_deletes_tokns(self):
        account = self.factory.make_account()
        self.factory.make_oauth_token(account=account)

        account.set_password('fooBar123')
        self.assertEqual(account.oauth_tokens().count(), 0)

    def test_sets_password_with_django_cypher(self):
        password = 'Some password from 2013'
        account = self.factory.make_account()

        account.set_password(password)

        self.assertTrue(check_password(password, account.encrypted_password))

    def test_set_password_validation_failure(self):
        account = self.factory.make_account()
        original_password = account.accountpassword.password
        with self.assertRaises(ValidationError):
            account.set_password('short')

        account = Account.objects.get(id=account.id)
        self.assertEqual(account.accountpassword.password, original_password)

    def test_set_password_resets_password_reset_flag(self):
        account = self.factory.make_account()
        account.need_password_reset = True
        account.save()

        account.set_password(DEFAULT_USER_PASSWORD)

        self.assertFalse(account.need_password_reset)

    def test_validated_email_address_is_preferred_email_by_default(self):
        account = self.factory.make_account(email_validated=False)
        email_address = account.emailaddress_set.create(
            email='foobar@canonical.com',
            status=EmailStatus.VALIDATED)
        self.assertEqual(account.preferredemail.email, 'foobar@canonical.com')

        email_address = EmailAddress.objects.get(pk=email_address.id)
        self.assertEqual(email_address.status, EmailStatus.PREFERRED)

    def test_account_suspend_reset_password(self):
        account = self.factory.make_account(password=DEFAULT_USER_PASSWORD,
                                            status=AccountStatus.SUSPENDED)

        account_password = AccountPassword.objects.get(account=account)
        self.assertEqual(account_password.password, 'invalid')

    def test_save_when_readonly(self):
        readonly_manager = ReadOnlyManager()
        account = self.factory.make_account()
        assert account.status == AccountStatus.ACTIVE

        assert not settings.READ_ONLY_MODE
        readonly_manager.set_readonly()
        self.addCleanup(readonly_manager.clear_readonly)

        account.suspend()
        # refresh account from db
        account = Account.objects.get(id=account.id)
        self.assertEqual(account.status, AccountStatus.ACTIVE)

    def test_save_when_no_password(self):
        account = self.factory.make_account()
        # remove all passwords
        AccountPassword.objects.filter(account=account).delete()

        # reload account, to bypass django orm cache
        account = Account.objects.get(pk=account.pk)
        # trigger test
        account.status = AccountStatus.SUSPENDED
        account.save()

        self.assertEqual(
            AccountPassword.objects.filter(account=account).count(), 0)

    def test_sites_with_active_sessions_when_returns_empty_result(self):
        account = self.factory.make_account()
        sites_count = account.sites_with_active_sessions().count()
        self.assertEqual(sites_count, 0)

    def test_sites_with_active_sessions_when_one_site_is_active(self):
        account = self.factory.make_account()

        account.openidrpsummary_set.create(
            date_last_used=datetime.utcnow(),
            trust_root='http://trust-root.example.com',
            openid_identifier='openid-identifier')

        account.openidrpsummary_set.create(
            date_last_used=datetime.utcnow() - timedelta(days=3),
            trust_root='http://trust-root-2.example.com')

        sites_count = account.sites_with_active_sessions().count()
        self.assertEqual(sites_count, 1)

    @patch('identityprovider.models.account.datetime')
    def test_sites_with_active_sessions_use_utc(self, mock_datetime):
        mock_datetime.utcnow.return_value = datetime.utcnow()
        mock_datetime.side_effect = lambda *args, **kw: datetime(*args, **kw)

        account = self.factory.make_account()
        sites_count = account.sites_with_active_sessions().count()

        self.assertEqual(sites_count, 0)
        self.assertEqual(mock_datetime.utcnow.called, True)

    def test_warn_about_backup_device(self):
        account = self.factory.make_account()
        # use assertEqual instead to assertTrue to ensure attr is a boolean
        self.assertEqual(account.warn_about_backup_device, True)

    def test_need_backup_device_warning_no_devices(self):
        account = self.factory.make_account()
        assert account.warn_about_backup_device
        assert account.devices.count() == 0
        self.assertFalse(account.need_backup_device_warning)

    def test_need_backup_device_warning_one_device(self):
        account = self.factory.make_account()
        self.factory.make_device(account=account)

        assert account.warn_about_backup_device
        assert account.devices.count() == 1
        self.assertTrue(account.need_backup_device_warning)

    def test_need_backup_device_warning_one_device_warn_setting_false(self):
        account = self.factory.make_account()
        account.warn_about_backup_device = False
        account.save()
        self.factory.make_device(account=account)

        assert not account.warn_about_backup_device
        assert account.devices.count() == 1
        self.assertFalse(account.need_backup_device_warning)

    def test_need_backup_device_warning_two_devices(self):
        account = self.factory.make_account()
        self.factory.make_device(account=account)
        self.factory.make_device(account=account)

        assert account.warn_about_backup_device
        assert account.devices.count() == 2
        self.assertFalse(account.need_backup_device_warning)

    def test_suspend(self):
        account = self.factory.make_account()
        assert account.status != AccountStatus.SUSPENDED

        account.suspend()
        self.assertEqual(account.status, AccountStatus.SUSPENDED)

        # the status change was also saved in the DB
        account = Account.objects.get(id=account.id)
        self.assertEqual(account.status, AccountStatus.SUSPENDED)

    def test_set_deleted_no_user(self):
        account = self.factory.make_account()
        assert User.objects.filter(
            username=account.openid_identifier).count() == 0

        account.set_deleted()
        self.assertEqual(account.status, AccountStatus.DELETED)

        account = Account.objects.get(id=account.id)
        self.assertEqual(account.accountpassword.password, 'invalid')
        self.assertFalse(account.is_active)
        self.assertEqual(account.emails().count(), 0)

        user = User.objects.get(username=account.openid_identifier)
        self.assertFalse(user.is_active)

    def test_set_deleted_clears_tokens(self):
        account = self.factory.make_account()

        # creating oauth token also creates the User for the account
        account.create_oauth_token('foo')
        account.create_oauth_token('bar')

        account.set_deleted()

        user = User.objects.get(username=account.openid_identifier)
        self.assertFalse(user.is_active)
        self.assertEqual(user.oauth_consumer.token_set.count(), 0)

    def test_reset_password_default_email(self):
        account = self.factory.make_account()

        name = 'identityprovider.emailutils.send_password_reset_email'
        with patch(name) as mock_send_email:
            requested = account.reset_password()

        self.assertTrue(requested)
        mock_send_email.assert_called_once_With(
            account, account.preferredemail.email, reason='password_leak')
        account = Account.objects.get(id=account.id)
        self.assertTrue(account.need_password_reset)

    def test_reset_password_default_reason(self):
        account = self.factory.make_account()
        email = account.preferredemail.email

        name = 'identityprovider.emailutils.send_password_reset_email'
        with patch(name) as mock_send_email:
            requested = account.reset_password(email)

        self.assertTrue(requested)
        mock_send_email.assert_called_once_with(
            account, email, reason='password_leak')
        account = Account.objects.get(id=account.id)
        self.assertTrue(account.need_password_reset)

    def test_reset_password_custom_reason(self):
        account = self.factory.make_account()
        email = account.preferredemail.email

        name = 'identityprovider.emailutils.send_password_reset_email'
        with patch(name) as mock_send_email:
            requested = account.reset_password(email, reason='forgot_password')

        self.assertTrue(requested)
        mock_send_email.assert_called_once_with(
            account, email, reason='forgot_password')
        account = Account.objects.get(id=account.id)
        self.assertTrue(account.need_password_reset)

    def test_reset_password_with_invalid_email(self):
        account = self.factory.make_account()

        requested = account.reset_password('invalid@foo.com')
        self.assertFalse(requested)
        account = Account.objects.get(id=account.id)
        self.assertFalse(account.need_password_reset)


class AccountEmailsTestCase(SSOBaseTestCase):

    def setUp(self):
        super(AccountEmailsTestCase, self).setUp()
        self.account = self.factory.make_account()

    def test_limit_emails(self):
        assert self.account.emails().count() == 1

        expected = [self.account.preferredemail]
        for i in xrange(10):
            date_created = datetime.utcnow() + timedelta(seconds=i)
            email = EmailAddress.objects.create(
                account=self.account, email='emails-test-%s@example.com' % i,
                status=EmailStatus.NEW if i % 2 else EmailStatus.VALIDATED,
                date_created=date_created)
            expected.append(email)

        expected.reverse()

        # no limit should return all
        emails = self.account.emails()
        self.assertEqual(emails.count(), 11)
        self.assertEqual(list(emails), expected)

        # limit should return the latest 5
        emails = self.account.emails(limit=5)
        self.assertEqual(emails.count(), 5)
        self.assertEqual(list(emails), expected[:5])

    def test_verified_emails_has_preferred_first(self):
        account = self.factory.make_account()
        self.factory.make_email_for_account(account,
                                            status=EmailStatus.VALIDATED)
        emails = account.verified_emails()
        self.assertTrue(len(emails) > 1)
        self.assertEqual(emails[0].status, EmailStatus.PREFERRED)
        self.assertEqual(emails[1].status, EmailStatus.VALIDATED)


class AccountPasswordTestCase(SSOBaseTestCase):

    def test_unicode(self):
        account = Account(openid_identifier='oid', displayname='displayname')
        password = AccountPassword(account=account, password='password')
        self.assertEqual(unicode(password), u'Password for displayname')


class CreateOAuthTokenForAccountTestCase(SSOBaseTestCase):

    def setUp(self):
        super(CreateOAuthTokenForAccountTestCase, self).setUp()
        self.account = self.factory.make_account(
            displayname='test', email='test-oauth@example.com',
            password=DEFAULT_USER_PASSWORD)
        self.user = User.objects.create_user(
            self.account.openid_identifier,
            'test@example.com', 'password')

    def test_when_account_has_associated_consumer(self):
        consumer, _ = Consumer.objects.get_or_create(user=self.user)

        token = self.account.create_oauth_token('new-token')

        self.assertEqual(token.consumer.id, consumer.id)

    def test_get_or_create(self):
        consumer, _ = Consumer.objects.get_or_create(user=self.user)

        token, created = self.account.get_or_create_oauth_token('new-token')

        self.assertTrue(created)
        self.assertEqual(token.consumer.id, consumer.id)

        token2, created = self.account.get_or_create_oauth_token('new-token')
        self.assertFalse(created)
        self.assertEqual(token2, token)
        self.assertNotEqual(token2.updated_at, token.updated_at)

        token3, created = self.account.get_or_create_oauth_token('new-token2')
        self.assertTrue(created)
        self.assertNotEqual(token, token3)
        self.assertEqual(token3.consumer.id, consumer.id)

    def test_get_or_create_when_multiple_results(self):
        consumer, _ = Consumer.objects.get_or_create(user=self.user)

        # create multiple tokens
        token1 = consumer.token_set.create(name='new-token')
        token2 = consumer.token_set.create(name='new-token')
        token2.created_at = token1.created_at + timedelta(seconds=10)
        token2.save()

        token, created = self.account.get_or_create_oauth_token('new-token')

        self.assertEqual(token, token2)
        self.assertFalse(created)

    def test_get_when_account_has_no_associated_consumer(self):
        Consumer.objects.filter(user=self.user).delete()

        token = self.account.create_oauth_token('new-token')

        self.assertEqual(token.consumer.user.id, self.user.id)

    def test_get_or_create_when_account_has_no_associated_consumer(self):
        Consumer.objects.filter(user=self.user).delete()

        token, created = self.account.get_or_create_oauth_token('new-token')

        self.assertTrue(created)
        self.assertEqual(token.consumer.user.id, self.user.id)

    def test_invalidate_oauth_tokens(self):
        token, created = self.account.get_or_create_oauth_token('new-token')
        assert self.account.oauth_tokens().count() > 0

        self.account.invalidate_oauth_tokens()
        self.assertEqual(self.account.oauth_tokens().count(), 0)

    def test_limit_oauth_tokens(self):
        assert self.account.oauth_tokens().count() == 0

        expected = []
        for i in xrange(10):
            token = self.account.create_oauth_token('token-%s' % i)
            token.updated_at = token.created_at + timedelta(seconds=i)
            token.save()
            expected.append(token)

        expected.reverse()

        # no limit should return all
        tokens = self.account.oauth_tokens()
        self.assertEqual(tokens.count(), 10)
        self.assertEqual(list(tokens), expected)

        # limit should return the latest 5
        tokens = self.account.oauth_tokens(limit=5)
        self.assertEqual(tokens.count(), 5)
        self.assertEqual(list(tokens), expected[:5])

    def test_oauth_token_by_key(self):
        tokens = {}
        for i in xrange(5):
            token = self.account.create_oauth_token('token-%s' % i)
            tokens[token.token] = token

        for key, token in tokens.iteritems():
            self.assertEqual(self.account.oauth_token_by_key(key), token)

    def test_oauth_token_by_key_no_token(self):
        assert self.account.oauth_tokens().count() == 0
        self.assertIsNone(self.account.oauth_token_by_key('key'))


class AuthenticateUserTestCase(SSOBaseTestCase):

    def setUp(self):
        super(AuthenticateUserTestCase, self).setUp()
        self.email = 'test@example.com'
        self.password = DEFAULT_USER_PASSWORD
        self.account = self.make_account()
        self.user = User.objects.create_user(
            self.account.openid_identifier,
            self.email, self.password)

    def make_account(self):
        account = self.factory.make_account(email=self.email,
                                            password=self.password)
        return account

    def assert_failed_login(self, exception):
        with self.assertRaises(exception):
            authenticate_user(self.email, self.password)

    def test_account_suspended(self):
        self.account.status = AccountStatus.SUSPENDED
        self.account.save()
        self.assert_failed_login(AccountSuspended)

    def test_account_deactivated(self):
        self.account.status = AccountStatus.DEACTIVATED
        self.account.save()
        self.assert_failed_login(AccountDeactivated)

    def test_failed_login(self):
        self.password = 'wrong password'
        self.assert_failed_login(AuthenticationError)

    def test_new_email_allowed(self):
        email = self.account.emailaddress_set.get(email=self.email)
        email.status = EmailStatus.NEW
        email.save()

        result = authenticate_user(self.email, self.password)

        self.assertEqual(result, self.account)

    def test_authenticate_user(self):
        result = authenticate_user(self.email, self.password)
        self.assertEqual(result, self.account)

    def test_invalidated_email(self):
        email_obj = self.account.emailaddress_set.get(email=self.email)
        email_obj.invalidate()

        # this is not distinguishable if another user registered with
        # the same email from a password mismatch
        with self.assertRaises(AuthenticationError):
            authenticate_user(self.email, self.password)

    def test_login_by_phone_enabled(self):
        email = self.account.emailaddress_set.get(email=self.email)
        email.status = EmailStatus.NEW
        email.save()

        # enable the LOGIN_BY_PHONE flag selective on team membership
        team_name = 'tomb-riders'
        key = 'LOGIN_BY_PHONE'
        field_name = 'team'
        value = team_name
        condition_set = ('identityprovider.gargoyle.LPTeamConditionSet('
                         'lp_team)')
        self.conditionally_enable_flag(key, field_name, value, condition_set)

        # bind account with the team
        team = self.factory.make_team(name=team_name)
        self.factory.add_account_to_team(self.account, team)

        result = authenticate_user(self.email, self.password)
        self.assertEqual(result, self.account)

    def test_password_policy_error(self):
        unicode_error_message = u'FAIL\xac'
        validator = Mock()
        validator.match.return_value = True
        validator.validate.side_effect = ValidationError(unicode_error_message)

        VALIDATORS = 'identityprovider.validators._validators'
        with patch(VALIDATORS, [validator]):
            with self.assertRaises(PasswordPolicyError) as ctx:
                authenticate_user(self.email, self.password)
        exception = ctx.exception

        self.assertEqual(unicode(exception), unicode_error_message)
        self.assertEqual(exception.account, self.account)

        expected_calls = [
            call.match(self.account), call.validate(self.password)]
        self.assertEqual(validator.mock_calls, expected_calls)

    def test_account_need_password_reset(self):
        self.account.need_password_reset = True
        self.account.save()

        with self.assertRaises(PasswordPolicyError) as ctx:
            authenticate_user(self.email, self.password)

        self.assertIn('there was a security incident', str(ctx.exception))
        self.assertEqual(ctx.exception.account, self.account)
