from datetime import datetime, timedelta

from django.conf import settings
from django.test import TestCase
from django.test.client import RequestFactory
from mock import Mock, patch
from oath import hotp

from identityprovider.login import AuthenticationError
from identityprovider.models.twofactor import (
    TWOFACTOR_LOGIN,
    AuthenticationDevice,
    is_authenticated,
    is_fresh,
    is_twofactor_enabled,
    is_upgraded,
    login,
    logout,
    site_requires_twofactor_auth,
    sync,
)
from identityprovider.tests.factory import SSOObjectFactory


class TwoFactorTestCase(TestCase):

    def setUp(self):
        super(TwoFactorTestCase, self).setUp()
        self.request = RequestFactory().get('/')
        self.request.user = Mock()
        self.request.user.is_authenticated.return_value = True
        self.request.session = {}
        self.factory = SSOObjectFactory()

    def make_device(self, counter=0):
        account = self.factory.make_account()
        return self.factory.make_device(account, counter=counter)

    @patch('identityprovider.models.twofactor.datetime')
    def test_login_sets_session(self, mock_datetime):
        mock_datetime.utcnow.return_value = datetime(2012, 04, 01)
        login(self.request)
        self.assertEqual(self.request.session[TWOFACTOR_LOGIN],
                         mock_datetime.utcnow.return_value)

    def test_logout_clears_sessiosn(self):
        login(self.request)

        logout(self.request)
        self.assertFalse(TWOFACTOR_LOGIN in self.request.session)

    def test_is_fresh_when_no_twofactor(self):
        self.assertFalse(is_fresh(self.request))

    def test_is_fresh_when_twofactor(self):
        login(self.request)
        self.assertTrue(is_fresh(self.request))

    def test_is_fresh_with_old_session(self):
        login(self.request)
        # fake old session
        self.request.session[TWOFACTOR_LOGIN] -= timedelta(
            0, settings.TWOFACTOR_FRESHNESS + 1)
        self.assertFalse(is_fresh(self.request))

    def test_is_upgraded_when_no_twofactor(self):
        self.assertFalse(is_upgraded(self.request))

    def test_is_upgraded_when_twofactor(self):
        login(self.request)
        self.assertTrue(is_upgraded(self.request))

    def test_is_upgraded_with_old_session(self):
        login(self.request)
        # fake old session
        self.request.session[TWOFACTOR_LOGIN] -= timedelta(
            0, settings.TWOFACTOR_TTL)
        self.assertFalse(is_upgraded(self.request))

    def test_is_authenticated_anonymous(self):
        self.request.user.is_authenticated.return_value = False
        self.assertFalse(is_authenticated(self.request))

    def test_is_authenticated_twofactor_disabled(self):
        name = 'identityprovider.models.twofactor.is_twofactor_enabled'
        with patch(name) as mock_enabled:
            mock_enabled.return_value = False
            self.assertTrue(is_authenticated(self.request))

    def test_is_authenticated_twofactor_not_required(self):
        name = 'identityprovider.models.twofactor.is_twofactor_enabled'
        with patch(name) as mock_enabled:
            mock_enabled.return_value = True
            self.request.user.twofactor_required = False
            self.assertTrue(is_authenticated(self.request))

    def test_is_authenticated_and_upgraded(self):
        base = 'identityprovider.models.twofactor.'
        name = base + 'is_twofactor_enabled'
        with patch(name) as mock_enabled:
            mock_enabled.return_value = True
            self.request.user.twofactor_required = True
            name = base + 'is_upgraded'
            with patch(name) as mock_upgraded:
                self.assertEqual(is_authenticated(self.request),
                                 mock_upgraded.return_value)

    @patch('identityprovider.models.twofactor.is_twofactor_enabled')
    def test_site_requires_twofactor_auth_with_twofactor_disabled(
            self, mock_enabled):

        mock_enabled.return_value = False
        rpconfig = Mock()
        rpconfig.twofactor_required.return_value = True

        required = site_requires_twofactor_auth(self.request, None, rpconfig)
        self.assertFalse(required)

    @patch('identityprovider.models.twofactor.is_twofactor_enabled')
    def test_site_requires_twofactor_auth_with_twofactor_enabled_no_rpconfig(
            self, mock_enabled):

        mock_enabled.return_value = False

        required = site_requires_twofactor_auth(self.request, None, None)
        self.assertFalse(required)

    @patch('identityprovider.models.twofactor.is_twofactor_enabled')
    def test_site_requires_twofactor_auth_with_twofactor_enabled(
            self, mock_enabled):

        mock_enabled.return_value = True
        rpconfig = Mock()
        rpconfig.twofactor_required.return_value = True

        required = site_requires_twofactor_auth(self.request, None, rpconfig)
        self.assertTrue(required)

    def test_is_twofactor_enabled_in_readonly_mode(self):
        with patch.multiple(settings, READ_ONLY_MODE=True):
            self.assertFalse(is_twofactor_enabled(self.request))

    @patch('identityprovider.models.twofactor.gargoyle.is_active')
    def test_is_twofactor_enabled_in_readwrite_mode(self, mock_is_active):
        with patch.multiple(settings, READ_ONLY_MODE=False):
            enabled = is_twofactor_enabled(self.request)
            self.assertEqual(enabled, mock_is_active.return_value)

    def test_syncs_device_with_three_otps(self):
        """The device syncs itself based on the otps provided.

        These provided otps must be sequential, and the first of them must have
        been generated from a counter within a window of "current counter" and
        +SYNC_WINDOW around the current device counter."""

        device = self.make_device()
        original_counter = device.counter

        counters = range(original_counter + 1, original_counter + 4)
        otps = [hotp(device.key, counter) for counter in counters]

        self.assertTrue(device.sync(otps))
        device = AuthenticationDevice.objects.get(id=device.id)
        next_counter = counters[-1] + 1
        self.assertEqual(device.counter, next_counter)

    def test_syncs_device_with_otps_starting_from_same_counter(self):
        device = self.make_device()
        original_counter = device.counter

        counters = range(original_counter, original_counter + 2)
        otps = [hotp(device.key, counter) for counter in counters]

        self.assertTrue(device.sync(otps))
        device = AuthenticationDevice.objects.get(id=device.id)
        next_counter = counters[-1] + 1
        self.assertEqual(device.counter, next_counter)

    def test_doesnt_sync_if_an_otp_is_invalid(self):
        device = self.make_device()
        original_counter = device.counter

        counters = range(original_counter + 1, original_counter + 3)
        otps = [hotp(device.key, counter) for counter in counters]
        otps.append('oops')

        self.assertRaises(AuthenticationError, device.sync, otps)

    def test_syncs_otps_distant_from_counter(self):
        """The device syncs otps even after the next counter.

        For example, if the current counter is 2, the device can sync with
        otps from which the first one starts with 7."""

        original_counter = 2
        device = self.make_device(counter=original_counter)

        first_counter = 7
        last_counter = first_counter + 2
        counters = range(first_counter, last_counter + 1)
        otps = [hotp(device.key, counter) for counter in counters]

        self.assertTrue(device.sync(otps))
        device = AuthenticationDevice.objects.get(id=device.id)
        self.assertEqual(device.counter, last_counter + 1)

    def test_doesnt_sync_if_ahead_of_the_window(self):
        """The device can't sync if the otps start after the window.

        For example, if the current counter is 2, the device can't sync with
        otps from which the first one starts with 2 + SYNC_WINDOW + 1."""

        original_counter = 2
        device = self.make_device(counter=original_counter)

        first_counter = original_counter + settings.HOTP_SYNC_WINDOW + 1
        last_counter = first_counter + 2
        counters = range(first_counter, last_counter + 1)
        otps = [hotp(device.key, counter) for counter in counters]

        self.assertFalse(device.sync(otps))
        device = AuthenticationDevice.objects.get(id=device.id)
        self.assertEqual(device.counter, original_counter)

    def test_syncs_otps_when_in_the_window_limit(self):
        """The device syncs otps that start in the window limit."""

        original_counter = 4
        device = self.make_device(counter=original_counter)

        first_counter = original_counter + settings.HOTP_SYNC_WINDOW
        last_counter = first_counter + 3
        counters = range(first_counter, last_counter + 1)
        otps = [hotp(device.key, counter) for counter in counters]

        self.assertTrue(device.sync(otps))
        device = AuthenticationDevice.objects.get(id=device.id)
        self.assertEqual(device.counter, last_counter + 1)

    def test_syncs_otps_when_otps_are_high_numbers(self):
        original_counter = 97
        device = self.make_device(counter=original_counter)

        first_counter = 98
        last_counter = first_counter + 2
        counters = range(first_counter, last_counter + 1)
        otps = [hotp(device.key, counter) for counter in counters]

        self.assertTrue(device.sync(otps))
        device = AuthenticationDevice.objects.get(id=device.id)
        self.assertEqual(device.counter, last_counter + 1)

    def test_doesnt_sync_if_an_otp_is_misplaced(self):
        """The device can't sync when one if the provided otps is misplaced."""
        device = self.make_device()
        original_counter = device.counter

        counters = range(original_counter + 1, original_counter + 3)
        counters.append(original_counter + 20)
        otps = [hotp(device.key, counter) for counter in counters]

        self.assertFalse(device.sync(otps))
        device = AuthenticationDevice.objects.get(id=device.id)
        self.assertEqual(device.counter, original_counter)

    def test_stop_synching_in_the_first_device_that_syncs(self):
        account = Mock()
        devices = [Mock(), Mock(), Mock()]
        otps = ['otp1', 'otp2']

        account.devices.order_by.return_value = devices
        devices[0].sync.return_value = False
        devices[1].sync.return_value = True

        self.assertTrue(sync(account, otps))
        devices[0].sync.assert_called_once_with(otps)
        devices[1].sync.assert_called_once_with(otps)
        self.assertFalse(devices[2].called)

    def test_raises_error_if_none_of_the_devices_sync(self):
        account = Mock()
        devices = [Mock(), Mock(), Mock()]
        otps = ['otp1', 'otp2']

        account.devices.order_by.return_value = devices
        devices[0].sync.return_value = False
        devices[1].sync.return_value = False
        devices[2].sync.return_value = False

        self.assertRaises(AuthenticationError, sync, account, otps)
        devices[0].sync.assert_called_once_with(otps)
        devices[1].sync.assert_called_once_with(otps)
        devices[2].sync.assert_called_once_with(otps)
