import sys
from unittest import TestCase

import oops
from django.conf import settings
from django.test.client import RequestFactory
from mock import patch

from identityprovider.oops_config import (
    ErrorReporter,
    global_error_reporter,
)


class ErrorReporterTests(TestCase):

    def setUp(self):
        super(ErrorReporterTests, self).setUp()
        self.reports = []
        self.published_ids = []
        self._publisher_id = 0

    def test_raw_config(self):
        reporter = ErrorReporter()
        self.assertIsInstance(reporter.oops_config, oops.Config)

    def test_setup_returns_config(self):
        reporter = ErrorReporter()
        config = reporter.setup()
        self.assertEqual(config, reporter.oops_config)

    def test_setup_uses_settings(self):
        reporter = ErrorReporter()
        value = 'some value'
        with patch.multiple(settings, OOPSES=dict(template=dict(key=value))):
            reporter.setup()
        self.assertEqual(value, reporter.oops_config.template['key'])

    def test_has_global(self):
        self.assertIsInstance(global_error_reporter, ErrorReporter)

    def capture_oopses(self, report):
        self.reports.append(report)
        new_id = self._publisher_id
        self.published_ids.append(new_id)
        self._publisher_id += 1
        return [new_id]

    def test_trigger_oops_makes_oops(self):
        reporter = ErrorReporter()
        reporter.oops_config.publisher = self.capture_oopses
        reporter.trigger_oops(None)
        self.assertEqual(1, len(self.reports))

    def test_trigger_oops_returns_ids(self):
        reporter = ErrorReporter()
        reporter.oops_config.publisher = self.capture_oopses
        ids = reporter.trigger_oops(None)
        self.assertEqual(self.published_ids, ids)

    def test_trigger_oops_uses_exc_info(self):
        reporter = ErrorReporter()
        reporter.oops_config.publisher = self.capture_oopses
        try:
            1/0
        except Exception:
            exc_info = sys.exc_info()
        reporter.trigger_oops(exc_info=exc_info)
        self.assertEqual(exc_info[0].__name__, self.reports[0]['type'])

    def test_make_context_uses_environ(self):
        reporter = ErrorReporter()
        environ = dict(key='value')
        context = reporter.make_context(environ=environ)
        self.assertEqual(environ, context['wsgi_environ'])

    def test_make_context_uses_oops_context_from_environ(self):
        reporter = ErrorReporter()
        value = 'value'
        oops_context = dict(key=value)
        environ = {'oops.context': oops_context}
        context = reporter.make_context(environ=environ)
        self.assertEqual(value, context['key'])

    def test_make_context_uses_timeline_from_environ(self):
        reporter = ErrorReporter()
        timeline = object()
        environ = {'timeline.timeline': timeline}
        context = reporter.make_context(environ=environ)
        self.assertEqual(timeline, context['timeline'])

    def test_make_context_uses_url_from_request(self):
        reporter = ErrorReporter()
        request = RequestFactory().request()
        context = reporter.make_context(request=request)
        self.assertEqual(request.build_absolute_uri(), context['url'])

    def test_make_context_uses_environ_from_request(self):
        reporter = ErrorReporter()
        request = RequestFactory().request()
        context = reporter.make_context(request=request)
        self.assertEqual(request.environ, context['wsgi_environ'])
