import time
from datetime import date

from bzrlib.errors import NotBranchError
from gargoyle.testutils import switches
from mock import Mock, patch

from identityprovider.preflight import get_version_info
from identityprovider.tests.utils import (
    SSOBaseTestCase,
    TeamConditionSelectiveMixin,
)


class TestPreflight(SSOBaseTestCase, TeamConditionSelectiveMixin):

    URL = '/preflight/'
    key = 'PREFLIGHT'

    def setUp(self):
        self.email = 'email@domain.local'
        self.password = 'testpassword'
        super(TestPreflight, self).setUp()

        self.account = self.make_account()
        self.enable_flag()

    def test_anonymous(self):
        response = self.client.get(self.URL)

        self.assertEqual(404, response.status_code)

    @switches(PREFLIGHT=False)
    def test_flag_not_active(self):
        login = self.client.login(username=self.email, password=self.password)
        self.assertTrue(login)

        response = self.client.get(self.URL)

        self.assertEqual(404, response.status_code)

    @switches(PREFLIGHT=True)
    def test_success(self):
        login = self.client.login(username=self.email, password=self.password)
        self.assertTrue(login)

        response = self.client.get(self.URL)

        self.assertEqual(200, response.status_code)
        libs = set([x['name'] for x in response.context[0]['versions']])
        expected = set([
            'Django',
            'Platform',
            'Python',
            'django',
            'django_openid_auth',
            'identityprovider',
            'openid',
            'piston',
            'preflight',
            'ssoclient',
            'config',
        ])

        self.assertEqual(expected, libs)

    @patch('identityprovider.preflight.import_module')
    def test_version_from_branch(self, mock_import):
        mod = Mock()
        mod.__name__ = 'foo'
        mod.__file__ = '/some/path'
        mod.__version__ = '1.2.3'
        mock_import.return_value = mod

        name = 'identityprovider.preflight.Branch.open_containing'
        tip_date = date(2013, 6, 24)
        with patch(name) as mock_open:
            mock_revision = Mock()
            mock_revision.timestamp = time.mktime(tip_date.timetuple())
            mock_branch = Mock()
            mock_branch.nick = 'trunk'
            mock_branch.revno.return_value = 42
            mock_branch.repository.get_revision.return_value = mock_revision
            mock_open.return_value = (mock_branch, 'other/path')

            version_info = get_version_info('foo')

        self.assertEqual(version_info,
                         {'name': 'foo', 'version': 'trunk@42 (13.06.24)'})

    @patch('identityprovider.preflight.import_module')
    def test_version_from_version_attribute(self, mock_import):
        mod = Mock()
        mod.__name__ = 'foo'
        mod.__file__ = '/some/path'
        mod.__version__ = '1.2.3'
        mock_import.return_value = mod

        name = 'identityprovider.preflight.Branch.open_containing'
        with patch(name) as mock_open:
            mock_open.side_effect = NotBranchError(mod.__file__)

            version_info = get_version_info(mod)

        self.assertEqual(version_info,
                         {'name': 'foo', 'version': '1.2.3'})

    @patch('identityprovider.preflight.import_module')
    def test_version_without_version_attribute(self, mock_import):
        mod = Mock()
        mod.__name__ = 'foo'
        mod.__file__ = '/some/path'
        mock_import.return_value = mod

        name = 'identityprovider.preflight.Branch.open_containing'
        with patch(name) as mock_open:
            mock_open.side_effect = NotBranchError(mod.__file__)

            version_info = get_version_info(mod)

        self.assertEqual(version_info,
                         {'name': 'foo', 'version': 'version not available'})

    def test_config_branch_version(self):
        name = 'identityprovider.preflight.Branch.open_containing'
        tip_date = date(2013, 6, 24)
        with patch(name) as mock_open:
            mock_revision = Mock()
            mock_revision.timestamp = time.mktime(tip_date.timetuple())
            mock_branch = Mock()
            mock_branch.nick = 'config'
            mock_branch.revno.return_value = 42
            mock_branch.repository.get_revision.return_value = mock_revision
            mock_open.return_value = (mock_branch, '')

            version_info = get_version_info('config')

        self.assertEqual(version_info,
                         {'name': 'config',
                          'version': 'config@42 (13.06.24)'})
