# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import json
import os
import stat
import shutil
import tempfile
import socket
import urllib2

from StringIO import StringIO

from django.conf import settings
from django.contrib.auth.models import User

from identityprovider.readonly import ReadOnlyManager
from identityprovider.readonly import (
    _remote_req,
    get_server_atts,
    update_server,
)
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import SSOBaseTestCase


class ReadOnlyBaseTestCase(SSOBaseTestCase):
    """A base TestCase for backing up and restoring our readonly settings"""

    def setUp(self):
        super(ReadOnlyBaseTestCase, self).setUp()
        # XXX, 2013-03-29, nessita: Do not remove this patching since the
        # READ_ONLY_MODE is tweak in tests, and it should be guaranteed the
        # readonlymode stays False
        overrides = self.settings(
            READ_ONLY_MODE=False,
            READONLY_SECRET='testsecret',
            DBFAILOVER_FLAG_DIR=tempfile.mkdtemp(),
        )
        overrides.enable()
        self.addCleanup(overrides.disable)
        self.addCleanup(shutil.rmtree, settings.DBFAILOVER_FLAG_DIR, True)

        self.rm = ReadOnlyManager()


class RemoteRequestTestCase(SSOBaseTestCase):
    msg = 'hello'
    host = 'myhost'
    scheme = 'https'
    vhost = 'http://foobar.baz'

    def setup_mock_urlopen(self, func):
        self.orig_urlopen = urllib2.urlopen
        urllib2.urlopen = func

    def restore_orig_urlopen(self):
        urllib2.urlopen = self.orig_urlopen

    def mock_urlopen(self, req, data=None):
        self.req = req
        self.assertEqual(5, socket.getdefaulttimeout())
        return StringIO(self.msg)

    def test_plain_remote_req(self):
        self.setup_mock_urlopen(self.mock_urlopen)
        server = {'host': self.host}
        self.assertEqual(self.msg, _remote_req(**server))
        self.assertEqual(self.host, self.req.get_host())
        self.restore_orig_urlopen()

    def test_https_remote_req(self):
        self.setup_mock_urlopen(self.mock_urlopen)
        server = {'host': self.host, 'scheme': self.scheme}
        self.assertEqual(self.msg, _remote_req(**server))
        self.assertEqual(self.scheme, self.req.get_type())
        self.assertEqual(self.host, self.req.get_host())
        self.restore_orig_urlopen()

    def test_vhost_remote_req(self):
        self.setup_mock_urlopen(self.mock_urlopen)
        server = {
            'host': self.host,
            'scheme': self.scheme,
            'virtual_host': self.vhost,
        }
        self.assertEqual(self.msg, _remote_req(**server))
        headers = {'Host': self.vhost}
        self.assertEqual(headers, self.req.headers)
        self.assertEqual(self.scheme, self.req.get_type())
        self.assertEqual(self.host, self.req.get_host())
        self.restore_orig_urlopen()

    def test_remote_req_urllib2_error(self):
        def mock_urlopen(req, data=None):
            raise urllib2.URLError((-1, 'error'))

        self.setup_mock_urlopen(mock_urlopen)
        server = {'host': self.host}
        self.assertEqual(None, _remote_req(**server))
        self.restore_orig_urlopen()


class ReadOnlyFlagFilesTestCase(ReadOnlyBaseTestCase):

    def test_flag_files_in_right_directory(self):
        self.rm.set_readonly()
        flags = os.listdir(settings.DBFAILOVER_FLAG_DIR)
        self.assertTrue('db.readonly' in flags)

    def test_readonly_flag_files_have_right_mode(self):
        self.rm.set_readonly()
        mode = os.stat(self.rm.readonly_marker_file)[stat.ST_MODE]
        self.assertTrue(mode & stat.S_IRGRP)
        self.assertTrue(mode & stat.S_IWGRP)


class ReadOnlyDataTestCase(ReadOnlyBaseTestCase):

    def test_readonly_returns_404_for_get(self):
        response = self.client.get('/readonlydata')
        self.assertEqual(404, response.status_code)

    def test_readonly_returns_404_without_readonly_secret(self):
        response = self.client.post('/readonlydata')
        self.assertEqual(404, response.status_code)

    def test_readonly_returns_404_with_bad_secret(self):
        data = {'secret': 'something different'}
        response = self.client.post('/readonlydata', data)
        self.assertEqual(404, response.status_code)

    def test_readonly_returns_200_with_right_secret(self):
        data = {'secret': settings.READONLY_SECRET}
        response = self.client.post('/readonlydata', data)
        self.assertEqual(200, response.status_code)

    def test_readonly_data(self):
        post = {'secret': settings.READONLY_SECRET}
        response = self.client.post('/readonlydata', post)
        data = json.loads(response.content)
        self.assertTrue(u'readonly' in data)

    def test_readonly_set_readonly(self):
        post = {'secret': settings.READONLY_SECRET, 'action': 'set'}
        response = self.client.post('/readonlydata', post)
        data = json.loads(response.content)
        self.assertTrue(data['readonly'])

    def test_readonly_clear_readonly(self):
        post = {'secret': settings.READONLY_SECRET, 'action': 'clear'}
        response = self.client.post('/readonlydata', post)
        data = json.loads(response.content)
        self.assertFalse(data['readonly'])


class ReadOnlyViewsTestCase(ReadOnlyBaseTestCase):

    def login_with_staff(self):
        user = User.objects.create(username='admin')
        user.is_staff = True
        user.set_password('password')
        user.save()
        r = self.client.login(username='admin', password='password')
        self.assertTrue(r)

    def patch_urlopen(self, mock_urlopen=None):
        self.reqs = []

        def _mock_urlopen(req, data=None):
            self.reqs.append(req)
            return StringIO(json.dumps({}))

        if mock_urlopen is None:
            mock_urlopen = _mock_urlopen

        old = urllib2.urlopen
        urllib2.urlopen = mock_urlopen
        self.addCleanup(setattr, urllib2, 'urlopen', old)

    def test_readonly_admin(self):
        self.login_with_staff()

        new_setting = [
            {'SERVER_ID': 'localhost', 'SCHEME': 'http',
             'HOST': 'localhost', 'VIRTUAL_HOST': '',
             'PORT': '8000'}
        ]
        with self.settings(APP_SERVERS=new_setting):
            r = self.client.get('/readonly')

        self.assertTemplateUsed(r, 'admin/readonly.html')
        expected = {'appservers': [{'name': 'localhost', 'reachable': False}],
                    'clear_all_readonly': False,
                    'set_all_readonly': False}
        for item in r.context:
            for key, value in expected.items():
                self.assertEqual(item[key], value)

    def test_get_server_atts_server_unreachable(self):
        servers = [{'SERVER_ID': 'localhost', 'SCHEME': 'http',
                    'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'}]
        expected = {'appservers': [{'name': 'localhost', 'reachable': False}],
                    'clear_all_readonly': False,
                    'set_all_readonly': False}
        atts = get_server_atts(servers)
        self.assertEqual(atts, expected)

    def test_get_server_atts_data_error(self):

        def mock_loads(data):
            raise ValueError()

        old_loads = json.loads
        json.loads = mock_loads
        self.addCleanup(setattr, json, 'loads', old_loads)
        self.patch_urlopen()

        servers = [{'SERVER_ID': 'localhost', 'SCHEME': 'http',
                    'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'}]
        expected = {'appservers': [{'name': 'localhost', 'reachable': False}],
                    'clear_all_readonly': False,
                    'set_all_readonly': True}
        atts = get_server_atts(servers)
        self.assertEqual(atts, expected)

    def test_get_server_atts_readonly(self):

        def mock_urlopen(req, data=None):
            data = {'readonly': True}
            return StringIO(json.dumps(data))

        self.patch_urlopen(mock_urlopen=mock_urlopen)

        servers = [{'SERVER_ID': 'localhost', 'SCHEME': 'http',
                    'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'}]
        expected = {'appservers': [{'name': 'localhost', 'reachable': True,
                                    'readonly': True}],
                    'clear_all_readonly': True,
                    'set_all_readonly': False}
        atts = get_server_atts(servers)
        self.assertEqual(atts, expected)

    def test_readonly_confirm_get(self):
        self.login_with_staff()

        r = self.client.get('/readonly/localhost/set')
        self.assertTemplateUsed(r, 'admin/readonly_confirm.html')
        self.assertEqual(r.context['appserver'], 'localhost')
        self.assertEqual(r.context['action'], 'set')

    def test_readonly_confirm_post(self):
        self.patch_urlopen()
        self.login_with_staff()

        new_setting = [
            {'SERVER_ID': 'localhost', 'SCHEME': 'http',
             'HOST': 'localhost', 'VIRTUAL_HOST': '',
             'PORT': '8000'}
        ]
        with self.settings(APP_SERVERS=new_setting):
            r = self.client.post('/readonly/localhost/set')
        data = map(lambda x: x.data, self.reqs)
        self.assertEqual(data, [
            "action=set&secret=%s" % settings.READONLY_SECRET
        ])

        self.assertRedirects(r, '/readonly')

    def test_update_server_all_appservers(self):
        self.patch_urlopen()
        new_setting = [
            {'SERVER_ID': 'localhost', 'SCHEME': 'http',
             'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'},
            {'SERVER_ID': 'otherhost', 'SCHEME': 'http',
             'HOST': 'otherhost', 'VIRTUAL_HOST': '', 'PORT': '8000'},
        ]
        with self.settings(APP_SERVERS=new_setting):
            update_server('set')
        data = map(lambda x: x.data, self.reqs)
        self.assertEqual(data, [
            "action=set&secret=%s" % settings.READONLY_SECRET,
            "action=set&secret=%s" % settings.READONLY_SECRET
        ])

    def test_update_server_one_appserver(self):
        self.patch_urlopen()
        new_setting = [
            {'SERVER_ID': 'localhost', 'SCHEME': 'http',
             'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'},
            {'SERVER_ID': 'otherhost', 'SCHEME': 'http',
             'HOST': 'otherhost', 'VIRTUAL_HOST': '', 'PORT': '8000'},
        ]
        with self.settings(APP_SERVERS=new_setting):
            update_server('set', 'localhost')
        data = map(lambda x: x.data, self.reqs)
        self.assertEqual(data, [
            "action=set&secret=%s" % settings.READONLY_SECRET
        ])

    def test_update_server_one_connection(self):
        self.patch_urlopen()
        new_setting = [
            {'SERVER_ID': 'localhost', 'SCHEME': 'http',
             'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'},
        ]
        with self.settings(APP_SERVERS=new_setting):
            update_server('set', 'localhost')
        data = map(lambda x: x.data, self.reqs)
        self.assertEqual(data, [
            "action=set&secret=%s" % settings.READONLY_SECRET,
        ])

    def test_nop_db_in_readonly_mode(self):
        account = self.factory.make_account()

        self.rm.set_readonly()
        self.addCleanup(self.rm.clear_readonly)

        response = self.client.get('/+login', follow=True)
        # assert we're in readonly mode
        self.assertTrue(response.context['readonly'])
        # login
        csrf_token = response.context['request'].META['CSRF_COOKIE']
        data = {'email': account.preferredemail.email,
                'password': DEFAULT_USER_PASSWORD,
                'csrftoken': csrf_token}
        response = self.client.post('/+login', data=data)

        # assert logged in
        self.assertRedirects(response, '/')
