from contextlib import contextmanager
from unittest import TestCase

from mock import Mock
from timeline import Timeline

from identityprovider.timeline import (
    get_request_timeline,
    get_timing_function,
    maybe_contextmanager,
)


class GetRequestTimelineTests(TestCase):

    def test_no_timeline(self):
        request = Mock()
        request.META = dict()
        self.assertEqual(None, get_request_timeline(request))

    def test_timeline(self):
        request = Mock()
        timeline = object()
        request.META = {'timeline.timeline': timeline}
        self.assertEqual(timeline, get_request_timeline(request))


class GetTimingFunctionTests(TestCase):

    def test_no_timeline(self):
        timer_fn = get_timing_function(None)
        with timer_fn('category', 'detail'):
            # No crash
            pass

    def test_timeline(self):
        timeline = Timeline()
        timer_fn = get_timing_function(timeline)
        with timer_fn('category', 'detail'):
            pass
        self.assertEqual(1, len(timeline.actions))
        self.assertEqual('category', timeline.actions[0].category)
        self.assertEqual('detail', timeline.actions[0].detail)
        self.assertNotEqual(None, timeline.actions[0].duration)

    def test_finishes_action_even_with_exception(self):
        timeline = Timeline()
        timer_fn = get_timing_function(timeline)

        def do_test():
            with timer_fn('category', 'detail'):
                1/0
        self.assertRaises(ZeroDivisionError, do_test)
        self.assertEqual(1, len(timeline.actions))
        self.assertNotEqual(None, timeline.actions[0].duration)


class MaybeContextManagerTests(TestCase):

    def test_None(self):
        with maybe_contextmanager(None, 'a'):
            # no crash
            pass

    def test_passes_args(self):
        passed_args = []
        passed_kwargs = []

        @contextmanager
        def cm(*args, **kwargs):
            passed_args.append(args)
            passed_kwargs.append(kwargs)
            yield
        with maybe_contextmanager(cm, 'a', b='c'):
            pass
        self.assertEqual([('a',)], passed_args)
        self.assertEqual([dict(b='c')], passed_kwargs)

    def test_yields_value(self):
        @contextmanager
        def cm(*args, **kwargs):
            yield 'the value'
        with maybe_contextmanager(cm) as the_value:
            self.assertEqual('the value', the_value)
