# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import urllib2
from datetime import date, datetime, time, timedelta
from uuid import uuid4

from django.conf import settings
from django.contrib.auth.hashers import get_hasher
from django.contrib.auth.models import make_password
from django.contrib.sessions.models import Session
from django.test import TestCase
from gargoyle.testutils import switches
from mock import patch

from identityprovider.models.person import Person
from identityprovider.utils import (
    canonical_url,
    clean_sessions_until,
    http_request_with_timeout,
    polite_form_errors,
    redirection_url_for_token,
    validate_launchpad_password,
)
from identityprovider.tests.utils import (
    patch_settings,
    SSOBaseTestCase,
)
from identityprovider.utils import (
    encrypt_launchpad_password,
    get_current_brand,
    get_cypher_migration_status,
)


class CanonicalUrlTestCase(SSOBaseTestCase):

    def test_when_object_is_not_person(self):
        self.assertTrue(canonical_url(None) is None)

    def test_when_view_name_is_none(self):
        url = canonical_url(Person(name="test"), view_name=None)
        self.assertEqual(url, "https://launchpad.net/~test")

    def test_when_view_name_is_not_none(self):
        url = canonical_url(Person(name="test"), view_name="TEST")
        self.assertEqual(url, "https://launchpad.net/~test/TEST")


class PoliteFormErrorsTestCase(SSOBaseTestCase):

    def test_polite_errors(self):
        text = "Enter a valid e-mail address."
        errors = {'email': [text]}

        polite_form_errors(errors)

        self.assertTrue(errors['email'][0] != text)


class ValidateLaunchpadPasswordTestCase(SSOBaseTestCase):

    def test_when_there_is_binascii_error_is_raised(self):
        self.assertFalse(validate_launchpad_password('a', 'b'))


class HttpRequestWithTimeoutTestCase(TestCase):

    def setUp(self):
        super(HttpRequestWithTimeoutTestCase, self).setUp()
        self.exception = None
        self.output = None
        self.headers = None

        class MockDatafile(object):

            def __init__(self, output, headers):
                self.output = output
                self.headers = headers

            def read(self):
                return self.output

            def info(self):
                return self.headers

        def mock_urlopen(request):
            self.request = request
            if self.exception:
                raise urllib2.URLError("arg")
            return MockDatafile(self.output, self.headers)

        self.urlopen = urllib2.urlopen
        urllib2.urlopen = mock_urlopen

    def tearDown(self):
        urllib2.urlopen = self.urlopen

    def test_without_data_and_successful_outcome(self):
        self.output = "output"
        data, headers = http_request_with_timeout("http://example.com")

        self.assertEqual(data, "output")

    def test_without_data_and_failing(self):
        self.exception = True
        data, headers = http_request_with_timeout("http://example.com")
        self.assertEqual(data, None)

    def test_with_data_and_successful_outcome(self):
        self.output = "output"
        data, headers = http_request_with_timeout("http://example.com",
                                                  {"i1": "v1", "i2": "v2"})
        self.assertTrue("&" in self.request.data)


class GetCurrentBrandTestCase(TestCase):
    def test_defaults_to_ubuntuone_brand(self):
        with patch_settings(BRAND=None):
            brand = get_current_brand()

        self.assertEqual('ubuntuone', brand)

    def test_no_switch_but_brand_setting_lp(self):
        with patch_settings(BRAND='launchpad'):
            brand = get_current_brand()

        self.assertEqual('launchpad', brand)

    @switches(BRAND_LAUNCHPAD=True)
    def test_feature_switch_for_lp(self):
        brand = get_current_brand()

        self.assertEqual('launchpad', brand)

    def get_is_active_mock(self):
        patcher = patch('gargoyle.gargoyle.is_active')
        self.addCleanup(patcher.stop)
        return patcher.start()

    def test_default_on_error(self):
        """Any exception during get_current_brand results in default."""
        mock_is_active = self.get_is_active_mock()
        mock_is_active.side_effect = Exception('bang')

        with patch.multiple(settings, BRAND='ubuntuone'):
            brand = get_current_brand()

        self.assertEqual('ubuntuone', brand)


class RedirectionURLForTokenTestCase(SSOBaseTestCase):

    def test_if_token_is_none(self):
        url = redirection_url_for_token(None)
        self.assertEqual(url, "/")

    def test_if_token_is_not_none(self):
        token = "ABCDEFGH" * 2
        url = redirection_url_for_token(token)
        self.assertEqual(url, "/%s/+decide" % token)


class SessionTestCase(SSOBaseTestCase):

    def create_session(self, expire_date):
        return Session.objects.create(
            session_key=str(uuid4()),
            session_data='some data',
            expire_date=expire_date,
        )

    def test_deletes_sessions_created_until_provided_date(self):
        creation_date = date.today()
        expire_date = creation_date + timedelta(
            seconds=settings.SESSION_COOKIE_AGE)

        self.create_session(expire_date)
        self.create_session(expire_date - timedelta(days=1))
        self.create_session(expire_date - timedelta(days=2))
        session4 = self.create_session(expire_date + timedelta(days=1))

        with self.assertNumQueries(1):
            clean_sessions_until(creation_date)

        sessions = list(Session.objects.all())

        self.assertEqual(sessions, [
            session4,
        ])

    def test_deletes_sessions_with_limit(self):
        creation_date = date.today()
        expire_date = creation_date + timedelta(
            seconds=settings.SESSION_COOKIE_AGE)

        self.create_session(expire_date)
        self.create_session(expire_date - timedelta(days=1))
        session3 = self.create_session(expire_date - timedelta(days=2))
        session4 = self.create_session(expire_date + timedelta(days=1))
        session5 = self.create_session(expire_date + timedelta(days=2))

        with self.assertNumQueries(1):
            clean_sessions_until(creation_date, limit=2)

        sessions = list(Session.objects.all())

        self.assertEqual(sessions, [
            session3,
            session4,
            session5,
        ])

    def test_deletes_sessions_until_exact_date_and_not_a_second_more(self):
        creation_date = date.today()
        expire_date = creation_date + timedelta(
            seconds=settings.SESSION_COOKIE_AGE)

        self.create_session(expire_date)
        self.create_session(expire_date - timedelta(days=1))
        self.create_session(expire_date - timedelta(days=2))
        session4 = self.create_session(datetime.combine(expire_date,
                                                        time(second=1)))

        with self.assertNumQueries(1):
            clean_sessions_until(creation_date)

        sessions = list(Session.objects.all())

        self.assertEqual(sessions, [
            session4,
        ])

    def test_doesnt_delete_sessions_if_dry_running(self):
        creation_date = date.today()
        expire_date = creation_date + timedelta(
            seconds=settings.SESSION_COOKIE_AGE)

        session1 = self.create_session(expire_date)
        session2 = self.create_session(expire_date - timedelta(days=1))
        session3 = self.create_session(expire_date - timedelta(days=2))
        session4 = self.create_session(expire_date + timedelta(days=1))

        with self.assertNumQueries(1):
            clean_sessions_until(creation_date, dry_run=True)

        sessions = list(Session.objects.all())

        self.assertEqual(sessions, [
            session1,
            session2,
            session3,
            session4,
        ])


class CypherMigrationStatusTestCase(SSOBaseTestCase):

    def make_password(self, encrypted_password):
        self.factory.make_account(password=encrypted_password,
                                  password_encrypted=True)

    def test_gets_cypher_migration_status(self):
        self.make_password(make_password(str(uuid4())))
        self.make_password(encrypt_launchpad_password(str(uuid4())))
        self.make_password(make_password(str(uuid4())))

        result = get_cypher_migration_status()

        self.assertEqual(result['new_hashes'], 2)
        self.assertEqual(result['old_hashes'], 1)


class CypherConfigurationTestCase(SSOBaseTestCase):

    def test_uses_bcrypt_by_default(self):
        password = 'some pass'
        salt = get_hasher('bcrypt13').salt()

        default_password = make_password(password, salt=salt)
        expected_password = make_password(password, salt=salt,
                                          hasher='bcrypt13')

        self.assertEqual(default_password, expected_password)
