from django.core.exceptions import ValidationError
from django.utils.unittest import TestCase
from mock import (
    MagicMock,
    call,
    patch,
    sentinel,
)

from identityprovider import validators
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import SSOBaseUnittestTestCase
from identityprovider.validators import (
    BasicValidator,
    CanonicalValidator,
    Errors,
    get_password_validator,
    validate_password_policy,
)


ACCOUNT, PASSWORD = sentinel.account, sentinel.password
VALIDATORS_PATH = 'identityprovider.validators._validators'


class PasswordValidatorTestsMixin(object):

    def test_password_too_short(self):
        with self.assertRaises(ValidationError):
            self.validator('abcd3Fg')

    def test_password_invalid_chars(self):
        with self.assertRaises(ValidationError):
            self.validator(u'abcd\xe1Fgh')

    def test_password_none(self):
        with self.assertRaises(ValidationError):
            self.validator(None)

    def test_default_password_is_validated(self):
        self.assertEqual(self.validator(DEFAULT_USER_PASSWORD), None)

    def test_valid_password(self):
        self.assertEqual(self.validator('abcD3fgh'), None)


class PasswordPolicyValidatorTestCase(PasswordValidatorTestsMixin, TestCase):
    validator = lambda self, password: validate_password_policy(password)


class BasicValidatorPolicyTestCase(PasswordValidatorTestsMixin, TestCase):
    validator = BasicValidator().validate

    def test_match(self):
        validator = BasicValidator()
        result = validator.match(ACCOUNT)
        self.assertEqual(result, True)


class CanonicalValidatorTestCase(PasswordValidatorTestsMixin, TestCase):
    validator = CanonicalValidator().validate

    def test_invalid_missing_uppercase_and_digits(self):
        with self.assertRaises(ValidationError):
            self.validator('abcdefgh')

    def test_invalid_missing_digits(self):
        with self.assertRaises(ValidationError):
            self.validator('abcdefGH')

    def test_invalid_missing_uppercase(self):
        with self.assertRaises(ValidationError):
            self.validator('abcdef12')

    def test_valid(self):
        self.assertEqual(self.validator('abcdefA1'), None)

    def test_match_none(self):
        validator = CanonicalValidator()
        self.assertEqual(validator.match(None), False)

    def test_match_account_not_in_team(self):
        account = MagicMock()
        account.person_in_team.return_value = False

        validator = CanonicalValidator()
        self.assertEqual(validator.match(account), False)
        account.person_in_team.assert_called_once_with('canonical')

    def test_match_account_in_team(self):
        account = MagicMock()
        account.person_in_team.return_value = True

        validator = CanonicalValidator()
        self.assertEqual(validator.match(account), True)
        account.person_in_team.assert_called_once_with('canonical')


class PasswordValidatorTestCase(TestCase):

    def test_default_validators(self):
        defaults = validators._validators
        first, second = defaults
        self.assertIsInstance(first, CanonicalValidator)
        self.assertEqual(type(second), BasicValidator)

    def test_get_password_validator(self):
        VALIDATE_PATH = 'identityprovider.validators.validate_password_policy'
        with patch(VALIDATE_PATH) as mock_validate:
            validator = get_password_validator(ACCOUNT)
            validator(PASSWORD)

        mock_validate.assert_called_once_with(PASSWORD, ACCOUNT)

    def test_validate_password_policy_success(self):
        parent = MagicMock()
        validator1 = parent.validator1
        validator2 = parent.validator2
        validator3 = parent.validator3

        validator1.match.return_value = False
        validator2.match.return_value = True

        validators = [validator1, validator2, validator3]
        with patch(VALIDATORS_PATH, validators):
            validate_password_policy(PASSWORD, ACCOUNT)

        expected_calls = [
            call.validator1.match(ACCOUNT),
            call.validator2.match(ACCOUNT),
            call.validator2.validate(PASSWORD)
        ]
        self.assertEqual(parent.mock_calls, expected_calls)

    def test_validate_password_policy_fails(self):
        parent = MagicMock()
        validator1 = parent.validator1
        validator2 = parent.validator2

        validator1.match.return_value = False
        validator2.match.return_value = True
        validator2.validate.side_effect = ValidationError('foo bar baz')

        validators = [validator1, validator2]
        with patch(VALIDATORS_PATH, validators):
            with self.assertRaises(ValidationError):
                validate_password_policy(PASSWORD, ACCOUNT)

    def test_fallback_validation(self):
        with patch(VALIDATORS_PATH, []):
            with self.assertRaises(ValidationError):
                validate_password_policy(PASSWORD, ACCOUNT)


class ErrorsTestCase(SSOBaseUnittestTestCase):

    def test_collect_no_key(self):
        errors = Errors()
        with errors.collect():
            raise ValidationError('Some Error')
        self.assertEqual(errors['__all__'], ['Some Error'])

    def test_collect_no_key_merges_dict(self):
        errors = Errors()
        with errors.collect():
            raise ValidationError({'k': 'v', 'x': 'y'})
        self.assertEqual(errors['k'], ['v'])
        self.assertEqual(errors['x'], ['y'])

    def test_collect_with_key(self):
        errors = Errors()
        with errors.collect('key'):
            raise ValidationError('Some Error')
        self.assertEqual(errors['key'], ['Some Error'])

    def test_collect_with_key_flattens_dict(self):
        errors = Errors()
        with errors.collect('key'):
            raise ValidationError({'k': 'v', 'x': 'y'})
        self.assertEqual(sorted(errors['key'][0]), ['v', 'y'])
