from contextlib import contextmanager


def get_request_timeline(request):
    """Get the timeline for a request, if any.

    :return: a Timeline instance if the request has one, or None if not.
    """
    return request.META.get('timeline.timeline', None)


def get_timing_function(timeline):
    """Get a context manager that will put timing info in to a timeline.

    Given

      timer_fn = get_timing_function(timeline)

      def request(method, url):
          with timer_fn('http-{}'.format(method), url):
              <..make request..>

    calling request('GET', 'http://google.com/') will lead to an action
    being inserted in to the timeline with

       start: time at which "with timer_fn(..." was exectuted.
       duration: duration until the with block was left, i.e. duration of
            the request.
       category: 'http-GET'
       detail: 'http://google.com/'

    :param timeline: the Timeline instance in which to create the action,
        or None to take no timing information.
    :return: a context manager that can be called with info about the
        operation, and will record an action on the timeline for the duration
        between __enter__ and __exit__.
    """
    @contextmanager
    def timer_fn(category, detail, allow_nested=False):
        if not timeline:
            yield
        else:
            action = timeline.start(category, detail,
                                    allow_nested=allow_nested)
            try:
                yield
            finally:
                action.finish()
    return timer_fn


def get_request_timing_function(request):
    return get_timing_function(get_request_timeline(request))


@contextmanager
def maybe_contextmanager(cm, *args, **kwargs):
    """Allow an optional context manager to be used.

    When you take an optional context manager as an argument, it can be
    tricky to use it, as you can't do "with None:".

    This function allows for you to call

        with maybe_contextmanager(cm, args, ...):
            pass

    and it will work whether cm is a contextmanager or None.
    """
    if cm is None:
        yield
    else:
        with cm(*args, **kwargs) as info:
            yield info
