from contextlib import contextmanager
import re

from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _


PASSWORD_POLICY_ERROR = _('Password must be at least 8 characters long')
PASSWORD_POLICY_HELP_TEXT = PASSWORD_POLICY_ERROR
CANONICAL_PASSWORD_POLICY_ERROR = _(
    "Password must consist of lower- and upper-case characters, and at "
    "least one digit."
)
PCI_PASSWORD_POLICY_ERROR = _(
    "Your password matches a previous passowrd. PCI compliance rules "
    "forbid the reuse of recent passwords."
)


class BasicValidator(object):
    """Generic password validator.

    Accounts are valid if:
        - password contains at least 8 characters
        - password contains only ascii characters
    """

    def match(self, account):
        """Return True for all accounts."""
        # Note that for new users account will be None
        # The basic validator always matches
        return True

    def validate(self, password):
        """Check the password complies with the policy.

        Validation is successful if:
            - password contains at least 8 characters
            - password contains only ascii characters

        Raises ValidationError on failure.
        """
        # Implement the basic password policy
        try:
            # ascii only!
            str(password)
        except UnicodeEncodeError:
            raise ValidationError(_("Invalid characters in password"))
        if password is None or len(password) < 8:
            raise ValidationError(PASSWORD_POLICY_ERROR)


class CanonicalValidator(BasicValidator):
    """Password validator for Canonical accounts.

    Accounts are valid if:
        - account is member of the 'canonical' team
        - password contains at least 8 characters
        - password contains only ascii characters
        - password contains at least one lower case letter
        - password contains at least one upper case letter
        - password contains at least one number
    """

    def match(self, account):
        """Return True if the account is part of the 'canonical' team."""
        return account is not None and account.person_in_team('canonical')

    def validate(self, password):
        """Check the password complies with the policy.

        Validation is successful if:
            - password contains at least one lower case letter
            - password contains at least one upper case letter
            - password contains at least one number

        Raises ValidationError on failure.
        """
        super(CanonicalValidator, self).validate(password)

        has_lower = False
        has_upper = False
        has_number = False
        if re.search(r'[a-z]', password):
            has_lower = True
        if re.search(r'[A-Z]', password):
            has_upper = True
        if re.search(r'[0-9]', password):
            has_number = True

        if not all([has_lower, has_upper, has_number]):
            raise ValidationError(CANONICAL_PASSWORD_POLICY_ERROR)


class PCIValidator(BasicValidator):
    """Apply PCI compliance password rules"""

    def match(self, account):
        """Return True if the account is in the 'PCI' team."""
        # store the account
        self.account = account
        return (account is not None and
                account.person_in_team(settings.PCI_COMLIANCE_GROUP))

    def validate(self, password):
        """Check the new password doesn't match any of the last four
        passwords."""
        super(CanonicalValidator, self).validate(password)

        # import here to avoid circular imports
        from identityprovider.auth import LaunchpadBackend

        account = self.account
        backend = LaunchpadBackend()
        for accountpassword in account.accountpassword_set.all()[1:]:
            try:
                backend._validate_raw_password(
                    account.accountpassword, password, update=False)
            except ValidationError:
                pass
            else:
                raise ValidationError(PCI_PASSWORD_POLICY_ERROR)


_validators = [CanonicalValidator(), BasicValidator(), PCIValidator()]


def validate_password_policy(password, account=None):
    """Validate password against matching policy.

    The password is tested against the first matching policy.
    Policies are selected based on the account provided.

    Raises ValidationError on failure.
    """
    for validator in _validators:
        if validator.match(account):
            # return to exit as we have found a match
            return validator.validate(password)

    # this should never happen as the BasicValidator should always match!
    raise ValidationError(PASSWORD_POLICY_ERROR)


def get_password_validator(account):
    """Return an appropriate password validation function for the account."""
    def validator(password):
        validate_password_policy(password, account)
    return validator


class Errors(dict):
    GENERAL = "__all__"

    @contextmanager
    def collect(self, key=None):
        try:
            yield
        except ValidationError as e:
            if key is not None:
                self[key] = e.messages
            else:
                # If you pass in an empty dict, django will overwrite the
                # reference to an ordinary dict. Srsly.
                self['Dont kill me, Django!'] = None
                e.update_error_dict(self)
                del self['Dont kill me, Django!']
