# coding: utf-8

# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import re
import urllib
import urllib2
from datetime import date
from functools import partial
from StringIO import StringIO
from urlparse import urlsplit

from django.conf import settings
from django.contrib.auth.models import check_password
from django.contrib.sessions.models import Session
from django.core import mail
from django.core.exceptions import ValidationError
from django.core.urlresolvers import reverse
from django.http import QueryDict
from django.utils.html import escape
from gargoyle.testutils import switches
from mock import patch, Mock
from openid.message import IDENTIFIER_SELECT
from pyquery import PyQuery
from timeline import Timeline
from u1testutils.logging import LogHandlerTestCase

from identityprovider import signed
from identityprovider.fields import (
    INVALID_EMAIL,
    ONE_TIME_PWD_ERROR,
    REQUIRED_FIELD,
)
from identityprovider.forms import PASSWORD_LEAKED
from identityprovider.login import LOGIN_INCORRECT
from identityprovider.models import (
    Account,
    AuthToken,
    EmailAddress,
    OpenIDRPConfig,
    twofactor,
)
from identityprovider.models.authtoken import create_token
from identityprovider.models.captcha import Captcha, CaptchaResponse
from identityprovider.models.const import AccountStatus, EmailStatus, TokenType
from identityprovider.readonly import ReadOnlyManager
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import (
    MockHandler,
    SSOBaseTestCase,
    patch_settings,
)
from identityprovider.views import server
from webui import decorators
from webui.views import ui


DISPLAYNAME = 'Mark Shuttleworth'


def _args_list(flow_name, key=None):
    if key is None:
        key = 'success'
    return [
        ((flow_name,), {'key': 'requested', 'rpconfig': None}),
        ((flow_name,), dict(rpconfig=None, key=key)),
    ]


class BaseTestCase(SSOBaseTestCase):

    email = 'mark@example.com'
    new_email = 'person@example.com'

    def setUp(self):
        super(BaseTestCase, self).setUp()

        p = patch('webui.views.ui.stats.increment')
        self.mock_increment = p.start()
        self.addCleanup(p.stop)

        self.data = {'email': self.email,
                     'password': DEFAULT_USER_PASSWORD}
        self.account = self.factory.make_account(**self.data)

    def post_new_account(
            self,
            email=None,
            token=None,
            follow=False,
            captcha=True):
        if email is None:
            email = self.new_email
        if token is None:
            url = reverse('new_account')
        else:
            url = reverse('new_account', kwargs=dict(token=token))
        data = {
            'displayname': DISPLAYNAME,
            'email': email,
            'password': 'Testing123',
            'passwordconfirm': 'Testing123',
            'accept_tos': True
        }
        if captcha:
            data['recaptcha_challenge_field'] = 'ignored'
            data['recaptcha_response_field'] = 'ignored'

        return self.client.post(url, data, follow=follow)

    def _token_url(self):
        token_string = self.mock_send_email.call_args[0][2]['token']
        token = AuthToken.objects.get(token=token_string)
        return token.absolute_url

    def authenticate(self):
        self.client.login(username=self.data['email'],
                          password=self.data['password'])

    def request_when_captcha_fails(self, url, data):
        class MockCaptcha(object):
            def __init__(self, *args):
                pass

            def verify(self, solution, ip_addr, email):
                self.message = 'no-challenge'
                return False

            @classmethod
            def new(cls, env):
                return cls()

        with patch.object(ui, 'Captcha', MockCaptcha):
            r = self.client.post(url, data)

        return r


class VerifyCaptureResponseTests(SSOBaseTestCase):

    @patch.object(Captcha, '_open')
    def test_uses_timeline_from_request(self, mock_open):
        mock_open.return_value = CaptchaResponse(200, StringIO('true\na'))
        request = Mock()
        timeline = Timeline()
        request.META = {'timeline.timeline': timeline}
        request.POST = {'recaptcha_challenge_field': 'captcha-id'}
        request.environ = {'REMOTE_ADDR': '127.0.0.1'}
        with patch.multiple(settings, DISABLE_CAPTCHA_VERIFICATION=False):
            ui._verify_captcha_response(None, request, None)
        self.assertEqual(1, len(timeline.actions))
        self.assertEqual('captcha-verify', timeline.actions[0].category)


class SpanishUIViewsTestCase(BaseTestCase):

    def setUp(self):
        super(SpanishUIViewsTestCase, self).setUp()
        language_code = settings.LANGUAGE_CODE
        settings.LANGUAGE_CODE = 'es'
        self.addCleanup(setattr, settings, 'LANGUAGE_CODE', language_code)

    def test_new_account_in_spanish(self):
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        r = self.post_new_account(email=self.email)

        self.assertEqual(r.status_code, 200)
        self.assertEqual(len(mail.outbox), 1)


class UIViewsBaseTestCase(BaseTestCase):

    def setUp(self):
        super(UIViewsBaseTestCase, self).setUp()
        self.mock_send_email = self._apply_patch(
            'identityprovider.emailutils.send_branded_email'
        )


class LoginTestCase(UIViewsBaseTestCase):

    def test_login_with_unicode_email(self):
        email = u'mark@客example.com'
        password = DEFAULT_USER_PASSWORD
        Account.objects.create_account(DISPLAYNAME, email, password)
        self.data['email'] = email.encode('utf-8')
        r = self.client.post(reverse('login'), self.data)

        self.assertRedirects(r, reverse('account-index'))

    def test_login_with_unicode_password(self):
        self.data['password'] = u'test客'.encode('utf-8')
        r = self.client.post(reverse('login'), self.data)

        self.assertFormError(r, 'form', None, LOGIN_INCORRECT)

    def test_login_with_next(self):
        self.data['next'] = reverse('account-edit')
        r = self.client.post(reverse('login'), self.data)
        self.assertRedirects(r, reverse('account-edit'))

    def test_login_when_account_inactive(self):
        _password = self.account.accountpassword.password

        for status, _ in AccountStatus._get_choices():
            if status == AccountStatus.ACTIVE:
                # skip as this is already tested by test_login_with_next
                continue

            self.account.status = status
            self.account.save()
            # reset password in case previous status was suspended, which
            # resets the password
            self.account.accountpassword.password = _password
            self.account.accountpassword.save()

            self.data['next'] = reverse('account-edit')
            r = self.client.post(reverse('login'), self.data)

            if status == AccountStatus.SUSPENDED:
                self.assertContains(r, 'Your account has been suspended. '
                                       'Please contact login support to '
                                       're-enable it')
            else:
                self.assertContains(r, 'Your account has been deactivated. '
                                       'To reactivate it, please reset your '
                                       'password')

    def test_login_without_next_with_token(self):
        token = 'a' * 16
        r = self.client.post(reverse('login', kwargs=dict(token=token)),
                             self.data)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(r['Location'], "http://testserver/%s/" % token)

    def test_login_without_next_nor_token(self):
        r = self.client.post(reverse('login'), self.data)
        self.assertRedirects(r, reverse('account-index'))

    @switches(LOGIN_BY_PHONE=False)
    def test_login_with_not_validated_email_permitted(self):
        self.account.emailaddress_set.create(
            email="mark-2@example.com", status=EmailStatus.NEW)
        self.data['email'] = 'mark-2@example.com'
        r = self.client.post(reverse('login'), self.data)
        self.assertRedirects(r, reverse('account-index'))

    @patch('webui.views.ui.get_rpconfig_from_request')
    def test_login_with_unverified_email_disabled_by_rpconfig(
            self, mock_get_rpconfig):

        rpconfig = OpenIDRPConfig(trust_root='http://localhost/foo',
                                  allow_unverified=False)
        mock_get_rpconfig.return_value = rpconfig

        # make account unverified
        self.account.emailaddress_set.all().update(status=EmailStatus.NEW)
        assert not self.account.is_verified
        self.data['email'] = 'mark@example.com'

        # use a token in the request to fake an openid request
        token = 'a' * 16
        with switches(LOGIN_BY_PHONE=False):
            r = self.client.post(reverse('login', kwargs={'token': token}),
                                 self.data)

        self.assertFormError(r, 'form', None, LOGIN_INCORRECT)

    def test_login_with_invalidated_email_error(self):
        email_obj = EmailAddress.objects.get(email="mark@example.com")
        email_obj.invalidate()

        r = self.client.post(reverse('login'), self.data)
        self.assertFormError(
            r, 'form', None,
            "This email address has been invalidated. "
            "Please contact login support.")

    def test_login_with_invalidated_email_error_when_email_readded(self):
        email_obj = EmailAddress.objects.get(email="mark@example.com")
        email_obj.invalidate()

        self.factory.make_account(email=self.email, password="password")
        r = self.client.post(reverse('login'), self.data)
        self.assertFormError(r, 'form', None, LOGIN_INCORRECT)

    def test_suspended_account_should_require_a_password_reset(self):
        """Checking regression in #620462."""
        self.account.suspend()

        # Try to log in using just suspended account
        response = self.client.post(
            reverse('login'), data=dict(email=self.email,
                                        password=DEFAULT_USER_PASSWORD))
        # This should produce an error
        suspended = ('<span class="error">Your account has been suspended. '
                     'Please contact login support to re-enable it</span>')
        self.assertContains(response, suspended)

        # re-enable the account
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        response = self.client.post(
            reverse('login'), data=dict(email=self.email,
                                        password=DEFAULT_USER_PASSWORD))

        # If the url is still the same, that means that the log in was
        # unsuccessful.
        self.assertContains(
            response, '<span class="error">%s</span>' % LOGIN_INCORRECT)

    def test_suspended(self):
        r = self.client.get(reverse('suspended'))
        self.assertTemplateUsed(r, 'account/suspended.html')

    def test_includes_create_account_form_for_u1_brand(self):
        r = self.client.get(reverse('login'))
        self.assertIn('create_account_form', r.context)

    def test_login_password_policy_problem_with_reset(self):
        validator = Mock()
        validator.match.return_value = True
        validator.validate.side_effect = ValidationError('THE REASON')

        VALIDATORS = 'identityprovider.validators._validators'
        with patch(VALIDATORS, [validator]):
            r = self.client.post(reverse('login'), self.data)

            # The form should redirect to reset password view,
            # not back to login
            self.assertEqual(r.status_code, 302)
            redirect_url = r'(/token/.*/\+resetpassword/%s)' % self.email
            url_match = re.search(redirect_url, r['location'])
            self.assertIsNotNone(url_match)
            reset_password_url = url_match.groups()[0]

            r = self.client.get(reset_password_url)
            self.assertContains(
                r, escape("Your password doesn't meet security constraints"))
            self.assertContains(r, 'THE REASON')

        # next request passes validation
        resp = self.client.post(reset_password_url, {
            'passwordconfirm': 'Foo12345', 'password': 'Foo12345'})
        self.assertRedirects(resp, reverse('account-index'))

    def test_login_password_policy_problem_with_need_password_reset(self):
        self.account.need_password_reset = True
        self.account.save()

        r = self.client.post(reverse('login'), self.data)

        # The form should redirect to reset password view,
        # not back to login
        self.assertEqual(r.status_code, 302)
        redirect_url = r'(/token/.*/\+resetpassword/%s)' % self.email
        url_match = re.search(redirect_url, r['location'])
        self.assertIsNotNone(url_match)
        reset_password_url = url_match.groups()[0]

        r = self.client.get(reset_password_url)
        self.assertContains(r, 'Your account is at risk')
        self.assertContains(
            r, 'there was a security incident on another website')

        # next request passes validation
        resp = self.client.post(reset_password_url, {
            'passwordconfirm': 'Foo12345', 'password': 'Foo12345'})
        self.assertRedirects(resp, reverse('account-index'))

    def test_login_password_policy_problem_reusing_token(self):
        AuthToken.objects.create(
            requester=self.account, email=self.email,
            requester_email=self.email, token='foobar',
            token_type=TokenType.PASSWORDRECOVERY)

        validator = Mock()
        validator.match.return_value = True
        validator.validate.side_effect = ValidationError('THE REASON')

        VALIDATORS = 'identityprovider.validators._validators'
        with patch(VALIDATORS, [validator]):
            r = self.client.post(reverse('login'), self.data)

        # The foobar token should have been used
        self.assertEqual(r.status_code, 302)
        regex = r'(/token/foobar/\+resetpassword/%s)' % self.email
        action_match = re.search(regex, r['location'])
        self.assertIsNotNone(action_match)

    def test_login_password_policy_problem_with_old_token(self):
        AuthToken.objects.create(
            requester=self.account, email=self.email,
            requester_email=self.email, token='foobar',
            date_consumed=date.today(), token_type=TokenType.PASSWORDRECOVERY)

        validator = Mock()
        validator.match.return_value = True
        validator.validate.side_effect = ValidationError('THE REASON')

        VALIDATORS = 'identityprovider.validators._validators'
        with patch(VALIDATORS, [validator]):
            r = self.client.post(reverse('login'), self.data)

        # The foobar token should *not* have been used
        self.assertEqual(r.status_code, 302)
        regex = r'(/token/(.*)/\+resetpassword/%s)' % self.email
        action_match = re.search(regex, r['location'])
        self.assertIsNotNone(action_match)
        self.assertNotEquals(action_match.groups()[1], 'foobar')


class LogoutTestCase(UIViewsBaseTestCase):

    def test_logout(self):
        self.authenticate()
        r = self.client.get(reverse('logout'))
        self.assertTemplateUsed(r, 'registration/logout.html')

    def test_logout_preserve_token(self):
        self.authenticate()
        token = 'a' * 16
        # This strange dance with session is necessary to overcome way in which
        # test.Client returns session (recreating it on every Client.session
        # access)
        session = self.client.session
        request = {'openid.mode': 'checkid_setup',
                   'openid.trust_root': 'http://localhost/',
                   'openid.return_to': 'http://localhost/',
                   'openid.identity': IDENTIFIER_SELECT}
        openid_server = server._get_openid_server()
        orequest = openid_server.decodeRequest(request)
        expected = signed.dumps(orequest, settings.SECRET_KEY)
        session[token] = expected
        session.save()

        self.client.get(reverse('logout', kwargs=dict(token=token)))
        self.assertEqual(expected, self.client.session.get(token))

    def test_logout_displays_ga_snippet(self):
        self.authenticate()
        rpconfig = OpenIDRPConfig.objects.create(
            trust_root='https://www.example.com/openid/',
            ga_snippet='[["_setAccount", "12345"]]')

        token = self._assign_token_to_rpconfig(rpconfig)
        response = self.client.get(reverse('logout', kwargs=dict(token=token)))
        self.assertContains(response, "_gaq.push(['_setAccount', '12345']);")

    def test_claim_token_for_password_recovery_no_preferredemail(self):
        for email_obj in self.account.emailaddress_set.all():
            email_obj.status = EmailStatus.NEW
            email_obj.save()

        self.client.post(reverse('forgot_password'), {'email': self.email})

        # we send an email anyway despite there not being a validated email
        self.assertEqual(self.mock_send_email.call_count, 1)


class EnterTokenTestCase(UIViewsBaseTestCase):

    def setUp(self):
        super(EnterTokenTestCase, self).setUp()
        self.data = {'email': self.email}
        # patch ratelimit_extra to avoid "Rate limit exceeded" error
        mock_should_ratelimit = self._apply_patch(
            'webui.views.ui.ratelimit_extra.should_ratelimit')
        mock_should_ratelimit.return_value = False

    def test_enter_unexisting_token(self):
        r = self.client.post(reverse('enter_token'), {
            'email': 'fake@example.com',
            'confirmation_code': '0'})
        self.assertContains(r, 'Unknown confirmation code')

    def test_claim_unexisting_token(self):
        r = self.client.get(
            reverse('claim_token', kwargs=dict(authtoken=0)), self.data)
        self.assertEqual(r.status_code, 404)

    def test_claim_token_for_unexisting_token_type(self):
        token = self.factory.make_authtoken(email='fake@example.com')
        token.token_type = 9999
        token.save()
        r = self.client.get(
            reverse('claim_token', kwargs=dict(authtoken=token.token)),
            self.data)
        self.assertEqual(r.status_code, 404)

    def test_claim_consumed_token(self):
        self.client.post(reverse('forgot_password'), {'email': self.email})
        url = self._token_url()

        pwd = 'Password1'
        r = self.client.post(url, {'password': pwd, 'passwordconfirm': pwd})
        self.assertRedirects(r, reverse('account-index'))

        self.client.cookies.clear()

        pwd = 'Password2'
        r = self.client.post(url, {'password': pwd, 'passwordconfirm': pwd})
        self.assertRedirects(r, reverse('bad_token'))

    def test_token_form_validation(self):
        r = self.client.post(reverse('enter_token'))
        self.assertContains(r, '<span class="error"')


class NewEmailTestCase(UIViewsBaseTestCase):

    def test_config_email(self):
        self.authenticate()

        redirection_url = "http://example.com/foo/bar/baz"
        token = self.factory.make_authtoken(
            token_type=TokenType.VALIDATEEMAIL, email=self.email,
            redirection_url=redirection_url)

        r = self.client.post(token.absolute_url, {'post': 'yes'})
        self.assertEqual(r.status_code, 302)
        self.assertEqual(r['Location'], redirection_url)

    def test_config_email_no_redirection_url(self):
        self.authenticate()

        token = self.factory.make_authtoken(
            token_type=TokenType.VALIDATEEMAIL, email=self.email,
            redirection_url=None)

        r = self.client.post(token.absolute_url, {'post': 'yes'})
        self.assertEqual(r.status_code, 302)
        self.assertEqual(
            r['Location'],
            'http://testserver' + reverse('account-index')
        )

    def test_bad_token(self):
        r = self.client.get(reverse('bad_token'))
        self.assertEqual(r.status_code, 200)

    def test_non_field_errors_is_not_in_html(self):
        r = self.client.post(reverse('enter_token'), {
            'email': 'fake@example.com',
            'confirmation_code': '0'})
        # non field error section immediately follows a <p>
        self.assertNotContains(r, '<p><span class="error"')
        self.assertContains(r, '<span class="error"')

    def test_logout_to_confirm(self):
        r = self.client.get(reverse('logout_to_confirm'))
        self.assertEqual(r.status_code, 200)

    def test_logout_to_confirm_use_request_context(self):
        response = self.client.get(reverse('logout_to_confirm'))
        msg = ('You have attempted to confirm an account while logged in as '
               'someone else.')
        self.assertContains(response, msg)


class ForgotPasswordTestCase(LogHandlerTestCase, UIViewsBaseTestCase):

    def setUp(self):
        super(ForgotPasswordTestCase, self).setUp()
        # patch ratelimit_extra to avoid "Rate limit exceeded" error
        mock_should_ratelimit = self._apply_patch(
            'webui.views.ui.ratelimit_extra.should_ratelimit')
        mock_should_ratelimit.return_value = False

    def test_forgot_password_includes_rp_analytics(self):
        rpconfig = OpenIDRPConfig.objects.create(
            trust_root='http://example.com/',
            ga_snippet='[["_setAccount", "12345"]]')
        token = self._assign_token_to_rpconfig(rpconfig)
        response = self.client.get(
            reverse('forgot_password', kwargs=dict(token=token)))
        self.assertContains(response, "_gaq.push(['_setAccount', '12345']);")

    def test_forgot_password_email_when_account_active(self):
        self.client.post(reverse('forgot_password'), {'email': self.email})

        self.assertEqual(self.mock_send_email.call_count, 1)

    def test_forgot_password_email_when_account_deactivated(self):
        """ Deactivated accounts can be reactivated using the reset password
        functionality. Bug #556878 """
        # make sure the account is deactivated
        self.account.status = AccountStatus.DEACTIVATED
        self.account.save()

        self.client.post(reverse('forgot_password'), {'email': self.email})

        self.assertEqual(self.mock_send_email.call_count, 1)
        self.assertEqual(self.account.preferredemail.email, self.email)

    def test_forgot_password_email_when_account_noaccount(self):
        # make sure the account is deactivated
        self.account.status = AccountStatus.NOACCOUNT
        self.account.save()

        self.client.post(reverse('forgot_password'), {'email': self.email})

        self.assertEqual(self.mock_send_email.call_count, 1)
        self.assertEqual(self.account.preferredemail.email, self.email)

    def test_forgot_password_email_when_account_suspended(self):
        # make sure the account is suspended
        self.account.suspend()

        self.client.post(reverse('forgot_password'), {'email': self.email})

        self.assertEqual(self.mock_send_email.call_count, 0)

    def test_forgot_password_invalid_form(self):
        r = self.client.post(reverse('forgot_password'))
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'registration/forgot_password.html')
        self.assertFormError(r, 'form', 'email', REQUIRED_FIELD)

    def test_forgot_password_with_token(self):
        token1 = create_token(16)
        self.client.post(
            reverse('forgot_password', kwargs=dict(token=token1)),
            {'email': self.email}
        )
        # should be only one
        auth_token = AuthToken.objects.get(
            email=self.email, token_type=TokenType.PASSWORDRECOVERY)
        self.assertEqual(auth_token.redirection_url,
                         reverse('server-decide', kwargs=dict(token=token1)))

    def test_reset_password_when_account_active(self):
        self.client.post(reverse('forgot_password'), {'email': self.email})

        # make sure the account is active
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))

    def test_reset_password_default_action(self):
        self.client.post(reverse('forgot_password'), {'email': self.email})

        # make sure the account is active
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        # confirm account
        data = {'password': 'Password3', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertContains(r, ' action=""')

    def test_reset_password_email_is_case_insensitive(self):
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # update requester email in token
        auth_token = AuthToken.objects.get(
            email=self.email, token_type=TokenType.PASSWORDRECOVERY)
        auth_token.requester_email = self.email.upper()
        assert self.email.upper() != self.email
        auth_token.save()

        # make sure the account is active
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))

    def test_reset_password_email_not_found(self):
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # update requester email in token
        auth_token = AuthToken.objects.get(
            email=self.email, token_type=TokenType.PASSWORDRECOVERY)
        auth_token.requester_email = 'another-email@example.com'
        auth_token.save()

        # make sure the account is active
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}

        mock_is_enabled = Mock(return_value=True)
        with patch.object(self.root_logger, 'isEnabledFor', mock_is_enabled):
            with patch('webui.views.ui.logger', self.root_logger):
                r = self.client.post(self._token_url(), data)

        self.assertRedirects(r, reverse('bad_token'))
        self.assertLogLevelContains(
            'WARNING', "Could not reset password, invalid email address "
            "provided in the token: %s (account %s)" % (
                auth_token.requester_email, self.account.id))

    def test_reset_password_with_lowered_password_requirements(self):
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # make sure the account is active
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        # confirm account
        data = {'password': 'password', 'passwordconfirm': 'password'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))

        account = Account.objects.get(pk=self.account.pk)
        # check password changed
        self.assertTrue(
            check_password('password', account.encrypted_password))

    def test_reset_password_when_account_active_no_password(self):
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        self.account.accountpassword.delete()
        # make sure the account is active
        self.account.status = AccountStatus.ACTIVE
        self.account.save()

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))

    def test_reset_password_when_account_deactivated(self):
        """Prevent password reset for unverified emails."""
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # make sure the account is deactivated
        self.account.status = AccountStatus.DEACTIVATED
        self.account.save()

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))
        account = Account.objects.get(id=self.account.id)
        self.assertEqual(account.status, AccountStatus.ACTIVE)
        self.assertEqual(
            account.preferredemail.status, EmailStatus.PREFERRED)

    def test_reset_password_when_account_deactivated_but_notverified(self):
        """Prevent password reset for unverified emails."""
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # make sure the account is deactivated
        self.account.status = AccountStatus.DEACTIVATED
        self.account.save()
        self.account.emailaddress_set.update(status=EmailStatus.NEW)

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))
        account = Account.objects.get(id=self.account.id)
        self.assertEqual(account.status, AccountStatus.ACTIVE)
        self.assertEqual(
            account.preferredemail.status, EmailStatus.PREFERRED)

    def test_reset_password_when_account_deactivated_no_preferred_email(self):
        # make sure the account is deactivated and preferred email is removed
        self.account.status = AccountStatus.DEACTIVATED
        # Blow out the cached value
        self.account.save()
        self.account.emailaddress_set.update(status=EmailStatus.NEW)

        self.account = Account.objects.get_by_email(self.email)

        self.assertEqual(self.account.preferredemail.email, self.email)

        self.client.post(reverse('forgot_password'), {'email': self.email})

        # we sent an email anyway, despite there not being a validated email
        self.assertEqual(self.mock_send_email.call_count, 1)

    def test_reset_password_when_account_no_verified_emails(self):
        self.account.emailaddress_set.update(status=EmailStatus.NEW)
        self.client.post(reverse('forgot_password'), {'email': self.email})
        # we sent an email anyway, despite there not being a validated email
        self.assertEqual(self.mock_send_email.call_count, 1)

    def test_reset_password_unverified_email_but_account_verified_email(self):
        verified = self.account.verified_emails()
        unverified = self.factory.make_email_for_account(
            self.account, status=EmailStatus.NEW)
        self.client.post(reverse('forgot_password'),
                         {'email': unverified.email})
        # email sent to verified email addresses instead
        self.assertEqual(self.mock_send_email.call_count, len(verified))

    def test_reset_password_when_account_noaccount(self):
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # make sure the account is deactivated
        self.account.status = AccountStatus.NOACCOUNT
        self.account.save()

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))

    def test_reset_password_when_account_suspended(self):
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        self.account.suspend()

        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}

        # confirm account
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('bad_token'))

    def test_reset_password_invalid_form(self):
        # get valid session
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # test view
        r = self.client.post(self._token_url())
        self.assertTemplateUsed(r, 'registration/reset_password.html')
        self.assertFormError(r, 'form', 'password', REQUIRED_FIELD)

    def test_reset_password_get(self):
        # get valid session
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # test view
        r = self.client.get(self._token_url())
        self.assertTemplateUsed(r, 'registration/reset_password.html')
        for context in r.context:
            self.assertEqual(context['form'].errors, {})

    def test_bad_method(self):
        r = self.client.put(reverse('forgot_password'), {'email': self.email})
        self.assertEqual(r.status_code, 405)

    def test_reset_password_invalidate_tokens(self):
        # make sure the account is active
        assert self.account.status == AccountStatus.ACTIVE

        token = self.account.create_oauth_token('new-token')
        r = self.client.post(reverse('forgot_password'), {'email': self.email})

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))
        self.assertNotIn(token, self.account.oauth_tokens())

    def test_reset_password_no_redirection_url(self):
        self.client.post(reverse('forgot_password'), {'email': self.email})
        auth_token = AuthToken.objects.get(
            email=self.email, token_type=TokenType.PASSWORDRECOVERY)
        auth_token.redirection_url = None
        auth_token.save()

        # confirm account
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertRedirects(r, reverse('account-index'))

    def test_reset_password_leaked_password_used(self):
        self.client.post(reverse('forgot_password'), {'email': self.email})

        # make sure the account is active
        self.account.status = AccountStatus.ACTIVE
        self.account.save()
        # create a leaked credential
        self.factory.make_leaked_credential(email=self.email,
                                            password='Password1')

        # verify current password is not the one we're trying to set
        self.assertFalse(check_password('Password1',
                                        self.account.encrypted_password))
        # attempt to submit a leaked password
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        r = self.client.post(self._token_url(), data)
        self.assertFormError(r, 'form', 'password', [PASSWORD_LEAKED])
        self.assertContains(r, 'there was a security incident')
        # refresh account from db
        account = Account.objects.get(id=self.account.id)
        # verify password hasn't changed
        self.assertFalse(check_password('Password1',
                                        account.encrypted_password))


class LoggedInForgotPasswordTestCase(ForgotPasswordTestCase):
    # forgot_password and reset_password views should also work
    # for logged in users, repeat tests with an authenticated user

    def setUp(self):
        super(LoggedInForgotPasswordTestCase, self).setUp()
        self.authenticate()


class ConfirmEmailTestCase(UIViewsBaseTestCase):

    def _setup_account_with_new_email(self, account=None):
        if account is None:
            account = self.account

        # create new email account
        email = self.factory.make_email_for_account(
            account, email=self.new_email, status=EmailStatus.NEW)
        # create verification token
        token = self.factory.make_authtoken(
            token_type=TokenType.VALIDATEEMAIL, email=email.email)
        return email, token

    def get_email_token_validation_response(self, account=None):
        email, token = self._setup_account_with_new_email(account)
        response = self.client.get(token.absolute_url)
        return response, token

    def test_confirm_email_not_existing(self):
        self.authenticate()
        email, token = self._setup_account_with_new_email()
        # delete email account
        email.delete()
        # use verification token
        r = self.client.get(token.absolute_url)
        self.assertEqual(r.status_code, 404)

    def test_confirm_email_get(self):
        self.authenticate()

        email, token = self._setup_account_with_new_email()
        # use verification token
        self.client.get(token.absolute_url)

        # verify token has not been consumed, and email has not been
        # verified yet
        token = AuthToken.objects.get(
            token=token.token, token_type=TokenType.VALIDATEEMAIL)
        self.assertEqual(token.date_consumed, None)
        email = EmailAddress.objects.get(email=email.email)
        self.assertEqual(email.status, EmailStatus.NEW)

    def test_confirm_email_as_another_user_fails(self):
        self.authenticate()
        account = self.factory.make_account()

        r, token = self.get_email_token_validation_response(account)

        self.assertEqual(r.status_code, 404)
        self.assertEqual(0, AuthToken.objects.filter(id=token.id).count())
        email = EmailAddress.objects.get(email=self.new_email)
        self.assertEqual(EmailStatus.NEW, email.status)

    def test_confirm_email_when_not_logged_in_redirects_to_login(self):
        r, token = self.get_email_token_validation_response(self.account)

        kwargs = dict(authtoken=token, email_address=self.new_email)
        url = (reverse('login') + '?next=' +
               urllib2.quote(reverse('confirm_email', kwargs=kwargs)))
        self.assertRedirects(r, url)

    def test_confirm_email_when_not_logged_in_shows_message(self):
        email, token = self._setup_account_with_new_email()
        response = self.client.get(token.absolute_url, follow=True)
        self.assertContains(
            response, 'Please log in to use this confirmation code')

    def test_invalid_email_does_not_blow_up(self):
        r = self.post_new_account(email='what<~@ever.com')
        self.assertEqual(200, r.status_code)
        self.assertContains(r, INVALID_EMAIL)


class CaptchaVerificationTestCase(BaseTestCase):
    success_status = 302

    def setUp(self):
        super(CaptchaVerificationTestCase, self).setUp()
        mock_handler = MockHandler()
        mock_handler.set_next_response('false\nno-challenge')
        Captcha.opener = urllib2.build_opener(mock_handler)
        self.addCleanup(setattr, Captcha, 'opener', None)

        old_disable = settings.DISABLE_CAPTCHA_VERIFICATION
        settings.DISABLE_CAPTCHA_VERIFICATION = False
        self.addCleanup(setattr, settings, 'DISABLE_CAPTCHA_VERIFICATION',
                        old_disable)

    def test_new_account_when_form_validation_fails(self):
        r = self.post_new_account()
        self.assertTemplateUsed(r, 'registration/new_account.html')
        msg = 'It appears that our captcha service was unable to load'
        self.assertContains(r, msg)

    def test_new_account_captcha_whitelist(self):
        email = 'canonicaltest@gmail.com'
        pattern = '^canonicaltest(?:\+.+)?@gmail\.com$'
        overrides = dict(
            DISABLE_CAPTCHA_VERIFICATION=False,
            EMAIL_WHITELIST_REGEXP_LIST=[pattern],
        )
        with patch_settings(**overrides):
            response = self.post_new_account(email=email)
            self.assertEqual(response.status_code, self.success_status)

    def test_new_account_captcha_whitelist_with_uuid(self):
        email = 'canonicaltest+something@gmail.com'
        pattern = '^canonicaltest(?:\+.+)?@gmail\.com$'
        overrides = dict(
            DISABLE_CAPTCHA_VERIFICATION=False,
            EMAIL_WHITELIST_REGEXP_LIST=[pattern],
        )
        with patch_settings(**overrides):
            response = self.post_new_account(email=email)
            self.assertEqual(response.status_code, self.success_status)

    @switches(CAPTCHA=True)
    def test_new_account_captcha_whitelist_fail(self):
        email = 'notcanonicaltest@gmail.com'
        pattern = '^canonicaltest(?:\+.+)?@gmail\.com$'
        overrides = dict(
            DISABLE_CAPTCHA_VERIFICATION=False,
            EMAIL_WHITELIST_REGEXP_LIST=[pattern],
        )
        with patch_settings(**overrides):
            response = self.post_new_account(email=email)
            msg = 'It appears that our captcha service was unable to load'
            self.assertContains(response, msg)


class CookiesTestCase(SSOBaseTestCase):

    def setUp(self):
        super(CookiesTestCase, self).setUp()
        _disable_cookie_check = decorators.disable_cookie_check
        decorators.disable_cookie_check = False
        Session.objects.all().delete()
        self.addCleanup(setattr, decorators, 'disable_cookie_check',
                        _disable_cookie_check)

    def _get(self, url, *params):
        r = self.client.get(url)
        # The following line makes this cookieless:
        self.client.cookies.clear()
        if 300 <= r.status_code < 400:
            location = r['Location']
            scheme, netloc, path, query, fragment = urlsplit(location)
            return self._get(path, QueryDict(query))
        else:
            return r

    def _check_url(self, url):
        r = self._get(url)
        self.assertContains(r, '<title>Cookies required</title>')

        r = self.client.get(url, follow=True)
        self.assertNotContains(r, '<title>Cookies required</title>')

    def _count_sessions(self):
        return Session.objects.all().count()

    def test_login(self):
        self._check_url(reverse('login'))

    def test_account_index(self):
        url = reverse('account-index')

        r = self.client.get(url)

        self.assertEqual(r.status_code, 200)
        self.assertNotContains(r, '<title>Cookies required</title>')

    def test_forgot_password(self):
        self._check_url(reverse('forgot_password'))

    def test_new_account(self):
        self._check_url(reverse('new_account'))

    def test_cookie_is_sessionless(self):
        self.client.get(reverse('login'))
        self.assertEqual(self._count_sessions(), 0)

    def test_cookie_redirects_using_relative_path_only(self):
        path = '/path;part1;part2?q1=a&q2=b'
        next_url = urllib.quote_plus('http://example.com' + path)

        # testing indirectly via a decorated view, in order
        # to test the normal use case
        url = reverse('forgot_password') + '?next=' + next_url
        # trigger cookie check
        response = self.client.get(url)
        # view redirects to view for checking cookie support
        self.assertEqual(response.status_code, 302)
        self.assertIn('+cookie', response['location'])
        # follow redirect to next url
        response = self.client.get(response['location'])
        self.assertEqual(response.status_code, 302)
        # django will always build an absolte url when returning a 302 response
        # since this is being generated as part of the test run, the full
        # url will be http://testserver/...
        expected_url = 'http://testserver%s?next=%s' % (
            reverse('forgot_password'), urllib.quote_plus(path))
        self.assertEqual(response['location'], expected_url)

    def test_cookie_redirects_when_no_query_string(self):
        # testing indirectly via a decorated view, in order
        # to test the normal use case
        url = reverse('new_account')
        # trigger cookie check
        response = self.client.get(url)
        # view redirects to view for checking cookie support
        self.assertEqual(response.status_code, 302)
        self.assertIn('+cookie', response['location'])
        # follow redirect to next url
        response = self.client.get(response['location'])
        self.assertEqual(response.status_code, 302)
        # django will always build an absolte url when returning a 302 response
        # since this is being generated as part of the test run, the full
        # url will be http://testserver/...
        expected_url = 'http://testserver%s' % url
        self.assertEqual(response['location'], expected_url)


class LoginFlowStatsTestCase(BaseTestCase):

    args_list = partial(_args_list, flow_name='flows.login')

    def test_login_stats(self):
        self.client.get(reverse('login'))
        self.client.post(reverse('login'), self.data)
        self.assertEqual(self.mock_increment.call_args_list, self.args_list())

    def test_login_error_stats(self):
        self.client.get(reverse('login'))
        # make sure to pass invalid data
        self.client.post(reverse('login'), {'account': 'invalid'})
        self.assertEqual(self.mock_increment.call_args_list,
                         self.args_list(key='error'))


class ForgotPasswordFlowStatsTestCase(UIViewsBaseTestCase):

    args_list = partial(_args_list, flow_name='flows.forgot_password')

    def setUp(self):
        super(ForgotPasswordFlowStatsTestCase, self).setUp()
        # patch ratelimit_extra to avoid "Rate limit exceeded" error
        mock_should_ratelimit = self._apply_patch(
            'webui.views.ui.ratelimit_extra.should_ratelimit')
        mock_should_ratelimit.return_value = False

    def forgot_password(self, data=None):
        self.client.get(reverse('forgot_password'))
        if data is None:
            data = {'email': self.email}
        self.client.post(reverse('forgot_password'), data)

    def reset_password(self, valid_password=True):
        data = {'password': 'Password1', 'passwordconfirm': 'Password1'}
        if not valid_password:
            data['passwordconfirm'] = 'invalid'
        self.client.post(self._token_url(), data)

    def test_forgot_password(self):
        self.forgot_password()
        self.reset_password()
        self.assertEqual(self.mock_increment.call_args_list, self.args_list())

    def test_forgot_password_form_error(self):
        self.forgot_password(data={'foo': 'invalid'})
        self.assertEqual(self.mock_increment.call_args_list,
                         self.args_list(key='error.form'))

    @patch('webui.views.ui.verify_token_string')
    def test_reset_password_token_error(self, mock_verify_token_string):
        mock_verify_token_string.return_value = None
        self.forgot_password()
        self.reset_password()
        self.assertEqual(self.mock_increment.call_args_list,
                         self.args_list(key='error.token'))

    def test_reset_password_form_error(self):
        self.forgot_password()
        self.reset_password(valid_password=False)
        self.assertEqual(self.mock_increment.call_args_list,
                         self.args_list(key='error.form'))


class BaseTwoFactorLoginTestCase(SSOBaseTestCase):
    email = 'user2f@example.com'
    password = DEFAULT_USER_PASSWORD
    logout_link = 'a[href="/+logout"]'
    url = reverse('twofactor')

    def setUp(self):
        super(BaseTwoFactorLoginTestCase, self).setUp()
        self.account = self.factory.make_account('2f User', self.email,
                                                 self.password)
        self.switch = switches(TWOFACTOR=True)
        self.switch.__enter__()
        self.addCleanup(self.switch.__exit__, None, None, None)

    def do_login(self, **kwargs):
        data = {'email': self.email, 'password': self.password}
        token = kwargs.pop('token', None)
        data.update(kwargs)
        if token:
            url = reverse('login', kwargs={'token': token})
        else:
            url = reverse('login')
        return self.client.post(url, data)

    def do_twofactor(self, oath_token, next=None):
        data = {'oath_token': oath_token}
        if next:
            data['next'] = next
        return self.client.post(self.url, data)


class TwoFactorLoginTestCase(BaseTwoFactorLoginTestCase):

    def test_twofactor_login_disabled_when_readonly(self):
        rm = ReadOnlyManager()
        rm.set_readonly()
        self.addCleanup(rm.clear_readonly)

        with switches(TWOFACTOR=True):
            self.do_login()
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 404)

    @patch('webui.views.ui.twofactor.site_requires_twofactor_auth')
    def test_when_site_does_not_require_twofactor_no_oath_field(
            self, mock_site):
        mock_site.return_value = False
        response = self.client.get(reverse('login'))
        tree = PyQuery(response.content)
        self.assertEqual(len(tree.find('input[name="oath_token"]')), 0)

    @patch('webui.views.ui.twofactor.site_requires_twofactor_auth')
    def test_when_site_requires_twofactor_oath_field_is_present(
            self, mock_site):
        mock_site.return_value = True
        response = self.client.get(reverse('login'))
        tree = PyQuery(response.content)
        inputs = tree.find('input[name="oath_token"]')
        self.assertEqual(len(inputs), 1)
        oath = inputs[0]
        self.assertEqual(oath.attrib['autocomplete'], 'off')

    @patch('webui.views.ui.twofactor.site_requires_twofactor_auth')
    def test_when_site_requires_twofactor_oath_field_is_present_lp(
            self, mock_site):
        mock_site.return_value = True
        response = self.client.get(reverse('login'))
        tree = PyQuery(response.content)
        inputs = tree.find('input[name="oath_token"]')
        self.assertEqual(len(inputs), 1)
        oath = inputs[0]
        self.assertEqual(oath.attrib['autocomplete'], 'off')

    @patch('webui.views.ui.twofactor.user_requires_twofactor_auth')
    def test_when_user_requires_twofactor_redirected(self, mock_user):
        mock_user.return_value = True
        r = self.do_login()
        self.assertRedirects(r, self.url)

    @patch('webui.views.ui.twofactor.user_requires_twofactor_auth')
    def test_when_user_requires_twofactor_redirected_with_token(
            self, mock_user):
        mock_user.return_value = True
        token = 'a' * 16
        r = self.do_login(token=token)
        self.assertRedirects(r, reverse('twofactor', kwargs=dict(token=token)))

    @patch('webui.views.ui.twofactor.user_requires_twofactor_auth')
    def test_when_user_requires_twofactor_redirected_with_next(
            self, mock_user):
        mock_user.return_value = True
        r = self.do_login(next=reverse('account-edit'))
        self.assertRedirects(r, '/two_factor_auth?next=/%2Bedit')

    @patch('webui.views.ui.twofactor.site_requires_twofactor_auth')
    @patch('webui.views.ui.authenticate')
    def test_full_twofactor_login_redirected_to_next(
            self, mock_auth, mock_site):
        mock_auth.return_value = True
        mock_site.return_value = True
        r = self.do_login(next=reverse('account-edit'), oath_token='123456')
        mock_auth.assert_called_once_with(self.account, '123456')
        self.assertRedirects(r, reverse('account-edit'))

    @patch('webui.views.ui.twofactor.site_requires_twofactor_auth')
    def test_full_twofactor_login_illformed_token_displays_error(
            self, mock_site):
        mock_site.return_value = True
        response = self.do_login(next=reverse('account-edit'),
                                 oath_token='XXX')
        tree = PyQuery(response.content)
        self.assertEqual(len(tree.find('input[name="oath_token"]')), 1)
        error = tree.find('#oathtoken span.error')
        self.assertEqual(len(error), 1)
        self.assertEqual(error[0].text.strip(), ONE_TIME_PWD_ERROR)

    @patch('webui.views.ui.twofactor.site_requires_twofactor_auth')
    @patch('webui.views.ui.authenticate')
    def test_full_twofactor_login_invalid_code_displays_error(
            self, mock_auth, mock_site):
        mock_auth.side_effect = ui.AuthenticationError("ERRORMSG")
        mock_site.return_value = True
        response = self.do_login(next=reverse('account-edit'),
                                 oath_token='123456')
        tree = PyQuery(response.content)
        self.assertEqual(len(tree.find('input[name="oath_token"]')), 1)
        error = tree.find('#oathtoken span.error')
        self.assertEqual(len(error), 1)
        self.assertEqual(error[0].text.strip(), "ERRORMSG")
        mock_auth.assert_called_once_with(self.account, '123456')

    @patch('webui.views.ui.twofactor.site_requires_twofactor_auth')
    @patch('webui.views.ui.authenticate')
    def test_full_twofactor_login_invalid_code_displays_unicode_error(
            self, mock_auth, mock_site):
        mock_auth.side_effect = ui.AuthenticationError(u"客")
        mock_site.return_value = True
        response = self.do_login(next=reverse('account-edit'),
                                 oath_token='123456')
        tree = PyQuery(response.content.decode('utf8'))
        self.assertEqual(len(tree.find('input[name="oath_token"]')), 1)
        error = tree.find('#oathtoken span.error')
        self.assertEqual(len(error), 1)
        self.assertTrue(error[0].text == u"客")

    @patch('webui.views.ui.authenticate')
    def test_twofactor_step_redirected_to_next(self, mock_auth):
        mock_auth.return_value = True
        self.client.login(username=self.email, password=self.password)
        r = self.do_twofactor(next=reverse('account-edit'),
                              oath_token='123456')
        self.assertRedirects(r, reverse('account-edit'))
        mock_auth.assert_called_once_with(self.account, '123456')

    def test_twofactor_step_illformed_token_displays_error(self):
        self.client.login(username=self.email, password=self.password)
        response = self.do_twofactor(next=reverse('account-edit'),
                                     oath_token='XXX')
        tree = PyQuery(response.content)
        self.assertEqual(len(tree.find('input[name="oath_token"]')), 1)
        error = tree.find('#oathtoken span.error')
        self.assertEqual(len(error), 1)
        self.assertEqual(error[0].text.strip(), ONE_TIME_PWD_ERROR)

    @patch('webui.views.ui.authenticate')
    def test_twofactor_step_invalid_code_displays_error(self, mock_auth):
        mock_auth.side_effect = ui.AuthenticationError("ERRORMSG")
        self.client.login(username=self.email, password=self.password)
        response = self.do_twofactor(next=reverse('account-edit'),
                                     oath_token='123456')
        tree = PyQuery(response.content)
        self.assertEqual(len(tree.find('input[name="oath_token"]')), 1)
        error = tree.find('#oathtoken span.error')
        self.assertEqual(len(error), 1)
        self.assertEqual(error[0].text.strip(), "ERRORMSG")
        mock_auth.assert_called_once_with(self.account, '123456')

    @patch('webui.views.ui.twofactor.user_requires_twofactor_auth')
    def test_two_factor_includes_rp_analytics(self, mock_user_requires_2f):
        mock_user_requires_2f.return_value = True
        rpconfig = OpenIDRPConfig.objects.create(
            trust_root='http://example.com/',
            ga_snippet='[["_setAccount", "12345"]]')
        token = 'averyrandomtoken'
        response = self.do_login(token=token)
        self._assign_token_to_rpconfig(rpconfig, token=token)
        url = reverse('twofactor', kwargs=dict(token=token))
        self.assertRedirects(response, url)

        response = self.client.get(url)

        self.assertContains(response, "_gaq.push(['_setAccount', '12345']);")

    def test_twofactor_provides_logout(self):
        self.client.login(username=self.email, password=self.password)
        response = self.client.get(self.url)

        tree = PyQuery(response.content)
        logout_link = tree.find(self.logout_link)
        self.assertTrue(len(logout_link) > 0)

    def test_show_user_details(self):
        self.client.login(username=self.email, password=self.password)
        response = self.client.get(self.url)

        self.assertContains(response, self.account.displayname)

    def test_doesnt_show_sync_link_on_get(self):
        self.client.login(username=self.email, password=self.password)
        response = self.client.get(self.url)

        self.assertNotContains(response, reverse('sync_2fa'))

    def test_shows_sync_link_when_otp_is_invalid(self):
        self.client.login(username=self.email, password=self.password)

        response = self.do_twofactor(oath_token='123456')

        self.assertContains(response, reverse('sync_2fa'))


class ResetPasswordLimitTestCase(SSOBaseTestCase):

    def request_password_reset(self, authtoken):
        url = reverse('reset_password', args=[authtoken, 'foo@foo.com'])
        response = self.client.get(url)
        return response

    @patch('webui.views.ui.ratelimit_extra.get_counters')
    def test_invalid_authtoken_below_threshold(self, mock_get_counters):
        mock_get_counters.return_value = {None: 1}

        authtoken = 'invalid'
        response = self.request_password_reset(authtoken)
        self.assertRedirects(response, '/+bad-token')

    @patch('webui.views.ui.ratelimit_extra.get_counters')
    def test_too_many_invalid_authtokens(self, mock_get_counters):
        mock_get_counters.return_value = {None: 20}

        authtoken = 'invalid'
        response = self.request_password_reset(authtoken)
        self.assertEqual(response.status_code, 403)


class TwoFactorLoginWarningFailuresTestCase(BaseTwoFactorLoginTestCase):

    def setUp(self):
        super(TwoFactorLoginWarningFailuresTestCase, self).setUp()
        # patch limitlogin to avoid "Rate limit exceeded" error
        self._apply_patch('webui.views.ui.limitlogin')
        self.team = 'foobarteam'
        p = patch_settings(
            TWOFACTOR_FAILURE_NOTIFICATION_THRESHOLD=2,
            TWOFACTOR_FAILURE_NOTIFICATION_EMAILS=['someone@somewhere.com'],
            TWOFACTOR_FAILURE_NOTIFICATION_TEAMS=[self.team])
        p.start()
        self.addCleanup(p.stop)

        assert self.client.login(username=self.email, password=self.password)
        assert self.account.twofactor_attempts == 0
        assert not self.account.person_in_team(self.team)

    def set_threshold_delta(self, delta):
        threshold = settings.TWOFACTOR_FAILURE_NOTIFICATION_THRESHOLD
        self.account.twofactor_attempts = threshold + delta
        assert self.account.twofactor_attempts >= 0
        self.account.save()

    def set_team_to_account(self, team):
        self.factory.add_account_to_team(self.account,
                                         self.factory.make_team(team))
        assert self.account.person_in_team(team)

    def fail_twofactor_form_invalid(self):
        # the following will ensure an invalid form
        response = self.do_twofactor(oath_token='foobar')
        errors = response.context['form'].errors
        assert errors == {
            'oath_token': ['Please enter a 6-digit or 8-digit one-time '
                           'password']
        }

        return response

    def fail_twofactor_token_invalid(self):
        # form is valid but oath token is not valid
        response = self.do_twofactor(oath_token='123456')
        errors = response.context['form'].errors
        assert errors == {'__all__': ['The password is invalid']}

        return response

    def assert_too_many_failures_warning_not_sent(self):
        self.assertEqual(len(mail.outbox), 0)

    def assert_too_many_failures_warning_sent(self, request):
        self.assertEqual(len(mail.outbox), 1)
        info = (self.account.twofactor_attempts, self.account.displayname,
                self.account.openid_identifier)
        body = mail.outbox[0].body
        for i in info:
            self.assertIn(str(i), body)
        for k, v in request.iteritems():
            self.assertIn(repr(k), body)
            self.assertIn(repr(v), body)

    def test_dont_notify_below_threshold_form_invalid(self):
        self.set_team_to_account(self.team)
        self.fail_twofactor_form_invalid()

        self.assert_too_many_failures_warning_not_sent()

    def test_dont_notify_below_threshold_auth_error(self):
        self.set_team_to_account(self.team)
        self.fail_twofactor_token_invalid()

        self.assert_too_many_failures_warning_not_sent()

    def test_team_mismatch_form_invalid(self):
        self.set_threshold_delta(-1)
        self.set_team_to_account('foobar')
        self.fail_twofactor_form_invalid()

        self.assert_too_many_failures_warning_not_sent()

    def test_team_mismatch_auth_error(self):
        self.set_threshold_delta(-1)
        self.set_team_to_account('foobar')
        self.fail_twofactor_token_invalid()

        self.assert_too_many_failures_warning_not_sent()

    def test_too_many_failures_form_invalid(self):
        self.set_threshold_delta(-1)
        self.set_team_to_account(self.team)
        response = self.fail_twofactor_form_invalid()

        self.assert_too_many_failures_warning_sent(request=response.request)

    def test_on_too_many_failures_auth_error(self):
        self.set_threshold_delta(-1)
        self.set_team_to_account(self.team)
        response = self.fail_twofactor_token_invalid()

        self.assert_too_many_failures_warning_sent(request=response.request)


class TwoFactorLoginLPTestCase(TwoFactorLoginTestCase):
    """This testcase inherits its tests from TwoFactorLoginTestCase in order to
    run exactly the same tests against the LP templates"""

    logout_link = 'form[action="/+logout"]'

    def setUp(self):
        super(TwoFactorLoginLPTestCase, self).setUp()
        brand_patch = patch_settings(BRAND='launchpad')
        brand_patch.start()
        self.addCleanup(brand_patch.stop)

    def test_lp_templates_render_correctly(self):
        """Test that the lp twofactor template renders properly"""
        self.do_login()
        response = self.client.get(self.url)
        self.assertNotIn('Email address', response.content)
        self.assertNotIn('Password', response.content)


class TwoFactorFunctionsTestCase(SSOBaseTestCase):

    def setUp(self):
        super(TwoFactorFunctionsTestCase, self).setUp()
        self.mock_logger = self._apply_patch(
            'identityprovider.models.twofactor.logger')

    def _test_user_requires_twofactor_auth(
            self, require2f, has_devices, returns):
        request = Mock()
        account = Mock(spec=Account)
        account.openid_identifier = 'abc1234'
        account.twofactor_required = require2f
        account.has_twofactor_devices.return_value = has_devices
        with patch(
                'webui.views.ui.twofactor.is_twofactor_enabled') as mock_flag:
            mock_flag.return_value = True

            r = twofactor.user_requires_twofactor_auth(request, account)
            self.assertEqual(returns, r)
            mock_flag.assert_called_once_with(request)

        return account

    def test_not_require_twofactor_auth_with_devices(self):
        self._test_user_requires_twofactor_auth(
            require2f=False, has_devices=True, returns=False)
        self.assertFalse(self.mock_logger.warning.called)

    def test_not_require_twofactor_auth_without_devices(self):
        self._test_user_requires_twofactor_auth(
            require2f=False, has_devices=False, returns=False)
        self.assertFalse(self.mock_logger.warning.called)

    def test_require_twofactor_auth_without_devices(self):
        account = self._test_user_requires_twofactor_auth(
            require2f=True, has_devices=False, returns=False)
        self.mock_logger.warning.assert_called_once_with(
            'Account %s requires two-factor but has no devices',
            account.openid_identifier)

    def test_require_twofactor_auth_with_devices(self):
        self._test_user_requires_twofactor_auth(
            require2f=True, has_devices=True, returns=True)
        self.assertFalse(self.mock_logger.warning.called)

    def _make_mock_account(self, *args):
        def _make_device(ret):
            m = Mock()
            m.authenticate.return_value = ret
            return m
        devices = [_make_device(a) for a in args]
        account = Mock()
        account.devices.order_by.return_value = devices
        return account, devices

    def test_authenticate_with_no_valid_devices_throws(self):
        account, _ = self._make_mock_account(False, False)
        self.assertRaises(ui.AuthenticationError, ui.authenticate,
                          account, '123456')

    def test_authenticate_with_first_device_succeeds(self):
        account, devices = self._make_mock_account(True, False)
        self.assertTrue(ui.authenticate(account, '123456'))
        self.assertTrue(devices[0].authenticate.called)
        self.assertFalse(devices[1].authenticate.called)

    def test_authenticate_with_last_device_succeeds(self):
        account, devices = self._make_mock_account(False, True)
        self.assertTrue(ui.authenticate(account, '123456'))
        self.assertTrue(devices[0].authenticate.called)
        self.assertTrue(devices[1].authenticate.called)


class ComboViewTestCase(SSOBaseTestCase):

    def test_combine_files(self):
        response = self.client.get(
            settings.COMBO_URL +
            '?yui/3.3.0/widget/widget-base.js'
            '&identityprovider/ubuntuone/js/ie/html5shiv.js')

        self.assertEqual(response['content-type'], 'text/javascript')
        self.assertNotContains(response, '[missing]')
        self.assertContains(response, '/* yui/3.3.0/widget/widget-base.js */')
        self.assertContains(
            response,
            '/* identityprovider/ubuntuone/js/ie/html5shiv.js */')


class StaticPageTestCase(SSOBaseTestCase):

    def test_static_view(self):
        response = self.client.get(reverse('faq'))
        self.assertTemplateUsed(response, 'static/faq.html')

    def test_static_view_custom_context(self):
        response = self.client.get(reverse('ubuntuone-account'))

        # this is known to be true for this static page
        self.assertTrue(response.context['use_app_links'])
