# Copyright 2010, 2012 Canonical Ltd.  This software is licensed under
# the GNU Affero General Public License version 3 (see the file
# LICENSE).

from django.conf import settings
from django.contrib import auth, messages
from django.contrib.auth.decorators import login_required
from django.core.urlresolvers import reverse
from django.template import RequestContext
from django.template.loader import render_to_string
from django.template.response import TemplateResponse
from django.http import (
    HttpResponseNotAllowed,
    HttpResponseRedirect,
)
from django.utils.decorators import method_decorator
from django.utils.translation import ugettext as _
from django.views.generic.base import View, TemplateResponseMixin

from gargoyle import gargoyle
from ssoclient.v2 import errors as api_errors

from identityprovider.apiutils import get_api_client
from identityprovider.forms import (
    PASSWORD_LEAKED,
    GenericEmailForm,
    NewAccountForm,
    TwoFactorSyncForm,
)
from identityprovider.login import AuthenticationError
from identityprovider.models.twofactor import sync, login
from identityprovider.models.account import Account
from identityprovider.stats import stats
from identityprovider.utils import (
    get_current_brand,
    polite_form_errors,
    redirection_url_for_token,
    add_non_field_error,
)
from identityprovider.views.utils import get_rpconfig_from_request

from webui.constants import (
    EMAIL_EXISTS_ERROR,
    VERIFY_EMAIL_MESSAGE,
)
from webui.decorators import (
    check_readonly,
    dont_cache,
    limitlogin,
    redirect_home_if_logged_in,
    require_twofactor_enabled,
    requires_cookies,
)
from webui.views.utils import (
    add_captcha_settings,
    display_email_sent,
    set_session_email,
)


ACCOUNT_CREATED = _('Your account was created successfully.')
CORRECT_ERRORS = _('Please correct the errors below.')
VERIFY_EMAIL_SENT = _('We have emailed %(email_to)s (from %(email_from)s), '
                      'please check your inbox to verify your email address.')

ERROR_PASSWORD_RESET_EMAIL_INVALIDATED = _(
    'The provided email address has been invalidated. It is no longer '
    'possible to reset the corresponding account\'s password.')
ERROR_PASSWORD_RESET_ACCOUNT_SUSPENDED = _(
    'The provided email address has been suspended. It is not possible '
    'to reset the corresponding account\'s password.')
ERROR_PASSWORD_RESET_ACCOUNT_DEACTIVATED = _(
    'The provided email address has been deactivated. It is not possible '
    'to reset the corresponding account\'s password.')
ERROR_PASSWORD_RESET_GENERIC = _(
    'It is not possible to reset this account\'s password.')
ERROR_PASSWORD_RESET_TOO_MANY_TOKENS = (
    'Too many active password reset requests. An email has already been sent '
    'to your account; please verify your inbox.')


@redirect_home_if_logged_in
@check_readonly
@requires_cookies
def new_account(request, token=None):
    captcha_required = (
        not gargoyle.is_active('OPTIONAL_CAPTCHA', request) or
        gargoyle.is_active('CAPTCHA', request)
    )
    captcha_error = ''
    captcha_error_message = None
    rpconfig = get_rpconfig_from_request(request, token)

    def collect_stats(key):
        stats.increment('flows.new_account', key=key, rpconfig=rpconfig)

    if request.method == 'GET':
        collect_stats('requested')
        form = NewAccountForm(initial={'email': request.GET.get('email')})

    elif request.method == 'POST':
        form = NewAccountForm(request.POST)
        if form.is_valid():
            data = dict((k, v) for k, v in form.cleaned_data.items()
                        if k in ('email', 'password', 'displayname'))
            data['captcha_id'] = request.POST.get(
                'recaptcha_challenge_field'
            )
            data['captcha_solution'] = request.POST.get(
                'recaptcha_response_field'
            )
            # we'll handle our own capture generation
            data['create_captcha'] = False

            url = redirection_url_for_token(token)

            try:
                api = get_api_client(request)
                api.register(data)

            except api_errors.InvalidData as e:
                # we shouldn't really get this, as the form will have validated
                # the data, but just in case...
                form._errors.update(e.extra)
                collect_stats('error.form')

            except api_errors.AlreadyRegistered:
                collect_stats('error.email')
                form._errors['email'] = [VERIFY_EMAIL_MESSAGE]

            except api_errors.CaptchaRequired as e:
                captcha_required = True
                collect_stats('captcha_required')

            except (api_errors.CaptchaFailure, api_errors.CaptchaError) as e:
                captcha_required = True
                captcha_error = '&error=' + e.extra.get('captcha_message', '')
                captcha_error_message = _('Incorrect captcha solution')
                collect_stats('error.captcha')

            else:
                collect_stats('success')

                # TODO: remove, but still needed atm
                set_session_email(request.session, data['email'])

                user = auth.authenticate(
                    username=data['email'],
                    password=data['password']
                )
                auth.login(request, user)

                msg_data = dict(email_to=data['email'],
                                email_from=settings.NOREPLY_FROM_ADDRESS)
                msgs = (ACCOUNT_CREATED, VERIFY_EMAIL_SENT % msg_data)
                messages.success(request, ' '.join(msgs), 'temporary')
                return HttpResponseRedirect(url)

        else:

            if get_current_brand() == 'ubuntuone':
                messages.error(request, CORRECT_ERRORS)

            collect_stats('error.form')

    context = {
        'form': form,
        'rpconfig': rpconfig,
        'token': token,
        'captcha_required': captcha_required,
        'captcha_error': captcha_error,
        'captcha_error_message': captcha_error_message,
    }
    if captcha_required:
        context = add_captcha_settings(context)

    context = RequestContext(request, context)

    if form.errors:
        err = form.errors.get('email', [''])[0]
        if err == VERIFY_EMAIL_MESSAGE:
            messages.error(request, EMAIL_EXISTS_ERROR)

        password_leaked = PASSWORD_LEAKED in form.errors.get('password', '')
        if password_leaked:
            error_msg = render_to_string('common/password_leaked_warning.txt',
                                         context)
            messages.warning(request, error_msg)

    template = 'registration/new_account.html'
    return TemplateResponse(request, template, context)


@check_readonly
@requires_cookies
def forgot_password(request, token=None):
    # Need this for displaying metrics per RP
    rpconfig = get_rpconfig_from_request(request, token)

    def collect_stats(key):
        stats.increment('flows.forgot_password', key=key, rpconfig=rpconfig)

    if request.method == 'GET':
        # track forgot password requests
        collect_stats('requested')
        form = GenericEmailForm(initial={'email': request.GET.get('email')})

    elif request.method == 'POST':
        form = GenericEmailForm(request.POST)
        if form.is_valid():
            email = form.cleaned_data['email']

            try:
                api = get_api_client(request)
                api.request_password_reset(email, token=token)
                set_session_email(request.session, email)

            except api_errors.InvalidData as e:
                for key, value in e.extra.items():
                    if not isinstance(value, list):
                        value = [value]
                    form.errors[key] = value
                collect_stats('error.form')

            except api_errors.EmailInvalidated as e:
                form.errors['__all__'] = [
                    ERROR_PASSWORD_RESET_EMAIL_INVALIDATED]
                collect_stats('error.email_invalidated')

            except api_errors.AccountSuspended as e:
                form.errors['__all__'] = [
                    ERROR_PASSWORD_RESET_ACCOUNT_SUSPENDED]
                collect_stats('error.account_suspended')

            except api_errors.AccountDeactivated as e:
                form.errors['__all__'] = [
                    ERROR_PASSWORD_RESET_ACCOUNT_DEACTIVATED]
                collect_stats('error.account_deactivated')

            except api_errors.CanNotResetPassword as e:
                form.errors['__all__'] = [ERROR_PASSWORD_RESET_GENERIC]
                collect_stats('error.can_not_reset_password')

            except api_errors.TooManyTokens as e:
                form.errors['__all__'] = [
                    ERROR_PASSWORD_RESET_TOO_MANY_TOKENS]
                collect_stats('error.too_many_tokens')

            else:
                msgs = {
                    'email_to': email,
                    'email_from': settings.NOREPLY_FROM_ADDRESS,
                    'support_form_url': settings.SUPPORT_FORM_URL,
                }

                if get_current_brand() == 'ubuntuone':
                    heading = 'Reset password'
                    reason = _(
                        "We have sent an email to %(email_to)s. To continue, "
                        "click on the link in your email, or enter the "
                        "confirmation code below."
                    ) % msgs
                else:
                    heading = 'Forgotten your password?'
                    reason = _(
                        "We've just emailed "
                        "%(email_to)s (from %(email_from)s) with "
                        "instructions on resetting your password.<br/><br/>"
                        "If the email address you provided has not been "
                        "verified we'll use your account's verified email "
                        "address instead. If you don't have a verified email "
                        "address please "
                        "<a href='%(support_form_url)s'>contact support</a>."
                    ) % msgs
                return display_email_sent(
                    request,
                    email,
                    heading,
                    reason,
                    _("Check that you&rsquo;ve actually "
                      "entered a subscribed email address."),
                    token=token,
                    rpconfig=rpconfig,
                )

        else:
            # track form errors
            collect_stats('error.form')
            polite_form_errors(form._errors)

    else:
        return HttpResponseNotAllowed(['GET', 'POST'])

    context = RequestContext(
        request, dict(form=form, rpconfig=rpconfig, token=token))
    template = 'registration/forgot_password.html'
    return TemplateResponse(request, template, context)


class Sync2faView(View, TemplateResponseMixin):
    template_name = 'registration/sync_2fa.html'
    hide_sidebar = True

    def _render_form(self, request, form, request_dict):
        context = RequestContext(request, {
            'form': form,
            'hide_sidebar': self.hide_sidebar,
            'next': request_dict.get('next'),
        })
        return self.render_to_response(context)

    def _sync_from_form(self, request, form):
        account = Account.objects.select_for_update().get(
            openid_identifier=request.user.username)
        otps = [form.cleaned_data[field] for field in form.fields.keys()]

        try:
            sync(account, otps)
        except AuthenticationError as exc:
            add_non_field_error(form, exc)
            return self._render_form(request, form, request.POST)
        else:
            login(request)
            next_url = request.POST.get('next')
            if not next_url:
                next_url = reverse('account-index')
            return HttpResponseRedirect(next_url)

    def get(self, request):
        form = TwoFactorSyncForm()
        return self._render_form(request, form, request.GET)

    def post(self, request):
        form = TwoFactorSyncForm(request.POST)

        if form.is_valid():
            return self._sync_from_form(request, form)
        else:
            return self._render_form(request, form, request.POST)

    @method_decorator(dont_cache)
    @method_decorator(limitlogin())
    # for this page, we need to be logged in but NOT twofactored
    @method_decorator(login_required)
    @method_decorator(require_twofactor_enabled)
    def dispatch(self, *args, **kwargs):
        return super(Sync2faView, self).dispatch(*args, **kwargs)


sync_2fa = Sync2faView.as_view()
